#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2001 Free Software Foundation
#
# FILE:
# GCommBase.py
#
# DESCRIPTION:
# Class that provides a basis for GNUe Comm Drivers.
#
# NOTES:
#


from gnue.common import GComm
from gnue.common import GDebug
import string, sys, thread



##############################################################################
#
# ClientAdapter
#
class Client:
  def request(self, service, params={}):
    pass

  def close():
    pass

  def runMethod(self, method, path, *args, **params):
    pass

  def setLocalArgumentChecking(self,enable):
    pass

  # Try to shut our services down
  def __del__(self):
    try:
      self.close()
    except:
      pass



##############################################################################
#
# ServerAdapter
#
class Server:
  def __init__(self, rpcdef, bindings, params):
    self._bindings = bindings
    self._rpcdef = rpcdef
    
  def allowObjectIntrospection(self,allow):
     # allow the remote client to get the type of an object by its handle
     # an to get other introspection details by that
     pass
   
  def serve(self):
    pass
  
  def serveAsNewThread(self):
    thread.start_new_thread(self.serve,())

  ## ObjectMethodChecking means, to check if the method of an object is
  ## included in the grpc
  ## in the service directory
  def setObjectMethodChecking(self,checking_on):
    self._objectMethodChecking=checking_on
    pass 
    
  def runService(self, method, data):
    pass
    
    
  #
  # Return an exception
  #
  def raiseException(self, exception, message, event=None):
    raise exception, message



##############################################################################
#
# ProxyObject
#
class ProxyObject:
  def __init__(self, adapter, attr, parent=None, **params):
    for param in params.keys():
      self.__dict__["_%s" % param] = params[param]
    self._adapter = adapter
    self._parent = parent
    self._attr = attr

    try:
      self._attrPath = parent._attrPath[:]
    except AttributeError:
      self._attrPath = []

    self._attrPath.append(attr)

    # print "Creating ProxyObject %s" % string.join(self._attrPath,'.')

    try:
      self._subobject = params['subobject']
    except KeyError:
      self._subobject = 0


  def __getattr__(self, attr):
    try:
      if attr[0] == '_':
        raise AttributeError, attr
      else:
        try:
          return self.__dict__[attr]
        except KeyError:
          po = self.__class__(self._adapter, attr, self,
                              subobject=1)
          self.__dict__[attr] = po
          return po

    except KeyError, mesg:
      raise AttributeError, mesg


  def __setattr__(self, attr, value):
    if attr[0] == '_':
      self.__dict__[attr] = value
    else:
      raise GComm.ProgrammingError




  def __call__(self, *args, **params):
    return self._adapter.runMethod(string.join(self._attrPath,'.'),
                                  *args, **params)
