#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2001 Free Software Foundation
#
# FILE:
# ObjectLibrarian.py
#
# DESCRIPTION:
# Helper module that keeps track of object references for use
# with non-Object interfaces
#
# NOTES:
#

import sha, time

class InvalidObjectReference(StandardError):
  pass

_objects = {}
_index = 0.0


def getObjectReference(object):
  try:
    return object._ObjectLibrarian__refid
  except AttributeError:

    # This is to introduce some obscurity to the id handles...
    global _index
    _index += time.clock()
    object._ObjectLibrarian__refid = \
        sha.new("%s%s%s"%(id(object), object, _index)).hexdigest()

    return object._ObjectLibrarian__refid


def archiveObject(object):
  _objects[getObjectReference(object)] = object


def retrieveObject(handle):
  try:
    return _objects[handle]
  except KeyError:
    raise InvalidObjectReference


def deferenceObject(object):
  try:
    del _objects[object._ObjectLibrarian__refid]
  except KeyError:
    pass
  except AttributeError:
    pass
