#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
# FILE:
# _dbsig/DBdriver.py
#
# DESCRIPTION:
# Generic implementation of dbdriver using Python DB-SIG v2
# specification.
#
# NOTES:
# The classes below are meant to be extended
#
# HISTORY:
#

from gnue.common import GDataObjects, GDebug, GConditions
import string
from gnue.common import GDebug
import types
import sys

class DBSIG_RecordSet(GDataObjects.RecordSet):

  def _haveHistory(self, table=None):
    if not table:
      auxTable = map(lambda x: x.strip(), self._parent._dataObject.table.split(','))
    else:
      auxTable = map(lambda x: x.strip(), table.split(','))
    for table in auxTable:
      if table.endswith('_data'):
        return 1
    return 0
    
  def getPostType(self):
    tmpPostType = ""
    if hasattr(self._parent._dataObject,"posttype"):
      tmpPostType =self._parent._dataObject.posttype.lower()
    return tmpPostType
    
  def _postChanges(self,autoCommit=1):
    if not self.isPending(): return
    auxTable = map(lambda x: x.strip(), self._parent._dataObject.table.split(','))

    if self._haveHistory() and self.getPostType() == "" or self.getPostType() == "transactional":
      return self._postChangesWHist(autoCommit)

    statements = []
    if self._deleteFlag:
      auxTable.reverse()
      for table in auxTable:
        statements.append(self._buildDeleteStatement(table))          
    elif self._insertFlag:
      for table in auxTable:
        statements.append(self._buildInsertStatement(table))          
    elif self._updateFlag:
      if len(auxTable) > 1:
        updTables = {}
        for iterField in self._modifiedFlags.keys():
          a = len (iterField.split ('.'))
          if (a == 2):
            [updTable, auxField]= iterField.split ('.')
            updTables[updTable] = ""
          else:
            # raise
            pass
        for iterTable in updTables.keys():
          statements.append(self._buildUpdateStatement(iterTable))            
      else:
        statements.append(self._buildUpdateStatement(auxTable[0]))          

    try:
      for statement in statements:
        GDebug.printMesg(5, "_postChanges: statement=%s" % statement)
        #print "statement: %s" % statement
        self._parent._update_cursor.execute(statement)
        # Set _initialData to be the just-now posted values
        if not self._deleteFlag:
          self._initialData = {}
          self._initialData.update(self._fields)
            
    except self._parent._dataObject._DatabaseError, err:
      GDebug.printMesg(1, "DBSIG_RecordSet._postChanges ERROR = %s " % err)
      print >> sys.stderr, "DBSIG_RecordSet._postChanges ERROR = %s " % err
      if autoCommit:
        self._parent._dataObject.rollback()
      raise GDataObjects.ConnectionError(str(err))

    if autoCommit:
      self._parent._dataObject.commit()
    
    self._updateFlag = 0
    self._insertFlag = 0
    self._deleteFlag = 0

    
    return 1

  def _postChangesWHist(self,autoCommit):
    auxTable = map(lambda x: x.strip(), self._parent._dataObject.table.split(','))
    auxHTable = None
    if hasattr(self._parent._dataObject,"historytable"):
      auxHTable =map(lambda x: x.strip(), self._parent._dataObject.historytable.split(','))
    statements = []
    try:
      if self._deleteFlag:
        auxTable.reverse()
        for table in auxTable:
          self._doDeleteStatementWHist(table)
      elif self._insertFlag:
        for table in auxTable:
          self._doInsertStatementWHist(table)
      elif self._updateFlag:
        if len(auxTable) > 1:
          updTables = {}
          for iterField in self._modifiedFlags.keys():
            a = len (iterField.split ('.'))
            if (a == 2):
              [updTable, auxField]= iterField.split ('.')
              updTables[updTable] = ""
            else:
              # raise
              pass
          for iterTable in updTables.keys():
            self._doDeleteStatementWHist(iterTable)
            self._doInsertStatementWHist(iterTable,1)
        else:
          self._doDeleteStatementWHist(auxTable[0])
          self._doInsertStatementWHist(auxTable[0],1)

    except self._parent._dataObject._DatabaseError, err:
      GDebug.printMesg(1, "DBSIG_RecordSet._postChangesWHist ERROR = %s " % err)
      print >> sys.stderr, "DBSIG_RecordSet._postChangesWHist ERROR = %s " % err
      if autoCommit:
        self._parent._dataObject.rollback()
      raise GDataObjects.ConnectionError(str(err))
          
    if autoCommit:
      self._parent._dataObject.commit()

    # Set _initialData to be the just-now posted values
    if not self._deleteFlag:
      self._initialData = {}
      self._initialData.update(self._fields)

    self._updateFlag = 0
    self._insertFlag = 0
    self._deleteFlag = 0

    return 1


  # If a vendor can do any of these more efficiently (i.e., use a known
  # PRIMARY KEY or ROWID, then override these methods. Otherwise, leave
  # as default.  Note that these functions are specific to DB-SIG based
  # drivers (i.e., these functions are not in the base RecordSet class)

  def _buildDeleteStatement(self, table):
    where = []
    if self._parent._dataObject._primaryKeys:
      for iterTable in self._parent._dataObject._primaryKeys.keys():
        if iterTable != table: continue
        for iterField in self._parent._dataObject._primaryKeys[iterTable]:
          auxWhere = "%s = '%s'"
          if self._initialData.has_key(table+"."+iterField):
            where.append(auxWhere % (iterField,self._initialData[table+"."+iterField]))
          else:
            where.append(auxWhere % (iterField,self._initialData[iterField]))            
    else:
      for field in self._initialData.keys():
        if self._parent.isFieldBound(field):
          a = len(field.split ('.'))
          if (a == 2):
            [auxTable, auxField]= field.split ('.')
            if string.count(self._parent._dataObject.table,',') and (auxTable != table):
              continue
          elif (a == 1):
            auxField= field
          else:
            # raise
            pass

          if self._initialData[field] == None:
            where.append ("%s IS NULL" % auxField)
          else:
            where.append ("%s=%s" % (auxField,
              self._parent._dataObject._toSqlString(self._initialData[field])))

    statement = "DELETE FROM %s WHERE %s" % \
       (table, string.join(where,' AND ') )
    return statement

  def _doDeleteStatementWHist(self, table):
    if not self._haveHistory(table): return 1
    today = "'%s'" % self._parent._dataObject.triggerExtensions.getTimeStamp()
    return self._parent._dataObject.triggerExtensions.atomicModify(today, \
                                                                   "%s._end_t" % (table), \
                                                                   self._parent._dataObject, \
                                                                   self, \
                                                                   "_end_t IS NULL",0)


  def _buildInsertStatement(self, table):
    vals = []
    fields = []
    
    pkFields = []
    if self._parent._dataObject._primaryKeys:
      pkFields.extend(self._parent._dataObject._primaryKeys[table])
    for field in self._fields.keys():
      if self._parent.isFieldBound(field):
        a = len (field.split ('.'))
        if (a == 2):
          [auxTable, auxField] = field.split ('.')
          if string.count(self._parent._dataObject.table,',') and (auxTable != table):
            continue
        elif (a == 1):
          auxField= field
        else:
          # raise
          pass

        if pkFields.count(auxField) > 0 : continue

        #try:
        #  if self._fields[field]:
        #    pkFields.remove(auxField)
        #  else:
        #    continue
        #except ValueError:
        #  pass
        
        fields.append(auxField)
        if self._fields[field] == None or self._fields[field] == '':
          vals.append ("NULL") #  % (self._fields[field]))
        else:
          vals.append (self._parent._dataObject._toSqlString(self._fields[field]))

    for iterField in pkFields:
      fields.append(iterField)
      tmpTable = table[0:24]
      seqValue = str(self._parent._dataObject.triggerExtensions.getSequence(tmpTable+"_"+iterField+"_seq"))
      vals.append(seqValue)
      self._fields[table+'.'+iterField] = seqValue

    for iterField in self._parent._dataObject._joinConditions.keys():
      [auxTable, auxField] = iterField.split('.')
      if auxTable == table and fields.count(auxField) == 0:
        fields.append(auxField)
        vals.append (self._parent._dataObject._toSqlString(\
            self._fields[self._parent._dataObject._joinConditions[iterField]]))

    if self._haveHistory(table):
      fields.append("_start_t")
      vals.append("'%s'" % self._parent._dataObject.triggerExtensions.getTimeStamp())

    return "INSERT INTO %s (%s) VALUES (%s)" % \
                  (table, string.join(fields,','), \
                   string.join(vals,',') )
                  
  def _doInsertStatementWHist(self, table, forUpdate=0):
    statementElement = self._buildInsertStatementWHist(table,forUpdate=forUpdate)
    statement = "INSERT INTO %s (%s) VALUES (%s)" % \
                (statementElement[0], string.join(statementElement[1],','), \
                 string.join(statementElement[2],',') ) 

    try:
      GDebug.printMesg(5, "_doInsertStatementWHist: statement=%s" % statement)
      self._parent._update_cursor.execute(statement)
      
    except self._parent._dataObject._DatabaseError, err:
      GDebug.printMesg(1, "_doInsertStatementWHist: ERROR = %s" % err)
      print >> sys.stderr, "DBSIG_RecordSet._doInsertStatementWHist: ERROR = %s " % err
      raise GDataObjects.ConnectionError(str(err))
 
    return 1

  def _buildInsertStatementWHist(self, table, field=None,forUpdate=0):
    if field and field.count('.') > 0:
      field = field.split('.')[1]
      
    vals = []
    fields = []
    
    pkFields = []
    if self._parent._dataObject._primaryKeys:
      pkFields.extend(self._parent._dataObject._primaryKeys[table])
    for iterField in self._fields.keys():
      if self._parent.isFieldBound(iterField):
        a = len (iterField.split ('.'))
        if (a == 2):
          [auxTable, auxField]= iterField.split ('.')
          if string.count(self._parent._dataObject.table,',') and (auxTable != table):
            continue
        elif (a == 1):
          auxField = iterField
        else:
          # raise
          pass

        if pkFields.count(auxField) > 0 : continue
        if field and field == auxField :
          continue

        #try:
        #  if self._fields[field]:
        #    pkFields.remove(auxField)
        #  else:
        #    continue
        #except ValueError:
        #  pass

        fields.append(auxField)
        if self._fields[iterField] == None or self._fields[iterField] == '':
          vals.append ("NULL") #  % (self._fields[iterField]))
        else:
          vals.append (self._parent._dataObject._toSqlString(self._fields[iterField]))

    for iterField in pkFields:
      fields.append(iterField)
      tmpTable = table[0:24]
      seqValue = str(self._parent._dataObject.triggerExtensions.getSequence(tmpTable+"_"+iterField+"_seq"))
      vals.append(seqValue)
      self._fields[table+'.'+iterField] = seqValue

    for iterField in self._parent._dataObject._joinConditions.keys():
      [auxTable, auxField] = iterField.split('.')
      if auxTable == table and fields.count(auxField) == 0:
        fields.append(auxField)
        vals.append (self._parent._dataObject._toSqlString(\
            self._fields[self._parent._dataObject._joinConditions[iterField]]))

    if self._haveHistory(table):
      fields.append("_start_t")
      if forUpdate:
        vals.append("'%s'" % self._parent._dataObject.triggerExtensions.getTimeStamp())
      else:
        vals.append("'1970-01-01'")

    statementElement = []
    statementElement.append(table)
    statementElement.append(fields)
    statementElement.append(vals)

    return statementElement
  
  def _buildUpdateStatement(self, table):
    updates = []
    for field in self._modifiedFlags.keys():
      a = len (field.split ('.'))
      if (a == 2):
        [auxTable, auxField]= field.split ('.')
        if string.count(self._parent._dataObject.table,',') and (auxTable != table):
          continue
      elif (a == 1):
        auxField= field
      else:
        # raise
        pass
      try:
          updates.append ("%s=%s" % (auxField, self._parent._dataObject._toSqlString(self._fields[field])))

      except KeyError:
        pass
    where = []
    if updates:
      if self._parent._dataObject._primaryKeys:
        for iterTable in self._parent._dataObject._primaryKeys.keys():
          if iterTable != table: continue
          for iterField in self._parent._dataObject._primaryKeys[iterTable]:
            auxWhere = "%s = '%s'"
            where.append(auxWhere % (iterField,self._initialData[table+"."+iterField]))
      else:
        for field in self._initialData.keys():
          a = len (field.split ('.'))
          if (a == 2):
            [auxTable, auxField]= field.split ('.')
            if string.count(self._parent._dataObject.table,',') and (auxTable != table):
              continue
          elif (a == 1):
            auxField= field
          else:
            # raise
            pass

          if self._initialData[field] == None:
            where.append ("%s IS NULL" % auxField)
          else:
            where.append ("%s=%s" % (auxField, self._parent._dataObject._toSqlString(self._initialData[field])))

    return "UPDATE %s SET %s WHERE %s" % \
       (table, string.join(updates,','), \
        string.join(where,' AND ') )


class DBSIG_ResultSet(GDataObjects.ResultSet):
  def __init__(self, dataObject, cursor=None, \
        defaultValues={}, masterRecordSet=None, fieldNames=None):
    GDataObjects.ResultSet.__init__(
           self,dataObject,cursor,defaultValues,masterRecordSet)
    self._recordSetClass = DBSIG_RecordSet
    self._fieldNames = []

    if self._cursor and self._cursor.description:
      if fieldNames:
        self._fieldNames = fieldNames
      else:
        for t in(self._cursor.description):
          self._fieldNames.append(t[0])
      GDebug.printMesg(5, "Field names set to %s" % self._fieldNames)

    self._recordCount = cursor.rowcount or 0

    GDebug.printMesg(5, 'ResultSet created')

  def _loadNextRecord(self):
    if self._cursor and self._cursor.description:
      rs = None

      try:
        rsets = self._cursor.fetchmany()
      except self._dataObject._DatabaseError, err:
        raise GDataObjects.ConnectionError(str(err))
      if rsets and len(rsets):
        for rs in(rsets):
          if rs:
            i = 0
            dict = {}
            for f in (rs):
              dict[string.lower(self._fieldNames[i])] = f
              i += 1
              
            self._cachedRecords.append (self._recordSetClass(parent=self, \
                                                             initialData=dict))
          else:
            return 0
        return 1
      else:
        return 0
    else:
      return 0


class DBSIG_DataObject(GDataObjects.DataObject):
  import re
  
  reJoinToTable = re.compile("(?:^[( ]*|JOIN\s)\s*(\w+(?:\s+as\s+\w+)?)", re.I)

  conditionElements = {
       'add':             (2, 999, '(%s)',                       '+'      ),
       'sub':             (2, 999, '(%s)',                       '-'      ),
       'mul':             (2, 999, '(%s)',                       '*'      ),
       'div':             (2, 999, '(%s)',                       '/'      ),
       'and':             (1, 999, '(%s)',                       ' AND '  ),
       'or':              (2, 999, '(%s)',                       ' OR '   ),
       'not':             (1,   1, '(NOT %s)',                   None     ),
       'negate':          (1,   1, '-%s',                        None     ),
       'null':            (1,   1, '(%s IS NULL)',               None     ),
       'notnull':         (1,   1, '(%s IS NOT NULL)',           None     ),
       'eq':              (2,   2, '(%s = %s)',                  None     ),
       'ne':              (2,   2, '(%s != %s)',                 None     ),
       'gt':              (2,   2, '(%s > %s)',                  None     ),
       'ge':              (2,   2, '(%s >= %s)',                 None     ),
       'lt':              (2,   2, '(%s < %s)',                  None     ),
       'le':              (2,   2, '(%s <= %s)',                 None     ),
       'like':            (2,   2, '(%s LIKE %s)',               None     ),
       'notlike':         (2,   2, '(%s NOT LIKE %s)',           None     ),
       'between':         (3,   3, '(%s BETWEEN %s AND %s)',     None     ),
       'notbetween':      (3,   3, '(%s NOT BETWEEN %s AND %s)', None     ),
       # These two are hacks... these are not really valid tags
       # Used when the 2nd value of EQ or NE is NULL.
       # These are probably not necessary now that null and not null exist
       '__iseq':          (2,   2, '(%s IS %s)',                 None     ),
       '__isne':          (2,   2, '(%s IS NOT %s)',             None     )}

  def __init__(self, strictQueryCount=1):
    GDataObjects.DataObject.__init__(self)

    GDebug.printMesg (1,"DB-SIG database driver backend initializing")

    # This should be set to the Single Quote escape character
    # (typically either "'" or "\"
    self._escapeSingleQuote = "'"
    self._resultSetClass = DBSIG_ResultSet
    self._DatabaseError = None
    self._strictQueryCount = strictQueryCount
    self._joinConditions = {}
    self._joinConditionsLoded = None
    
    # If a DB driver supports a unique identifier for rows,
    # list it here.  _primaryIdField is the field name (lower case)
    # that would appear in the recordset (note that this can be
    # a system generated format). If a primary id is supported,
    # _primaryIdFormat is the WHERE clause to be used. It will have
    # the string  % (fieldvalue) format applied to it.
    #
    # See Oracle drivers for example
    
    self._primaryKeys = {}
    #self._primaryIdField = None         # Internal recordset field name (lowercase!!!)
    #self._primaryIdSelect = ""  # Select clause
    #self._primaryIdFormat = "__gnue__ = '%s'"         # Where clause format

    # Internal flag to avoid consistently doing the same check
    # If this is set to 1 initially, then the
    self._primaryIdChecked = 0  # Internal flag


    # The date/time format used in insert/select statements
    # (based on format used for time.strftime())
    self._dateTimeFormat = "'%c'"


  def _toSqlString(self, value):
    try:
      return value.strftime(self._dateTimeFormat)
    except AttributeError:
      if value == None:
        return "NULL"
      elif type(value) == types.FloatType:
        return str(value)
      else:
        return "'%s'" % string.replace(str(value),"'",
                    "%s'" % self._escapeSingleQuote)


  # This should be over-ridden only if driver needs more than user/pass
  def getLoginFields(self):
    return [['_username', 'User Name',0],['_password', 'Password',1]]


  # Used by drivers with a unique id (like rowid) (see Oracle for example)
  def _checkForPrimaryId(self):
    self._primaryIdChecked = 1

  # Ejecuta un select crudo y devuelve un RecordSet.
  def _createRawResultSet(self, statement):
    rs = None
    try:
      cursor = self._dataConnection.cursor()
      cursor.execute(statement)
      rs = self._resultSetClass(self, cursor=cursor, masterRecordSet=None)
    except self._DatabaseError, err:
      raise GDataObjects.ConnectionError(str(err))

    return rs

  def _createResultSet(self, conditions={}, readOnly=0, masterRecordSet=None):
    if self._outerJoinCondition and not self._outerJoin:
      self._outerJoin = self._outerJoinConditionToSQL(self._outerJoinCondition)
      import re
      auxStr = self._outerJoin[:]
      self.table = ", ".join(self.reJoinToTable.findall(auxStr)) # Extrae las tablas a partir del self._outerJoin mediante
                                                                 # una expresin regular (gentileza de John)
    if not hasattr(self,'table'):
      return self._resultSetClass(self, cursor=self._dataConnection.cursor(), masterRecordSet=masterRecordSet)
    

    # Used by drivers with a unique id (like rowid)
    if not self._primaryIdChecked: self._checkForPrimaryId()

    try:
      cursor = self._dataConnection.cursor()

      # pull a record count for the upcomming query
      if self._strictQueryCount:
        recordCount = self._getQueryCount(conditions)

      cursor.arraysize = self.cache
      tableCount = string.count(self.table,',')+1
      tables = map(lambda x: x.strip(), self.table.split(','))
      
      if not self._prefixedFields:
        import gnue.common.GConditions as GConditions
        tmpCond = GConditions.GCondition()
        ceq = GConditions.GCeq(tmpCond)
        GConditions.GCConst(ceq,1,"number")
        GConditions.GCConst(ceq,0,"number")
        if self._primaryKeys and self._fieldReferences:
          for table in self._primaryKeys.keys():
            for field in self._primaryKeys[table]:
              if not self._fieldReferences.has_key(table+"."+field) and not self._fieldReferences.has_key(field):
                self._fieldReferences[table+"."+field] = ""
              
          self._prefixedFields = self._fieldReferences.keys()
  
        if tableCount > 1:
          if not self._prefixedFields:
            self._prefixedFields = []
            for iterTable in tables:
              cursor.execute(self._buildQuery(tmpCond,iterTable))
              if iterTable.count(' as '):
                iterTable = map(lambda x: x.strip(), iterTable.split(' as '))[1] #Esto es por si tiene ej: "from entity as client"
              for iterDesc in (cursor.description):
                self._prefixedFields.append(iterTable+'.'+iterDesc[0])
  
      #if masterRecordSet:
      #  print self._conditionToSQL(conditions)
      cursor.execute(self._buildQuery(conditions))


    except self._DatabaseError, err:
      raise GDataObjects.ConnectionError(str(err))

    if not self._prefixedFields:
      rs = self._resultSetClass(self, cursor=cursor, masterRecordSet=masterRecordSet)
    else:
      rs = self._resultSetClass(self, cursor=cursor, masterRecordSet=masterRecordSet, fieldNames=self._prefixedFields)

    if self._strictQueryCount:
      rs._recordCount = recordCount
    if readOnly:
      rs._readonly = readOnly

    return rs

  def _getQueryCount(self,conditions={}):
    cursor = self._dataConnection.cursor()

    cursor.execute(self._buildQueryCount(conditions))
    rs = cursor.fetchone()
    return int(rs[0])

  def commit(self):
    GDebug.printMesg (5,"DB-SIG database driver: commit()")

    try:
      self._dataConnection.commit()
      # Ahora seteamos la coneccin para que pase a modo "Read committed"
      self._dataConnection.cursor().execute("SET TRANSACTION ISOLATION LEVEL READ COMMITTED")
    except self._DatabaseError, value:
      raise GDataObjects.ConnectionError(str(err))
    
    self._beginTransaction()

  def rollback(self):
    GDebug.printMesg (5,"DB-SIG database driver: rollback()")

    try:
      self._dataConnection.rollback()
      # Ahora seteamos la coneccin para que pase a modo "Read committed"
      self._dataConnection.cursor().execute("SET TRANSACTION ISOLATION LEVEL READ COMMITTED")
    except:
      pass	# I'm SURE this isn't right (jcater)
                # But not all db's support transactions

    self._beginTransaction()


  # Used to convert a joinCondition tree to an sql from clause
  def _outerJoinConditionToSQL (self, element):
    result = ""
    if isinstance(element,GConditions.GJoin):
      nextJoin = None
      condition = None
      rightTable = None
      leftTable = None
      joinCondType = element.type.upper()
      for i in range(0, len(element._children)):
        if isinstance(element._children[i],GConditions.GCRightTable):
          rightTable = element._children[i].name
        elif isinstance(element._children[i],GConditions.GCLeftTable):
          leftTable = element._children[i].name
        elif isinstance(element._children[i],GConditions.GJoin):
          nextJoin = element._children[i]
        elif isinstance(element._children[i],GConditions.GCondition):
          condition = self._conditionToSQL(element._children[i]).replace('WHERE','ON')
      if rightTable and leftTable:
        result = " (%s %s JOIN %s %s) " % (leftTable,joinCondType,rightTable,condition)
        return result
      elif rightTable:
        result = " (%s %s JOIN %s %s) " % (self._outerJoinConditionToSQL(nextJoin),joinCondType,rightTable,condition)
        return result
      elif leftTable:
        result = " (%s %s JOIN %s %s) " % (leftTable,joinCondType,self._outerJoinConditionToSQL(nextJoin),condition)
        return result
      else:
        raise GConditions.ConditionNotSupported
       
    if element:
      return self._outerJoinConditionToSQL(element._children[0])

  # Used to convert a condition tree to an sql where clause
  def _conditionToSQL (self, condition):
    if condition == {} or condition == None:
      return ""
    elif type(condition) == types.DictType:
      cond = GConditions.buildConditionFromDict(condition)
    else:
      cond = condition

    if not len(cond._children):
      return ""
    elif len(cond._children) > 1:
      chillun = cond._children[:]
      cond._children = []
      _and = GConditions.GCand(cond)
      _and._children = chillun

    where = " WHERE %s" % (self.__conditionToSQL (cond._children[0]))
    self._joinConditionsLoded = 1
    GDebug.printMesg(5, where)
    return where

  #
  # Used internally by _conditionToSQL
  #
  # This code recursively travels down a condition tree replacing the objects
  # with a strings representation
  def __conditionToSQL (self, element):
    if type(element) != types.InstanceType:
      return "%s" % element
    else:
      # Note that we strip the GC from the object types and lowercase the rest
      otype = string.lower(element._type[2:])
      #print "Otype: ",otype
      if not self._joinConditionsLoded and \
         otype == "eq" and \
         len(element._children) == 2 and \
         string.lower(element._children[0]._type[2:]) == "cfield" and \
         string.lower(element._children[1]._type[2:]) == "cfield":
        [auxTable, auxField]= element._children[0].name.split ('.')
        try:
          pKeys = self._primaryKeys[auxTable]
          for pKey in pKeys:
            if pKey == auxField:
              self._joinConditions[element._children[1].name] = element._children[0].name
              break
        except KeyError:
          pass
        if not self._joinConditions.has_key(element._children[1].name):
          self._joinConditions[element._children[0].name] = element._children[1].name

      if otype == 'cfield':
        return "%s" % element.name
      elif otype == 'cconst':
        if element.value == None:
          return "NULL"
        elif element.type == 'number':
          return "%s" % element.value
        else:
          return self._toSqlString(element.value)
      elif otype == 'cparam':
        v = element.getValue()
        return (v == None and "NULL") or ("'%s'" % element.getValue())
      elif self.conditionElements.has_key(otype):
        for i in range(0, len(element._children)):
          element._children[i] = self.__conditionToSQL(element._children[i])
        if len(element._children) == 2 and \
           otype in ('eq','ne') and \
           element._children[1] == 'NULL':
           otype = "__is%s" % otype
        if len(element._children) < self.conditionElements[otype][0]:
          raise GConditions.ConditionError, \
            _('Condition element "%s" expects at least %s arguments; found %s') % \
                (otype, self.conditionElements[otype][0], len(element._children))
        if len(element._children) > self.conditionElements[otype][1]:
          raise GConditions.ConditionError, \
            _('Condition element "%s" expects at most %s arguments; found %s') % \
                (otype, self.conditionElements[otype][0], len(element._children))
        if self.conditionElements[otype][3] == None:
          return self.conditionElements[otype][2] % tuple(element._children)
        else:
          return self.conditionElements[otype][2] % \
            (string.join(element._children, self.conditionElements[otype][3]))
      else:
        raise GConditions.ConditionNotSupported, \
          _('Condition clause "%s" is not supported by this db driver.') % otype

  # Code necessary to force the connection into transaction mode...
  # this is usually not necessary (MySQL is one of few DBs that must force)
  def _beginTransaction(self):
    pass


class DBSIG_DataObject_Object:
  def __init__(self):
   # TODO: A dummy placeholder to let old db driver work
   # TODO: each can have their *_DataObject_Object object from __init__ing
   # TODO: DBSIG_DataObject_Object
   GDebug.printMesg(0, "Database driver needs updated to not initialize DBSIG_DataObject_Object")

  def _buildQuery(self, conditions={}, qTable=None):
    if not qTable:
      qTable = self.table
    GDebug.printMesg(7,'Implicit Fields: %s' % self._fieldReferences)

    if self._fieldReferences:
      # el prximo if pregunta si tiene composicin externa la consulta que se quiere realizar
      # y adems pregunta si en la condicin existe "1=0" porque es una consdicin (truco) que
      # se utiliza para averiguar los campos de una tabla y es solo para inicializar el objeto
      # que contiene el resultado de la cosulta y no para consultar realmente
      if self._outerJoin and not self._conditionToSQL(conditions).replace(" ","").count("1=0"):
        q = "SELECT %s FROM %s%s" % \
            (string.join(self._fieldReferences.keys(),","), self._outerJoin,
             self._conditionToSQL(conditions))
      else:
        q = "SELECT %s FROM %s%s" % \
            (string.join(self._fieldReferences.keys(),","), qTable,
             self._conditionToSQL(conditions))
      #if qTable.count(',') != 0:
      #  print "q in buildQuery with fieldRef",q
    else:
      self._primaryIdSelect = None
      # el prximo if pregunta si tiene composicin externa la consulta que se quiere realizar
      # y adems pregunta si en la condicin existe "1=0" porque es una consdicin (truco) que
      # se utiliza para averiguar los campos de una tabla y es solo para inicializar el objeto
      # que contiene el resultado de la cosulta y no para consultar realmente
      if self._outerJoin and not self._conditionToSQL(conditions).replace(" ","").count("1=0"):
        q = "SELECT * FROM %s%s" % (self._outerJoin, self._conditionToSQL(conditions))
      else:
        q = "SELECT * FROM %s%s" % (qTable, self._conditionToSQL(conditions))
      #if qTable.count(',') != 0:
      #  print "q in buildQuery without fieldRef",q

    if hasattr(self,'order_by'):
     q = "%s ORDER BY %s " % (q, self.order_by)

    GDebug.printMesg(5,q)
    return q

  def _buildQueryCount(self, conditions={}):
    if self._outerJoin:
      q = "SELECT count(*) FROM %s%s" % (self._outerJoin, self._conditionToSQL(conditions))
    else:
      q = "SELECT count(*) FROM %s%s" % (self.table, self._conditionToSQL(conditions))

    GDebug.printMesg(5,q)

    return q

class DBSIG_DataObject_SQL:
  def _buildQuery(self, conditions={}):
    # Obviously, this is in a pre-alpha state :)
    return "select zipcode, city, state from zipcode order by zipcode desc"

supportedDataObjects = {
  'object': DBSIG_DataObject_Object,
  'sql':    DBSIG_DataObject_SQL
}





