#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it 
# and/or modify it under the terms of the GNU General Public 
# License as published by the Free Software Foundation; either 
# version 2, or(at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be 
# useful, but WITHOUT ANY WARRANTY; without even the implied 
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public 
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000, 2001 Free Software Foundation
#
# FILE:
# _pgsql/DBdriver.py
#
# DESCRIPTION:
# A core Postgresql implementation of dbdriver the other
# postgresql drivers can extend
#
# NOTES:
#


from string import lower, join
import sys
import string
from gnue.common import GDebug, GDataObjects
from gnue.common.dbdrivers._dbsig.DBdriver \
   import DBSIG_RecordSet, DBSIG_ResultSet, DBSIG_DataObject, \
          DBSIG_DataObject_Object

#from gnue.common.dbdrivers._dbsig.DBdriver import DBSIG_DataObject_Object as PGSQL_DataObject_Object
from gnue.common.dbdrivers._dbsig.DBdriver import DBSIG_DataObject_SQL as PGSQL_DataObject_SQL

class PGSQL_RecordSet(DBSIG_RecordSet):
   pass
#   def _buildUpdateStatement(self, table):
#     updates = []
#     for field in self._modifiedFlags.keys():
#       a = len (field.split ('.'))
#       if (a == 2):
#         [auxTable, auxField]= field.split ('.')
#         if string.count(self._parent._dataObject.table,',') and (auxTable != table):
#           continue
#       elif (a == 1):
#         auxField= field
#       else:
#         # raise
#         pass

#       updates.append ("%s=%s" % (auxField,
#                                  self._parent._dataObject._toSqlString(self._fields[field])))

#     where = []
    
#     if updates:
#       if self._parent._dataObject._primaryKeys:
#         for iterTable in self._parent._dataObject._primaryKeys.keys():
#           if iterTable != table: continue
#           for iterField in self._parent._dataObject._primaryKeys[iterTable]:
#             auxWhere = "%s = '%s'"
#             where.append(auxWhere % (iterField,self._initialData[table+"."+iterField]))
#       else:
#         for field in self._initialData.keys():
#           a = len (field.split ('.'))
#           if (a == 2):
#             [auxTable, auxField]= field.split ('.')
#             if string.count(self._parent._dataObject.table,',') and (auxTable != table):
#               continue
#           elif (a == 1):
#             auxField= field
#           else:
#             # raise
#             pass

#           if self._initialData[field] == None:
#             where.append ("%s IS NULL" % auxField)
#           else:
#             where.append ("%s=%s" % (auxField, self._parent._dataObject._toSqlString(self._initialData[field])))

#     updateStmt = "BEGIN; SELECT %s FROM %s WHERE %s FOR UPDATE; " % \
#                  (string.join(updates,','), table, \
#                   string.join(where,' AND ') ) + \
#                  "UPDATE %s SET %s WHERE %s" % \
#                  (table, string.join(updates,','), \
#                   string.join(where,' AND ') ) + "; COMMIT;"

#     return updateStmt

class PGSQL_ResultSet(DBSIG_ResultSet):
  def __init__(self, dataObject, cursor=None, defaultValues={}, masterRecordSet=None, fieldNames=None):
    DBSIG_ResultSet.__init__(self, dataObject, \
                             cursor, defaultValues, masterRecordSet,fieldNames)
    self._recordSetClass = PGSQL_RecordSet

class PGSQL_DataObject(DBSIG_DataObject):

  conditionElements = {
       'add':             (2, 999, '(%s)',                       '+'      ),
       'sub':             (2, 999, '(%s)',                       '-'      ),
       'mul':             (2, 999, '(%s)',                       '*'      ),
       'div':             (2, 999, '(%s)',                       '/'      ),
       'and':             (1, 999, '(%s)',                       ' AND '  ),
       'or':              (2, 999, '(%s)',                       ' OR '   ),
       'not':             (1,   1, '(NOT %s)',                   None     ),
       'negate':          (1,   1, '-%s',                        None     ),
       'null':            (1,   1, '(%s IS NULL)',               None     ),
       'notnull':         (1,   1, '(%s IS NOT NULL)',           None     ),
       'eq':              (2,   2, '(%s = %s)',                  None     ),
       'ne':              (2,   2, '(%s != %s)',                 None     ),
       'gt':              (2,   2, '(%s > %s)',                  None     ),
       'ge':              (2,   2, '(%s >= %s)',                 None     ),
       'lt':              (2,   2, '(%s < %s)',                  None     ),
       'le':              (2,   2, '(%s <= %s)',                 None     ),
       'like':            (2,   2, '(%s ILIKE %s)',              None     ),
       'notlike':         (2,   2, '(%s NOT ILIKE %s)',          None     ),
       'between':         (3,   3, '(%s BETWEEN %s AND %s)',     None     ),
       'notbetween':      (3,   3, '(%s NOT BETWEEN %s AND %s)', None     ),
       # These two are hacks... these are not really valid tags
       # Used when the 2nd value of EQ or NE is NULL.
       # These are probably not necessary now that null and not null exist
       '__iseq':          (2,   2, '(%s IS %s)',                 None     ),
       '__isne':          (2,   2, '(%s IS NOT %s)',             None     )}

  def __init__(self, pgdriver=None, pgresultset=None):
    DBSIG_DataObject.__init__(self)
    self._connectString = 'host=%s dbname=%s user=%s password=%s port=%s'
    self._escapeSingleQuote = '\\'
    if pgdriver:
      self._pgdriver = pgdriver
      self._DatabaseError = self._pgdriver.DatabaseError
    if pgresultset:
      self._resultSetClass = pgresultset
    
  def connect(self, connectData={}):
    GDebug.printMesg(1,"Postgresql database driver initializing")
    try:
#
#  Ugly dneighbo hack as no python studs to ask questions for
# 
      try:
        port = connectData['port']
      except:
        port  = '5432'

#
#  End hack (note port variable used below and defined in __init above)
#
      self._dataConnection = self._pgdriver.connect(self._connectString %
                                                    (connectData['host'],
                                                     connectData['dbname'],
                                                     connectData['_username'],
                                                     connectData['_password'],
                                                     port))
      self.triggerExtensions = PGSQL_TriggerExtensions(self._dataConnection)
    except self._DatabaseError, value:
      GDebug.printMesg(1,"Connect String: %s" % (self._connectString % \
             (connectData['host'],
              connectData['dbname'],
              connectData['_username'],
              connectData['_password'],
              port)))
      GDebug.printMesg(1,"Exception %s " % value)
      raise GDataObjects.LoginError, value

    try:
      encoding = connectData['encoding']
      GDebug.printMesg(1,'Setting postgresql client_encoding to %s' % encoding)
      cursor = self._dataConnection.cursor()
      cursor.execute("SET CLIENT_ENCODING TO '%s'" % encoding)
      cursor.close()
    except KeyError:
      pass
    except self._DatabaseError: 
      try: 
        cursor.close()
      except: 
        pass

    self._postConnect()

  def _postConnect(self):
    self.triggerExtensions = PGSQL_TriggerExtensions(self._dataConnection)

  #
  # Schema (metadata) functions
  #

  # Return a list of the types of Schema objects this driver provides
  def getSchemaTypes(self):
    return [('view','View',1), ('table','Table',1)]

  # Return a list of Schema objects
  def getSchemaList(self, type=None):
    includeTables = (type in ('table','sources', None))
    includeViews = (type in ('view','sources', None))

    inClause = []
    if includeTables:
      inClause.append ("'r'")
    if includeViews:
      inClause.append ("'v'")

    # TODO: This excludes any system tables and views. Should it?
    statement = "select relname, relkind, oid from pg_class " + \
            "where relkind in (%s) " % (join(inClause,',')) + \
            "and relname not like 'pg_%' " + \
            "order by relname"

    cursor = self._dataConnection.cursor()
    cursor.execute(statement)

    list = []
    for rs in cursor.fetchall():
      list.append(GDataObjects.Schema(attrs={'id':rs[2], 'name':rs[0],
                         'type':rs[1] == 'v' and 'view' or 'table'},
                         getChildSchema=self.__getFieldSchema))

    cursor.close()
    return list


  # Find a schema object with specified name
  def getSchemaByName(self, name, type=None):
    statement = "select relname, relkind, oid from pg_class " + \
            "where relname = '%s'" % (name)

    cursor = self._dataConnection.cursor()
    cursor.execute(statement)

    rs = cursor.fetchone()
    if rs:
      schema = GDataObjects.Schema(attrs={'id':rs[2], 'name':rs[0],
                           'type':rs[1] == 'v' and 'view' or 'table'},
                           getChildSchema=self.__getFieldSchema)
    else:
      schema = None

    cursor.close()
    return schema

  # Get fields for a table
  def __getFieldSchema(self, parent):

    statement = "select attname, pg_type.oid, typname, " + \
            " attnotnull, atthasdef, atttypmod " + \
            "from pg_attribute, pg_type " + \
            "where attrelid = %d and " % (parent.id) + \
            "pg_type.oid = atttypid and attnum >= 0" + \
            "order by attnum"

    cursor = self._dataConnection.cursor()
    cursor.execute(statement)

    list = []
    for rs in cursor.fetchall():

      attrs={'id': rs[1], 'name': rs[0],
             'type':'field', 'nativetype': rs[2],
             'required': rs[3] and not rs[4]}

      if rs[2] in ('int8','int2','int4','numeric',
                   'float4','float8','money','bool'):
        attrs['datatype']='number'
      elif rs[2] in ('date','time','timestamp','abstime','reltime'):
        attrs['datatype']='date'
      else:
        attrs['datatype']='text'

      if rs[5] != -1:
        attrs['length'] = rs[5]

      list.append(GDataObjects.Schema(attrs=attrs))

    cursor.close()
    return list

class PGSQL_DataObject_Object(PGSQL_DataObject, \
      DBSIG_DataObject_Object):

  def __init__(self):
    PGSQL_DataObject.__init__(self)
    self._primaryIdChecked = 0


  def _buildQuery(self, conditions={}, qTable=None):
    if not qTable:
      qTable = self.table
    return DBSIG_DataObject_Object._buildQuery(self, conditions, qTable)

  #def _haveHistory(self, table=None):
  #  if not table:
  #    auxTable = string.split(string.replace(self.table,' ',''),',')
  #  else:
  #    auxTable = string.split(string.replace(table,' ',''),',')
  #  for table in auxTable:
  #     statement = "select a.attname from pg_class c, pg_attribute a "
  #     statement += "where c.relname = '%s' and a.attnum > 0 and a.attrelid = c.oid and a.attname = '_table' " % table
  #     try:
  #        cursor = self._dataConnection.cursor()
  #        cursor.execute(statement)
  #        rowCount = cursor.rowcount 
  #        cursor.close()
  #        if rowCount > 0:
  #           return rowCount
  #     except self._DatabaseError, err:
  #        print "DBSIG_RecordSet._haveHistory ERROR %s " % err
  #        pass
  #  return 0

  
  def _checkForPrimaryId(self):
    self._primaryIdChecked = 1

    if not hasattr(self,'table'):
       return
    #To get the names you can start a second query like:
    tables = map(lambda x: x.strip(), self.table.split(','))

    self._primaryKeys = {}
    for table in tables:
      statement = "SELECT indkey FROM pg_index, pg_class c WHERE indrelid=c.oid AND indisprimary AND relname='%s' " \
                  % table
      cursor = self._dataConnection.cursor()
      cursor.execute(statement)
      indkeys = ""
      fields = []
      try:
        rsets = cursor.fetchmany()
      except self._DatabaseError, err:
        raise GDataObjects.ConnectionError(str(err))
      if rsets and len(rsets):
        for rs in (rsets):
          if rs:
            for f in (rs):
              if len(indkeys) == 0:
                indkeys = indkeys + f
              else:
                indkeys = indkeys + "," + f
      cursor.close()

      if len(indkeys) != 0:
        statement = "SELECT attname FROM pg_class c, pg_attribute WHERE c.oid = attrelid and attnum in (%s) \
                   AND relname='%s'" % (indkeys,table)
        cursor = self._dataConnection.cursor()
        cursor.execute(statement)
        try:
          rsets = cursor.fetchmany()
        except self._DatabaseError, err:
           raise GDataObjects.ConnectionError(str(err))
        if rsets and len(rsets):
          for rs in (rsets):
            if rs:
              for f in (rs):
                fields.append(f)
        cursor.close()
        self._primaryKeys[table] = fields

#
#  Extensions to Trigger Namespaces
#
class PGSQL_TriggerExtensions:

  def __init__(self, connection):
    self.__connection = connection

  # Return the current date, according to database
  def getTimeStamp(self):
    return self.__singleQuery("select current_timestamp")

  def getUnixTimeStamp(self):
    return self.__singleQuery("select extract(epoch from current_timestamp)")

  # Return a sequence number from sequence 'name'
  def getSequence(self, name):
    return self.__singleQuery("select nextval('%s')" % name)

  def atomicModify(self, func, field, dataObject, record, cond=None, autoCommit=1):
    if string.count(dataObject.table,',') == 0:
      table = dataObject.table
    elif field.count('.'):
      table = field.split('.')[0]
    else:
      raise
    
    where = []
    
    if dataObject._primaryKeys:
      for iterTable in dataObject._primaryKeys.keys():
        if iterTable != table: continue
        for iterField in dataObject._primaryKeys[iterTable]:
          auxWhere = "%s = '%s'"
          where.append(auxWhere % (iterField,record._initialData[table+"."+iterField]))
    else:
      for field in record._initialData.keys():
        a = len (field.split ('.'))
        if (a == 2):
          [auxTable, auxField]= field.split ('.')
          if string.count(dataObject.table,',') and (auxTable != table):
            continue
        elif (a == 1):
          auxField= field
        else:
          # raise
          pass

        if record._initialData[field] == None:
          where.append ("%s IS NULL" % auxField)
        else:
          where.append ("%s=%s" % (auxField, dataObject._toSqlString(record._initialData[field])))
    if cond:
       where.append(cond)
       
    cursor = self.__connection.cursor()

    getValueStmt = "SELECT %s FROM %s WHERE %s FOR UPDATE; " % \
                   (field, table, \
                    string.join(where,' AND ') ) 

    try:
      GDebug.printMesg(5, "_atomicModify: statement=%s" % getValueStmt)       
      cursor.execute(getValueStmt)
      rv = cursor.fetchone()
    except:
      print >> sys.stderr, "DBdriver.py", "error! %s" % sys.exc_info()[1]
      GDebug.printMesg(1,"**** Unable to execute extension query")
      GDebug.printMesg(1,"**** %s" % sys.exc_info()[1])
      cursor.close()
      raise GDataObjects.ConnectionError(str(err))

    try:
      lAux = rv[0] 
    except:
       try:
          if autoCommit:
             self.__connection.rollback()
          cursor.close()
       except:
          print >> sys.stderr, "DBdriver.py", "error! %s" % sys.exc_info()[1]
          GDebug.printMesg(1,"**** Unable to execute extension query")
          GDebug.printMesg(1,"**** %s" % sys.exc_info()[1])
          cursor.close()
          raise GDataObjects.ConnectionError(str(sys.exc_info()[1]))

    if field.count('.') > 0:
      auxField = field.split('.')[1]
    else:
      auxField = field

    if str(type(func)).count("'function'") > 0:
       value = func(rv[0])
    else:
       value = func

    updateStmt = "UPDATE %s SET %s WHERE %s" % \
                 (table, " %s = %s " % (auxField, str(value)), \
                  string.join(where,' AND ') ) 

    try:
      GDebug.printMesg(5, "_atomicModify: statement=%s" % updateStmt)
      cursor.execute(updateStmt)
      if autoCommit:
         self.__connection.commit()
      #cursor.close()
    except:
      print >> sys.stderr, "DBdriver.py", "error! %s" % sys.exc_info()[1]
      GDebug.printMesg(1,"**** Unable to execute extension query")
      GDebug.printMesg(1,"**** %s" % sys.exc_info()[1])
      cursor.close()
      raise GDataObjects.ConnectionError(str(sys.exc_info()[1]))

    return 1

  def atomicModifyForEntry(self, func, field, dataObject, record, cond=None,autoCommit=1):
    if string.count(dataObject.table,',') == 0:
      table = dataObject.table
    elif field.count('.'):
      table = field.split('.')[0]
    else:
      raise
    
    if field and field.count('.') > 0:
      field = field.split('.')[1]

    where = []
    
    if dataObject._primaryKeys:
      for iterTable in dataObject._primaryKeys.keys():
        if iterTable != table: continue
        for iterField in dataObject._primaryKeys[iterTable]:
          auxWhere = "%s = '%s'"
          where.append(auxWhere % (iterField,record._initialData[table+"."+iterField]))
    else:
      for iterField in record._initialData.keys():
        a = len (iterField.split ('.'))
        if (a == 2):
          [auxTable, auxField]= iterField.split ('.')
          if string.count(dataObject.table,',') and (auxTable != table):
            continue
        elif (a == 1):
          auxField= iterField
        else:
          # raise
          pass

        if record._initialData[iterField] == None:
          where.append ("%s IS NULL" % auxField)
        else:
          where.append ("%s=%s" % (auxField, dataObject._toSqlString(record._initialData[iterField])))
    if cond:
       where.append(cond)

    if record._haveHistory(table):
       where.append("_end_t IS NULL")
       
    cursor = self.__connection.cursor()
    
    getValueStmt = "SELECT %s FROM %s WHERE %s FOR UPDATE; " % \
                   (field, table, \
                    string.join(where,' AND ') ) 

    try:
      GDebug.printMesg(5, "_atomicModify: statement=%s" % getValueStmt)       
      cursor.execute(getValueStmt)
      rv = cursor.fetchone()
    except:
      print >> sys.stderr, "DBdriver.py", "error! %s" % sys.exc_info()[1]
      GDebug.printMesg(1,"**** Unable to execute extension query")
      GDebug.printMesg(1,"**** %s" % sys.exc_info()[1])
      cursor.close()
      raise GDataObjects.ConnectionError(str(sys.exc_info()[1]))

    try:
      lAux = rv[0] 
    except:
       try:
          if autoCommit:
             self.__connection.rollback()
          print >> sys.stderr, "DBdriver.py", "error! %s" % sys.exc_info()[1]
          GDebug.printMesg(1,"**** Unable to execute extension query")
          GDebug.printMesg(1,"**** %s" % sys.exc_info()[1])
          raise GDataObjects.ConnectionError("el registro a actaluzar no existe")
       except:
          print >> sys.stderr, "DBdriver.py", "error! %s" % sys.exc_info()[1]
          GDebug.printMesg(1,"**** Unable to execute extension query")
          GDebug.printMesg(1,"**** %s" % sys.exc_info()[1])
          raise GDataObjects.ConnectionError(str(sys.exc_info()[1]))

    if field.count('.') > 0:
      auxField = field.split('.')[1]
    else:
      auxField = field

    if str(type(func)).count("'function'") > 0:
       value = func(rv[0])
    else:
       value = func

    if record._haveHistory(table) and record.getPostType() == "" or record.getPostType() == "transactional":
       updateStmt = "UPDATE %s SET _end_t = '%s' WHERE %s" % \
                    (table, dataObject.triggerExtensions.getTimeStamp(), \
                     string.join(where,' AND ') )
       statementElement = record._buildInsertStatementWHist(table,field=field)
       statementElement[1].append(field)
       statementElement[2].append(str(value))

       insertStmt = "INSERT INTO %s (%s) VALUES (%s)" % \
                         (statementElement[0], string.join(statementElement[1],','), \
                          string.join(statementElement[2],',') ) 

    else:
       updateStmt = "UPDATE %s SET %s WHERE %s" % \
                    (table, " %s = %s " % (auxField, str(value)), \
                     string.join(where,' AND ') ) 
  

    try:
      GDebug.printMesg(5, "_atomicModifyForEntry: statement=%s" % updateStmt)
      cursor.execute(updateStmt)
      if record._haveHistory(table) and record.getPostType() == "" or record.getPostType() == "transactional":
         GDebug.printMesg(5, "_atomicModifyForEntry: statement=%s" % insertStmt)
         cursor.execute(insertStmt)
      #cursor.close()
    except:
      print >> sys.stderr, "DBdriver.py", "error! %s" % sys.exc_info()[1]
      GDebug.printMesg(1,"**** Unable to execute extension query")
      GDebug.printMesg(1,"**** %s" % sys.exc_info()[1])
      cursor.close()
      raise GDataObjects.ConnectionError(str(sys.exc_info()[1]))

    if autoCommit:
      self.__connection.commit()

    return 1

  # Run the SQL statement 'statement'
  def sql(self, statement):
    cursor = self.__connection.cursor()
    try:
      cursor.execute(statement)
      cursor.close()
    except:
      cursor.close()
      raise

  # Used internally
  def __singleQuery(self, statement):
    cursor = self.__connection.cursor()
    try:
      cursor.execute(statement)
      rv = cursor.fetchone()
      cursor.close()
    except:
      print >> sys.stderr, "DBdriver.py", "You've got your bar in my foo! And you've got your foo on my bar!  Two great reams that ream well together!"
      GDebug.printMesg(1,"**** Unable to execute extension query")
      GDebug.printMesg(1,"**** %s" % sys.exc_info()[1])
      cursor.close()
      return None

    try:
      return rv[0]
    except:
      return None



