#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000, 2001 Free Software Foundation
#
# FILE:
# db2/DBdriver.py
#
# DESCRIPTION:
# Driver to provide access to data stored in a DB2 database
#
# NOTES:
#   ftp://people.linuxkorea.co.kr/pub/DB2/
#
#   Supported attributes (via connections.conf or <database> tag)
#
#     service=    This is the DB2 DSN for your connection  (required)
#

import string
import sys

from string import lower
# from gnue.forms import GFOptions
from gnue.forms.GFError import DBError
from gnue.common import GDebug, GDataObjects
from gnue.common.dbdrivers._dbsig.DBdriver \
   import DBSIG_RecordSet, DBSIG_ResultSet, DBSIG_DataObject, \
          DBSIG_DataObject_SQL, DBSIG_DataObject_Object

try:
  import DB2
except ImportError, message:
  raise GConnections.AdapterNotInstalled, \
         _("Driver not installed: db2 for DB2 [%s]") % message


class DB2_RecordSet(DBSIG_RecordSet):
  def _postChanges(self):
    values = []
    fields = []
    fieldString = ""
    for field in (self._modifiedFields.keys()):
      fields.append(field + " = %s")
      values.append(self._fields[field])

    statement = 'UPDATE %s SET %s WHERE %s = %s'


class DB2_ResultSet(DBSIG_ResultSet):
  def __init__(self, dataObject, cursor=None, defaultValues={}, masterRecordSet=None):
    DBSIG_ResultSet.__init__(self, dataObject, \
            cursor, defaultValues, masterRecordSet)
    self._recordSetClass = DB2_RecordSet

class DB2_DataObject(DBSIG_DataObject):
  def __init__(self):
    DBSIG_DataObject.__init__(self)
    self._DatabaseError = "Database Error"

  def connect(self, connectData={}):
    GDebug.printMesg(1,"DB2 database driver initializing")
    try:
      self._dataConnection = DB2.connect(uid=connectData['_username'],
                   pwd=connectData['_password'],
                   dsn=connectData['service'])
      self.triggerExtensions = TriggerExtensions(self._dataConnection)
    except self._DatabaseError, value:
      raise GDataObjects.LoginError, value

  #
  # Schema (metadata) functions
  #

  # Return a list of the types of Schema objects this driver provides
  def getSchemaTypes(self):
    return [ ('table',    'Table',1),
             ('view',     'View', 1),
             ('alias',    'Alias',1),
             ('summary',  'Summary Table',1) ]


  # Return a list of Schema objects
  def getSchemaList(self, type=None):

    where_user = ""
    if type == None:
      where_type = ['A','S','T','V']
    else:
      where_type = [string.upper(type[0])]


    statement = \
      "select tabschema||'.'||tabname||'.'||type full_name, " + \
        "tabschema||'.'||tabname table_name, " + \
        "type table_type " + \
        "from syscat.tables where type in ('%s') and status = 'N' %s " \
              % (string.join(where_type,"','"), where_user) + \
          "order by tabname "

    GDebug.printMesg(5,statement)

    cursor = self._dataConnection.cursor()
    cursor.execute(statement)

    list = []
    for rs in cursor.fetchall():
      list.append(GDataObjects.Schema(attrs={'id':string.lower(rs[0]), 'name':rs[1],
                         'type':rs[2]},
                         getChildSchema=self.__getFieldSchema))

    cursor.close()
    return list


  # Find a schema object with specified name
  def getSchemaByName(self, name, type=None):

    where_user = ""
    parts = string.split(string.upper(name),'.')
    name = parts[-1]
    if len(parts) > 1:
      schema = " and tabschema='%s'" % parts[-2]
    else:
      schema = ""

    statement = \
      "select tabschema||'.'||tabname||'.'||type full_name, " + \
        "tabschema||'.'||tabname table_name, " + \
        "type table_type " + \
        "from syscat.tables where tabname == '%s' and status = 'N' %s " \
              % (name, schema) + \
          "order by tabname "

    GDebug.printMesg(5,statement)

    cursor = self._dataConnection.cursor()
    cursor.execute(statement)

    list = []
    for rs in cursor.fetchall():
      list.append(GDataObjects.Schema(attrs={'id':string.lower(rs[0]), 'name':rs[1],
                         'type':rs[2]},
                         getChildSchema=self.__getFieldSchema))

    cursor.close()

    try:
      return list[0]
    except:
      return None



  # Get fields for a table
  def __getFieldSchema(self, parent):

    # TODO: This does not support user-defined datatypes...
    # TODO: it will always report such as TEXT-like fields.

    schema, name, type = string.split(parent.id,'.')

    cursor = self._dataConnection.cursor()

    if type == 'a':
      statement = "select base_tabschema, base_tabname " + \
                  "from syscat.tables " + \
                  "where tabschema = '%s' and tabname='%s'" % (schema, name)

      GDebug.printMesg(5,statement)

      cursor.execute(statement)
      rs = cursor.fetchone()
      schema, name = rs

    statement = \
       "select tabschema||'.'||tabname||'.'||colname, " + \
       "colname, typename, nulls, length, scale, typeschema " + \
       "from syscat.columns" + \
       "where tabschema = '%s' and tabname = '%s' " % (schema, name) + \
       "order by colno"

    GDebug.printMesg(5,statement)

    cursor.execute(statement)

    list = []
    for rs in cursor.fetchall():

      attrs={'id': rs[0], 'name': rs[1],
             'type':'field', 'nativetype': rs[2],
             'required': rs[3] == 'Y'}

      if rs[2] in ('SMALLINT','INTEGER','INT','FLOAT','DOUBLE',
                   'DECIMAL','BIGINT','REAL','DEC','NUMERIC','NUMBER',
                   'NUM'):
        attrs['precision'] = rs[5]
        attrs['datatype'] = 'number'
      elif rs[2] in ('DATE','TIME','TIMESTAMP'):
        attrs['datatype'] = 'date'
      else:
        attrs['datatype'] = 'text'

      if rs[5] != 0:
        attrs['length'] = rs[4]

      list.append(GDataObjects.Schema(attrs=attrs))

    cursor.close()
    return tuple(list)


class DB2_DataObject_Object(DB2_DataObject, \
      DBSIG_DataObject_Object):

  def __init__(self):
    DB2_DataObject.__init__(self)

  def _buildQuery(self, conditions={}, qTable = None): 
    if not qTable:
       qTable = self.table
    return DBSIG_DataObject_Object._buildQuery(self, conditions, qTable)


class DB2_DataObject_SQL(DB2_DataObject, \
      DBSIG_DataObject_SQL):
  def __init__(self):
    # Call DBSIG init first because DB2_DataObject needs to overwrite
    # some of its values
    DBSIG_DataObject_SQL.__init__(self)
    DB2_DataObject.__init__(self)

  def _buildQuery(self, conditions={}):
    return DBSIG_DataObject_SQL._buildQuery(self, conditions)



supportedDataObjects = {
  'object': DB2_DataObject_Object,
  'sql':    DB2_DataObject_SQL
}

