#
# Copyright 2000-2002 Free Software Foundation
#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# FILE:
# odbc-db2/DBdriver.py
#
# DESCRIPTION:
# Driver to provide access to data via the public domain win32all ODBC Driver
# using DB2 specific catalog data
#
# NOTES:
# Only works under Win32... requires the win32all extensions.
# (http://aspn.activestate.com/ASPN/Downloads/ActivePython/Extensions/Win32all)
#
#   Supported attributes (via connections.conf or <database> tag)
#
#     service=   This is the ODBC DSN= string to use.
#
#


import sys, string, types
from gnue.common import GDebug
from gnue.common import GDataObjects
from gnue.common.dbdrivers.odbc.DBdriver import ODBC_DataObject_Object

class ODBCDB2_DataObject_Object(ODBC_DataObject_Object):


  #
  # Schema (metadata) functions
  #

  # Return a list of the types of Schema objects this driver provides
  def getSchemaTypes(self):
    return [ ('table',    'Table',1),
             ('view',     'View', 1),
             ('alias',    'Alias',1) ]


  # Return a list of Schema objects
  def getSchemaList(self, type=None):

    where_user = ""
    if type == None:
      where_type = ['A','T','V']
    else:
      where_type = [string.upper(type[0])]


    statement = \
      "select creator||'.'||name||'.'||type, " + \
        "creator||'.'||name, " + \
        "type, name " + \
        "from sysibm.systables where type in ('%s') %s " \
              % (string.join(where_type,"','"), where_user) + \
          "order by name "

    GDebug.printMesg(5,statement)

    cursor = self._dataConnection.cursor()
    cursor.execute(statement)

    list = []
    for rs in cursor.fetchall():
      list.append(GDataObjects.Schema(attrs={'id':rs[0],
                           'name':rs[1],
                           'type':{'A':'alias','T':'table','V':'view'}[rs[2]]},
                            getChildSchema=self.__getFieldSchema))

    cursor.close()
    return list


  # Find a schema object with specified name
  def getSchemaByName(self, name, type=None):

    parts = string.split(string.upper(name),'.')
    name = parts[-1]
    if len(parts) > 1:
      creator = " and creator='%s'" % parts[-2]
    else:
      creator = ""

    statement = \
      "select creator||'.'||name||'.'||type, " + \
        "creator||'.'||name, " + \
        "type, name " + \
        "from sysibm.systables where name = '%s' %s " \
              % (name, creator) + \
          "order by name "

    GDebug.printMesg(5,statement)

    cursor = self._dataConnection.cursor()
    cursor.execute(statement)

    list = []
    for rs in cursor.fetchall():
      list.append(GDataObjects.Schema(attrs={'id':rs[0], 'name':rs[1],
                         'type':{'A':'alias','T':'table','V':'view'}[rs[2]]},
                         getChildSchema=self.__getFieldSchema))

    cursor.close()

    try:
      return list[0]
    except:
      return None



  # Get fields for a table
  def __getFieldSchema(self, parent):

    # TODO: This does not support aliases that point to remote databases
    # TODO: I have no clue how to attach to a remote database to grab
    # TODO: the "real" table's column information :(

    creator, name, type = string.split(parent.id,'.')

    cursor = self._dataConnection.cursor()

    if type == 'A':
      statement = "select tbcreator, tbname " + \
                  "from sysibm.systables " + \
                  "where creator = '%s' and name='%s'" % (creator, name)

      GDebug.printMesg(5,statement)

      cursor.execute(statement)
      rs = cursor.fetchone()
      creator, name = rs

    statement = \
       "select tbcreator||'.'||tbname||'.'||name, " + \
       "name, coltype, nulls, length, scale, colno " + \
       "from sysibm.syscolumns " + \
       "where tbcreator = '%s' and tbname = '%s' " % (creator, name) + \
       "order by colno"

    GDebug.printMesg(5,statement)

    cursor.execute(statement)

    list = []
    for rs in cursor.fetchall():

      attrs={'id': rs[0], 'name': rs[1],
             'type':'field', 'nativetype': rs[2],
             'required': rs[3] == 'N'}

      if rs[2] in ('SMALLINT','INTEGER','INT','FLOAT','DOUBLE',
                   'DECIMAL','BIGINT','REAL','DEC','NUMERIC','NUMBER',
                   'NUM'):
        attrs['precision'] = rs[5]
        attrs['datatype'] = 'number'
      elif rs[2] in ('DATE','TIME','TIMESTAMP'):
        attrs['datatype'] = 'date'
      else:
        attrs['datatype'] = 'text'

      if rs[5] != 0:
        attrs['length'] = rs[4]

      list.append(GDataObjects.Schema(attrs=attrs))

    cursor.close()
    return tuple(list)



supportedDataObjects = {
  'object': ODBCDB2_DataObject_Object,
}



