#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2002 Free Software Foundation
#
# FILE:
# GFDisplayHandler.py
#
# DESCRIPTION:
#
# NOTES:
#

import sys, string, types

from gnue.common.FormatMasks import DateMask, TextMask, NumberMask
from gnue.common import GDebug
from GFEvent import *

try:
  from mx.DateTime import DateTime
except ImportError:
  print """
   This GNUe tool requires mxDateTime to be installed

   You can find the mxDateTime at 

   http://www.lemburg.com/files/python/mxDateTime.html

"""
  import sys
  sys.exit()

#############################################################################
#
# Base handler
#
# Do not use this handler directly.
# You should be using one of the
# other types based on this one.
#
class FieldDisplayHandler(GFEventAware):


  #####################
  #
  # General methods
  #

  def setValue(self, value):
    self.modified = 0
    self.value = value
    self._buildDisplay()


  def getValue(self):
    return self.value


  def getDisplay(self):
    return self.display


  def getDisplayFiller(self, value):
    return self._buildDisplayHelper(value, 0)


  def generateRefreshEvent(self):
    # TODO: this should probably actually happen in UIwxpython!
    if self.handleCR:
      self.dispatchEvent(GFEvent('updateEntryEditor',
           object = self.entry,
           display=self.display, # string.replace(self.display,'\n','\r\n'),
           cursor=self.cursor + len(string.split(self.display[:self.cursor+1],'\n'))-1,
           selection=self.getSelectionArea(),
         ))
    else:
      self.dispatchEvent(GFEvent('updateEntryEditor',
           object = self.entry,
           display=self.display,
           cursor=self.cursor,
           selection=self.getSelectionArea(),
         ))


  #####################
  #
  # Editing methods
  #
  
  #
  # beginEdit
  # called when a widget first gets focus
  #
  def beginEdit(self, event):
    self.editing = 1
    self.modified = 0

    # TODO: Replace with formatter
    self.setValue(self.entry.getValue())
    self.work = self._buildDisplayHelper(self.value, 1)
    self._buildDisplay()
    if self.entry.readonly:
      self.cursor = 0
    else:
      self.cursor = len(self.display)

  #
  # endEdit
  # called when a widget loses focus or
  # when enter is hit
  #
  def endEdit(self, event):
    if not self.editing:
      return

    if not self._loadedAllowedValues:
      self.entry.allowedValues()
      self._loadedAllowedValues = 1


    # If this event returns __error__, then
    # the input was invalid and the entry is
    # still in editing mode.  __errortext__
    # will contain an appropriate error message.
    #
    # If no error, then event.modified is true
    # if the field was modified and the event's
    # result is the final entry value.

    self.selection1 = None

    if self.modified:
      if self._buildValue():
        if self.entry._allowedValues and not self.entry._allowedValues.has_key(str(self.value)):
          self.work = ""
          event.__error__ = 1
          event.__errortext__ = _("Invalid value '%s' for field") % self.value
          return

        self.editing = 0
        event.__results__ = self.value
        event.modified = self.modified
        self.entry.setValue(self.value)

        self._buildDisplay()

      else:
        event.__error__ = 1
        event.__errortext__ = _("The entered value\n ") \
                            + " '%s'\n" % self.work \
                            + _("is not a valid entry.")


  def addText(self, event):

    value = event.text

    if (self.entry.readonly and not self.entry._block.mode=='query'):
      event.__error__ = 1
      event.__errortext__ = _("Cannot modify field.  Form is read only")
      return
    elif self.entry.editOnNull and \
         self.value != None and \
         self.value != "" and \
         not self.entry._block._resultSet.current.isPending():
      event.__error__ = 1
      event.__errortext__ = _("Cannot modify field after initial creation.")
      return

    # Do some appropriate checks on the text
    # TODO: This will be moved to the mask system!!

    if self.entry._lowercase:
      value = string.lower(event.text)
    if self.entry._uppercase:
      value = string.upper(event.text)

    if hasattr(self.entry,'max_length') and \
       len(self.work)  + len(value) > self.entry.max_length:
      return

    if ( self.entry._numeric and \
         self.entry._block.mode == 'normal' ):
      for char in value:
        if not (char in string.digits or char in '.'):
          return

    # To do overstrike, we'll fudge by first "highlighting"
    # the character to replace, then use the selection logic.

    if  hasattr(event,'overstrike') and event.overstrike and \
        self.selection1 == None:
      self.selection1 = self.cursor
      self.selection2 = self.selection1 + 1

    if self.selection1 != None:
      # If text is selected, then we will replace

      s1 = min(self.selection1, self.selection2)
      s2 = max(self.selection1, self.selection2)

      self.work = self.work[:s1]  \
                   + value        \
                   + self.work[s2:]

      self.selection1 = None
      self.cursor = s1 + len(value)

    else:
      # Otherwise just graft the new text in place
      
      self.work = self.work[:self.cursor] \
                   + value                \
                   + self.work[self.cursor:]

      self.cursor += len(event.text)

    
    event.__dropped__ = 1
    self.modified = 1
    self._buildDisplay()


  # Delete backwards one character
  def backspace(self, event):

    # If we have a "selection",
    # then act like a "delete"
    if self.selection1 != None:
      self.delete(event)
      return

    precurs = self.cursor
    self.moveCursorLeft(event)

    if self.cursor != precurs:
      event.overstrike = 1
      event.text = ""
      
      self.addText(event)

  # Delete forward one character
  def delete(self, event):
    event.overstrike = 1
    event.text = ""
    
    self.addText(event)


  #####################
  #
  # Cursor movement
  #

  def moveCursorRelative(self, relative):
    pass


  def moveCursor(self, event, selecting=0):
    if not selecting:
      self.selection1 = None

    self.cursor = min(event.position, len(self.display))


  def moveCursorLeft(self, event, selecting=0):
    if not selecting:
      self.selection1 = None

    if self.cursor > 0:
      self.cursor -= 1

  def moveCursorRight(self, event, selecting=0):
    if not selecting:
      self.selection1 = None

    if self.cursor < len(self.display):
      self.cursor += 1

  def moveCursorToEnd(self, event, selecting=0):
    if not selecting:
      self.selection1 = None

    self.cursor = len(self.display)


  def moveCursorToBegin(self, event, selecting=0):
    if not selecting:
      self.selection1 = None

    self.cursor = 0


  #####################
  #
  # Selection stuff
  #

  # Set the selection area
  def setSelectionArea(self, cursor1, cursor2):
    self.selection1 = min(cursor1, cursor2)
    self.selection2 = max(cursor1, cursor2)


  # Return the selected area as a tuple (or
  # None if no selection)
  def getSelectionArea(self):
    if self.selection1 == None:
      return None
    else:
      return ( min(self.selection1, self.selection2),
               max(self.selection1, self.selection2) )


  # Select the entire text of the entry and move
  # the cursor to the end
  def selectAll(self, event):

    self.selection1 = 0
    self.moveCursorToEnd(event,1)
    self.selection2 = self.cursor


  # Move the selection cursor to the left one unit
  def selectLeft(self, event):

    if self.selection1 == None:
      self.selection1 = self.cursor

    self.moveCursorLeft(event, 1)
    self.selection2 = self.cursor


  # Move the selection cursor to the right one unit
  def selectRight(self, event):

    if self.selection1 == None:
      self.selection1 = self.cursor

    self.moveCursorRight(event,1)
    self.selection2 = self.cursor


  # Select to the beginning of the entry
  def selectToBegin(self, event):

    if self.selection1 == None:
      self.selection1 = self.cursor

    self.moveCursorToBegin(event,1)
    self.selection2 = self.cursor


  # Select to the beginning of the entry
  def selectToEnd(self, event):

    if self.selection1 == None:
      self.selection1 = self.cursor

    self.moveCursorToEnd(event,1)
    self.selection2 = self.cursor


  # Copy to the clipboard
  def clipboardCopy(self, event):
    if self.selection1 != None:
      sel1, sel2 = self.getSelectionArea()
      self.dispatchEvent(GFEvent('setClipboard',
               text=self.display[sel1:sel2]))

    event.refreshDisplay = 0


  def clipboardCut(self, event):
    self.clipboardCopy(event)
    edevent = GFEvent("requestKEYPRESS", text="")
    self.dispatchEvent(edevent)


  def clipboardPaste(self, event):
    event.text = self.dispatchEvent(GFEvent('getClipboard'))
    if event.text != None:
      self.addText(event)


  #####################
  #
  # Internal methods
  #
  def __init__(self, entry, eventHandler, formatter=None):
    GFEventAware.__init__(self, eventHandler)

    self.entry = entry            # Our entry
    self.editing = 0              # In editing mode??
    self.modified = 0             # Have we been modified??
    self.value = None             # The latest db-compat value
    self.work = ""                # Our working value
    self.display = ""             # The latest display-formatted value
    self.selection1 = None        # Start of highlight
    self.selection2 = None        # End of highlight
    self.cursor = 0               # Cursor position
    self._loadedAllowedValues = 0 # the allowed values already been loaded

    # TODO: replace w/an event that asks the
    # TODO: UIdriver if this should happen!
    self.handleCR = sys.platform == "win32"

    # Needs to be set by the child classes
    self.formatter = formatter

    self.subevents = GFEventController()
    self.subevents.registerEventListeners( {

                 # "Entry" events
                 'requestKEYPRESS'     : self.addText,
                 'requestCURSORLEFT'   : self.moveCursorLeft,
                 'requestCURSORRIGHT'  : self.moveCursorRight,
                 'requestCURSOREND'    : self.moveCursorToEnd,
                 'requestCURSORHOME'   : self.moveCursorToBegin,
                 'requestCURSORMOVE'   : self.moveCursor,
                 'requestBACKSPACE'    : self.backspace,
                 'requestDELETE'       : self.delete,
                 'beginEDITMODE'       : self.beginEdit,
                 'endEDITMODE'         : self.endEdit,

                 # Selection/clipboard events
                 'requestSELECTALL'    : self.selectAll,
                 'requestSELECTTOHOME' : self.selectToBegin,
                 'requestSELECTTOEND'  : self.selectToEnd,
                 'requestSELECTLEFT'   : self.selectLeft,
                 'requestSELECTRIGHT'  : self.selectRight,
                 'requestCOPY'         : self.clipboardCopy,
                 'requestCUT'          : self.clipboardCut,
                 'requestPASTE'        : self.clipboardPaste,
        })


  def _buildValue(self):
    self.value = self.work
    return 1

  def _buildDisplayHelper(self, value, editing):
    if value == None:
      return ""
    else:
      return "%s" % value

  def _buildDisplay(self):
    if self.editing:
      self.display = self._buildDisplayHelper(self.work, 1)
    else:
      self.display = self._buildDisplayHelper(self.value, 0)








#############################################################################
#
#
#
class TextDisplayHandler(FieldDisplayHandler):
  def __init__(self, entry, eventHandler, displayMask, inputMask, formatMask):
##    FieldDisplayHandler.__init__(self, entry, eventHandler,
##            TextMask.TextMask(displayMask, inputMask, formatMask))
    FieldDisplayHandler.__init__(self, entry, eventHandler, None)



#############################################################################
#
# Handler for Numeric types
#
class NumberDisplayHandler(FieldDisplayHandler):
  def __init__(self, entry, eventHandler, displayMask, inputMask):
##    FieldDisplayHandler.__init__(self, entry, eventHandler,
##            DateMask.DateMask(displayMask, inputMask, displayMask))
    FieldDisplayHandler.__init__(self, entry, eventHandler, None)
    self.displayMask= displayMask

  def setValue(self, value):
    if value in (None,""):
      value = None
    else:
      value = float(value)

    return FieldDisplayHandler.setValue(self, value)


  # TODO: Replace with format mask
  def _buildDisplayHelper(self, val, editing):
    if val in (None,""):
      return ""

    if editing and type(val) == types.StringType:
      return val

    try:
      value = float(val)
    except ValueError:
      return str(val)

    # hack
    if self.displayMask:
      try:
        return self.displayMask % value
      except:
        pass

    try:
      if int(value) == value:
        value = "%d" % value
      else:
        value = "%s" % value
    except OverflowError:
      value = "%s" % value

    if not editing:
      # Comma-fy (Temporary!)
      if value[:1] == '-':
        minus = "-"
        value = value[1:]
      else:
        minus = ""

      try:
        whole, decimal = string.split(str("%.2f" % eval(value)),'.')
        value = ".%s" % decimal
      except ValueError:
        whole = value
        value = ""

      while len(whole) > 3:
        value = ",%s%s" % (whole[-3:], value)
        whole = whole[:-3]

      value = minus + whole + value

    return value


  def _buildValue(self):
    if not len(self.work):
      self.value = None
      return 1

    try:
      self.value = float(self.work)
      return 1
    except ValueError:
      return 0




#############################################################################
#
# Handler for Date types
#
class DateDisplayHandler(FieldDisplayHandler):
  def __init__(self, entry, eventHandler, displayMask, inputMask):
##    FieldDisplayHandler.__init__(self, entry, eventHandler,
##            DateMask.DateMask(displayMask, inputMask, displayMask))
    FieldDisplayHandler.__init__(self, entry, eventHandler, None)


  def setValue(self, value):
    return FieldDisplayHandler.setValue(self, value)


  # TODO: Replace with format mask
  def _buildDisplayHelper(self, value, editing):
    format = "%d/%m/%Y"
    
    if value in (None,""):
      return ""
    try:
      return value.strftime(format)
    except AttributeError:
      return str(value)


  def _buildValue(self):

    if not len(self.work):
      self.value = None
      return 1

    # TODO: Replace with format mask
    try:
      value = string.replace(string.replace(self.work,'.','/'),'-','/')

      # Support for quick entry like '123102' for '12/31/02'
      if len(value) in (6,8) and string.find(value,'/') == -1:
        day = value[:2]
        month = value[2:4]
        year = value[4:]
      else:
        day, month, year = string.split(value,'/')

      year = int(year)
      if year < 100:
        if year > 50:
          year = year + 1900
        else:
          year = year + 2000

      self.value = DateTime(year, int(month), int(day))

      return 1
    except:
      return 0



#############################################################################
#
# Handler for dropdowns
#
class DropdownDisplayHandler(FieldDisplayHandler):

  def __init__(self, *args, **params):

    FieldDisplayHandler.__init__(self, *args, **params)

    # My events...
    self.subevents.registerEventListeners( {
           'requestREPLACEVALUE' : self.replaceText  } )


  def beginEdit(self, event):

    if not self._loadedAllowedValues and not hasattr(self.entry,'_allowedValues'):
      self.entry.allowedValues()
      self._loadedAllowedValues = 1

    self.editing = 1
    self.modified = 0

    # TODO: Replace with formatter
    self.setValue(self.entry.getValue())

    if self.value == None:
      self.work = ""
    else:
      try:
        self.work = self.entry._allowedValues[str(self.value)]
      except KeyError:
        self.work = ""
        event.__error__ = 1
        event.__errortext__ = _("Invalid value '%s' for keyed pull-down field") % self.value

    self._buildDisplay()

    self.cursor = len(self.display)


  # TODO: Replace with format mask
  def _buildDisplayHelper(self, value, editing):
    self._dropdownIndex = 0
    if value in (None,""):
      return ""

    if editing:
      i = 0
      val = string.lower(value)
      for disp in self.entry._allowedValuesDescr:
        if disp[:len(val)] == val:
          display = self.entry._allowedValues[self.entry._allowedValuesReverse[disp]]
          self._dropdownIndex = i
          return str(display)

      return str(value)
    else:
      return self.entry._allowedValues[str(value)]

  def _buildDisplay(self):
    if self.editing:
      self.display = self._buildDisplayHelper(self.work, 1)
      if self.cursor > len(self.work):
        self.work = self.display
    else:
      self.display = self._buildDisplayHelper(self.value, 0)


  def replaceText(self, event):
    self.selection1 = 0
    self.selection2 = len(self.display)
    self.addText(event)


  def _buildValue(self):
    if self.work == "":
      self.value = '' # None
    else:
      try:
        self.value = self.entry._allowedValuesReverse[self.display]
      except KeyError:
        return 0
    return 1

#############################################################################
#
# Handler for textLookups
#
class TextLookupDisplayHandler(FieldDisplayHandler):

  def __init__(self, *args, **params):
    FieldDisplayHandler.__init__(self, *args, **params)

  def beginEdit(self, event):
    pass

  # TODO: Replace with format mask
  def _buildDisplayHelper(self, value, editing):
    if value in (None,""):
      return ""
    descr = ""
    if hasattr(self.entry,'datasource'):
      if hasattr(self.entry,'foreign_key'):
        foreignKeyContent = string.split(self.entry.foreign_key,'.')
        if len(foreignKeyContent) > 2:
          tmpFieldName = foreignKeyContent[1]+'.'+foreignKeyContent[2]
        else:
          tmpFieldName = foreignKeyContent[1]
      
      lResultSet = self.entry.datasource.createResultSet({tmpFieldName:value})
    if (lResultSet.firstRecord()):
      if hasattr(self.entry,'foreign_key_description'):
        descr = "%s" % lResultSet.current.getField(self.entry.foreign_key_description)

    return descr

  def _buildDisplay(self):
    self.display = self._buildDisplayHelper(self.value, 0)

  def endEdit(self,event):
    pass

#############################################################################
#
# Handler for Checkbox types
#
class CheckboxDisplayHandler(FieldDisplayHandler):

  def __init__(self, *args, **params):

    self.trueValue = gConfig("checkboxTrue")
    self.falseValue = gConfig("checkboxFalse")

    self.trueValues =  ('Y','y','T','t','1','x','X', self.trueValue)
    self.falseValues = ('N','n','F','f','0','', ' ', self.falseValue)

    FieldDisplayHandler.__init__(self, *args, **params)

    # My events...
    self.subevents.registerEventListeners( {
           'requestTOGGLECHKBOX' : self.toggle  } )


  def setValue(self, value):
    # Force to 0 or 1
#     print "setValue 1: %s" % value
    if ("%s" % value)[:1] in self.trueValues:
      self.value = 1
    elif ("%s" % value)[:1] in self.falseValues:
      self.value = 0
    else:
      self.value = value and 1 or 0
#     print "setValue 2: %s" % self.value

    self.modified = 0
    self._buildDisplay()


  def _buildValue(self):
    if int(self.work):
      return self.trueValue
    else:
      return self.falseValue


  # TODO: Replace with format mask
  def _buildDisplayHelper(self, value, editing):
#     print "bDH: v: %s, e: %s" % (value, editing)
#     from traceback import print_stack
#     import sys
#     print_stack (file=sys.stdout)
#     print '-------------'

    if ("%s" % value)[:1] in self.trueValues:
      return 1
    elif ("%s" % value)[:1] in self.falseValues:
      return 0
    else:
      return value and 1 or 0


  def getDisplay(self):
    return self.display


  def toggle(self, event):
#     from traceback import print_stack
#     import sys
#     print_stack (file=sys.stdout)
    if not self.editing:
      self.beginEdit (None)
#     print "toggle 1: %s" % self.work
    self.work= not self.work
    # self.setValue()
#     print "toggle 2: %s" % self.work
#     print '-------------'
    # print "toggle: %s" % event.data[1]
    self.modified= 1
    self._buildDisplay ()


  def beginEdit(self, event):
#     print "bE"

    self.editing = 1
    self.modified = 0

    self.setValue(self.entry.getValue())

    self.work = self.value

    self._buildDisplay()


  def endEdit(self, event):
#     print "eE"

    self.modified = 0

    self.value= self.work
    self.entry.setValue(self.value)

    self._buildDisplay()
    # self.setValue(self.work)
    self.editing = 0



  # Delete backwards one character
  def backspace(self, event):
    return

  # Delete forward one character
  def delete(self, event):
    return

  def moveCursorRelative(self, relative):
    return


  def moveCursor(self, event, selecting=0):
    return


  def moveCursorLeft(self, event, selecting=0):
    return


  def moveCursorRight(self, event, selecting=0):
    return


  def moveCursorToEnd(self, event, selecting=0):
    return


  def moveCursorToBegin(self, event, selecting=0):
    return


  #####################
  #
  # Selection stuff
  #

  # Set the selection area
  def setSelectionArea(self, cursor1, cursor2):
    return


  # Return the selected area as a tuple (or
  # None if no selection)
  def getSelectionArea(self):
    return


  # Select the entire text of the entry and move
  # the cursor to the end
  def selectAll(self, event):
    return


  # Move the selection cursor to the left one unit
  def selectLeft(self, event):
    return


  # Move the selection cursor to the right one unit
  def selectRight(self, event):
    return


  # Select to the beginning of the entry
  def selectToBegin(self, event):
    return


  # Select to the beginning of the entry
  def selectToEnd(self, event):
    return


  # Copy to the clipboard
  def clipboardCopy(self, event):
    return self.work


  def clipboardCut(self, event):
    return self.work


  def clipboardPaste(self, event):
    event.text = self.dispatchEvent(GFEvent('getClipboard'))
    if event.text != None:
      self.addText(event)



