#
# Copyright 2000-2002 Free Software Foundation
#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# FILE:
# GFEvent.py
#
# DESCRIPTION:
# Basic event passed between the GFForm, GFView, and GFController objects
#
# NOTES:
#

from gnue.common import GDebug
import string

#
# Basic event passed between items
#
# If event = GFEvent('MYEVENT',x=1, y=2, object=MyObject), then
#   event.event == 'MYEVENT', event.x = 1,
#   event.y == 2, and event.object == MyObject
#
class GFEvent:
  def __init__(self, event, data=None, **parms):

    self.__dict__.update(parms)

    # TODO: Get rid of data=
    self.data  = data

    self.__event__ = event
    self.__result__ = None
    self.__dropped__ = 0
    self.__error__ = 0
    self.__errortext__ = ""


  def getResult(self):
    return self.__result__

  def setResult(self, value):
    self.__result__ = value

  def getEvent(self):
    return self.__event__


#
# Base for and object that sends and receives events
#
class GFEventAware:

  def __init__(self, controller):
    self.__controller = controller
    self.dispatchEvent = controller.dispatchEvent
    self.registerEventListeners = controller.registerEventListeners

  def __del__(self):
    self.__controller = None
    self.dispatchEvent = None
    self.registerEventListeners = None


class GFEventController:

  def __init__(self):
    self.__incomingEvents = {}


  def registerEventListeners(self, events):
    for event in events.keys():
      try:
        self.__incomingEvents[event].append(events[event])
      except KeyError:
        self.__incomingEvents[event] = [events[event]]
 

  def dispatchEvent(self, event):
    #GDebug.printMesg(0, "Event: %s" % (event.__event__))
    try:
      handlers = self.__incomingEvents[event.__event__]
    except KeyError:
      GDebug.printMesg(3, "Unknown Event: %s " % (event.getEvent()))
      return

    for handler in handlers:
      ## GDebug.printMesg(10, "Event: %s in %s" % (event.__event__, handler))
      handler(event)
      if event.__error__ or event.__dropped__:
        break

    return event.__result__








