#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
# FILE:
# GFBlock.py
#
# DESCRIPTION:
#
# NOTES:
#
# HISTORY:
#

from gnue.forms.GFEvent import *
from gnue.forms.GFObjects.GFDataSource import GFDataSource

from gnue.common import GDebug
from gnue.common import GConditions
from GFObj import GFObj

import string

# These should really go somewhere else
TRUE = 1
FALSE = 0


############################################################
# GFBlock
#
#
class GFBlock(GFObj, GFEventAware):
  def __init__(self, parent=None):
    GFObj.__init__(self, parent, 'GFBlock')

    self.mode = 'normal'

    self._inits = [self.initialize]

    self._datasource = None
    self._resultSet = None
    self._dataSourceLink = None
    self._entryList = []

    self._currentRecord = 0
    self._recordCount = 0
    self._queryDefaults = {}
    self._queryValues = {}
    self._lastQueryValues = {}
    self._autocreate =  int(gConfig('autocreate'))
    self._gap = 0
    self._rows = 1
    
    self._scrollBars= []

    #
    # Trigger exposure
    #
    self._triggerGlobal = 1
    #self._triggerFunctions={'clear':{'function':self.processClear},
    self._triggerFunctions={'jumpRecord':{'function':self.jumpRecord,
                                          'global': 1,                                          
                                          },
                            'getRecordCount':{'function':self.getRecordCount},
                            'getCurrentRecord':{'function':self.getCurrentRecord},
                            'getCurrentRecordNumber':{'function':self.getCurrentRecordNumber},                            
                            'isSaved':{'function':self.isSaved},
                            'deleteRecord':{'function':self.deleteRecord},
                            'isEmpty':{'function':self.isEmpty},
                            'nextRecord':{'function':self.nextRecord},
                            'lastRecord':{'function':self.lastRecord},
                            'firstRecord':{'function':self.firstRecord},
                            'prevRecord':{'function':self.prevRecord},                            
                            'newRecord':{'function':self.newRecord},
                            'initQuery':{'function':self.initQuery},
                            'processQuery':{'function':self.processQuery},                            
                            'commit':{'function':self.commit},
                            'reset':{'function':self.reset},
                            'clear':{'function':self.clear},
                            'postChanges':{'function':self.postChanges},
                            }

  def __getitem__(self, index):

    cnt = self._resultSet.getRecordCount() - 1
    if index > cnt:
      raise IndexError

    #print "GFBlock", index
    self.jumpRecord(index)

    if cnt == index and self.isEmpty():
      #print "Bad, bad, bad"
      raise IndexError

    return self

  def _buildObject(self):

    if hasattr(self, 'rows'):
      self._rows = self.rows

    if hasattr(self, 'rowSpacer'):
      self._gap = self.rowSpacer

    if hasattr(self,'datasource'):
      self.datasource = string.lower(self.datasource)

    self.walk(self.__setChildRowSettings)
    return GFObj._buildObject(self)



  def initialize(self):
    self._form = self.findParentOfType('GFForm')
    self._page = self.findParentOfType('GFPage')

    # Initialize our events system
    GFEventAware.__init__(self, self._form._app.eventController)


    self.walk(self.buildEntryList)

    if not hasattr(self,'datasource') or not self.datasource:
      ds = GFDataSource(self)
      self.datasource = ds.name = "dts_%s" % self
      self._form._datasourceDictionary[ds.name] = ds
      ds._buildObject()
      ds.phaseInit()

    self._dataSourceLink = self._form._datasourceDictionary[self.datasource]
    self._dataSourceLink.registerResultSetListener(self._loadResultSet)


  def __setChildRowSettings(self, object):
    if hasattr(object,'rows'):
      object._rows = object.rows
    else:
      object._rows = self._rows

    if hasattr(object,'rowSpacer'):
      object._gap = object.rowSpacer
    else:
      object._gap = self._gap


  def buildEntryList(self, object):
    if object._type == 'GFEntry':
      self._entryList.append(object)

  def getResultSet(self):
    return self._resultSet

  def getCurrentRecord(self):
    return self.getResultSet().getCurrentRecord()

  def getCurrentRecordNumber(self):
    return self.getResultSet().getRecordNumber()

  def getRecordCount(self):
    return self._recordCount
    
  #
  # isSaved
  #
  # returns a positive value if the datasource the block is associated
  # with is saved
  #
  def isSaved(self):
    return self._resultSet == None or not self._resultSet.isPending()

  #
  # deleteRecord
  #
  # Doesn't really delete the record but marks it for
  # deletion during next commit
  #
  def deleteRecord(self):
    self._resultSet.current.delete()

  #
  # Load and process a resultset
  #
  def _loadResultSet(self, resultSet):
    self._resultSet = resultSet
    self._currentRecord = -1
    self._recordCount = 0
    resultSet._block = self

    if self._resultSet.firstRecord():
      self.switchRecord(0)
      self._recordCount = self._resultSet.getRecordCount()
    else:
      # TODO: This means no results were returned from a query.
      # TODO: It probably shouldn't create a new record :)  Display a message?
      self.newRecord()

  #
  # isEmpty()
  #
  def isEmpty(self):
    return self._resultSet.current.isEmpty()

  #
  #
  # Moves the proper record into editing position
  #
  def switchRecord(self, adjustment):
    if adjustment <> 0:
      self.processTrigger('ON-SWITCH')
    self._currentRecord = self._resultSet.getRecordNumber()
    for entry in self._entryList:
      #print entry.name, entry._block.name, entry.hidden
      entry.recalculateVisible( adjustment, self._currentRecord, self._recordCount)
      self._form.updateUIEntry(entry)

    # update the scrollbars...
    entryFounded = None
    for sb in self._scrollBars:
      for entry in self._entryList:
        try:
          if not entry.hidden and entry.rows > 1:
            entryFounded = entry
            break
        except:
          pass
      sb.update (entry.lowestVisible, self._recordCount)

  #
  # newRecord
  #
  # Adds a record to the current records in memory
  #
  def newRecord(self):
    if self._resultSet.insertRecord():
      self._recordCount = self._resultSet.getRecordCount()
      self.switchRecord(1)
      self.processTrigger('ON-NEWRECORD')

  def nextRecord(self):
    if self._resultSet.nextRecord():
      self._recordCount = self._resultSet.getRecordCount()
      self.switchRecord(1)
    elif self._autocreate and not self.isEmpty() and not self.restrictInsert:
      self.newRecord()

  def lastRecord(self):
    self._resultSet.lastRecord()
    self.switchRecord(0)

  def firstRecord(self):
    self._resultSet.firstRecord()
    self.switchRecord(0)

  def prevRecord(self):
    if self._resultSet.prevRecord():
      self.switchRecord(-1)

  def jumpRecord(self, recordNumber):
    if not self._resultSet.setRecord(recordNumber):
      self._resultSet.lastRecord()
    jump = self._resultSet.getRecordNumber() - self._currentRecord
    self._currentRecord = self._resultSet.getRecordNumber()
    self.switchRecord(jump)

  def postChanges(self):
    self.getForm().changeFocus(self)    
    self.getForm().postChanges(self)

  def commit(self):
    self.getForm().changeFocus(self)    
    self.getForm().commit(self)

  def getForm(self):
    return self._form

  #
  # processCommit
  #
  def processCommit(self):
    GDebug.printMesg(1, "processing commit on block %s"%self.name,1)
    self.mode='commit'

    self._resultSet.setRecord(self._precommitRecord)

    if not self._dataSourceLink.hasMaster():
      self._resultSet.post()
      self._dataSourceLink._dataObject.commit()

    self._recordCount = self._resultSet.getRecordCount()

    # If all our records were deleted, create an empty record
    if not self._recordCount:
      self.newRecord()
    else:
      self.jumpRecord(self._resultSet.getRecordNumber())

    self.mode='normal'

  #
  # processPostChanges
  #
  def processPostChanges(self):
    GDebug.printMesg(1, "processing post changes on block %s"%self.name,1)
    self.mode='commit'

    self._resultSet.setRecord(self._precommitRecord)

    if not self._dataSourceLink.hasMaster():
      if not str(self._resultSet).count('gnue.common.dbdrivers._empty.DBdriver.NIL_ResultSet'):
        self._resultSet.post({},0)

    self._recordCount = self._resultSet.getRecordCount()

    # If all our records were deleted, create an empty record
    if not self._recordCount:
      self.newRecord()
    else:
      self.jumpRecord(self._resultSet.getRecordNumber())

    self.mode='normal'


  #
  # processClear
  #
  def processClear(self):
    #self._dataSourceLink._dataObject.rollback()
    if not self._dataSourceLink.hasMaster():
      self._dataSourceLink.createEmptyResultSet()
    else:
      self._dataSourceLink.createEmptyResultSet(masterRecordSet=self._resultSet._masterRecordSet)
    self._currentRecord = 0
    self._recordCount = 0
    self.switchRecord(0)


  def reset(self):
    self.processRollback(1)

  def clear(self):
    self.processRollback()

  #
  # processRollback
  #
  # if recovering=0, then the user requested a rollback
  # if recovering=1, then a commit or such failed and we need to clean up
  # (but not lose state information)
  #
  def processRollback(self, recovering=0):
    if not recovering:
      self._currentRecord = 0
      self._recordCount = 0
      self._dataSourceLink.createEmptyResultSet()
    else:
      cachedRecord = self._dataSourceLink.getCurrentResultSet().getCachedRecords()
      for tmpRecord in cachedRecord:
        if tmpRecord.isInserted():
          tmpFields = tmpRecord.getFields() 
          for key in tmpFields.keys():
            tmpFields[key] = ''
          tmpRecord._emptyFlag = 1
        else:  
          tmpRecord.getFields().update(tmpRecord.getInitialData())
        tmpRecord.setUpdateFlag(0)

    #self._dataSourceLink._dataObject.rollback()
    self.switchRecord(0)


  #
  # initQuery and processQuery
  #
  def initQuery(self):

    # If Enter-Query is hit once, enter query mode
    # If Enter-Query is hit twice, bring back conditions from last query.
    # If Enter-Query is hit thrice, cancel the query and go into normal mode.
    if self.mode != 'query':
      self.mode = 'query'
      self._query2 = int(gConfig("RememberLastQuery"))
      self._queryValues = {}
      self._queryValues.update(self._queryDefaults)
      self.switchRecord(0)
    elif self._query2:
      self._query2 = 0
      self._queryValues = {}
      self._queryValues.update(self._lastQueryValues)
      self.switchRecord(0)
    else:
      self.mode = 'normal'
      self.switchRecord(0)

    # Removed to allow the GFForm object to deside when to clear the block
    # self.processRollback()

  def processQuery(self):
    # Implement sloppyQuery system
    if self.mode == 'query':
      for entry in self._entryList:
        if hasattr(entry,'sloppyQuery'):
          newValue = '%'
          value = entry._value
          for letter in value:
            if letter != "%":
              newValue += letter+"%"
          entry.setValue(newValue)

      self.mode = 'normal'

      self._lastQueryValues = {}
      self._lastQueryValues.update(self._queryValues)

      conditionLike = {}
      conditionEq = {}

      # Get all the user-supplied parameters from the entry widgets
      for entry in self._queryValues.keys():
        if entry._bound and entry.isQueryable() and len(str(self._queryValues[entry])):
          if entry.typecast == 'text':
            GDebug.printMesg(5,'Adding conditional LIKE (%s like %s)' % (entry.field, self._queryValues[entry]))
            conditionLike[entry.field] = self._queryValues[entry]
          else:
            GDebug.printMesg(5,'Adding conditional AND (%s=%s)' % (entry.field, self._queryValues[entry]))
            conditionEq[entry.field] = self._queryValues[entry]

      if len(conditionLike.keys()) and len(conditionEq.keys()): 
        GDebug.printMesg(5,'Combining like w/and (%s, %s)' % (conditionLike, conditionEq))
        conditions = GConditions.combineConditions( \
            GConditions.buildConditionFromDict(conditionLike, GConditions.GClike), \
            GConditions.buildConditionFromDict(conditionEq, GConditions.GCeq) )

      elif len(conditionLike.keys()): 
        conditions = GConditions.buildConditionFromDict(conditionLike, GConditions.GClike)

      elif len(conditionEq.keys()):
        conditions = GConditions.buildConditionFromDict(conditionEq, GConditions.GCeq)

      else: 
        conditions = {}

      self._dataSourceLink.createResultSet(conditions)
      self._recordCount = self._resultSet.getRecordCount()

      self.processTrigger('POST-QUERY')
      # update the scrollbars...
      for sb in self._scrollBars:
        sb.update (self._currentRecord, self._recordCount)

  def registerScrollBar (self, sb):
    self._scrollBars.append (sb)

