#
# Copyright 2000, 2001 Free Software Foundation
#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# FILE:
# GFParser.py
#
# DESCRIPTION:
# Class that contains a sax based xml processor for GNUE forms
#
# NOTES:
#

from gnue.common import GDataSource, GParser, GMenu

import copy, types
from gnue.common import GTypecast






########
########  Please keep this file neat !!!
########






#######################################################
# This method loads a form from an XML file and returns
# a GFForm object.  If initialize is 1 (default), then
# the form is initialized and ready to go.
#
# (initialize=0 is currently not used -- will probably
#  be used in the Forms Designer package where we will
#  not want the loaded form to connect to databases, etc)
#######################################################

def loadForm(buffer, app, initialize=1, parameters={}):
  return GParser.loadXMLObject (buffer, xmlFormsHandler, 'GFForm', 'form',
           initialize, attributes={"_app": app,
                               "_connections": app.connections}, initParameters=parameters)


xmlElements = None


def getXMLelements():

  global xmlElements

  if xmlElements == None:
    from gnue.forms import GFObjects, GFLibrary, GFTrigger, GFForm

    xmlElements = {
      'form': {
         'BaseClass': GFForm.GFForm,
         'Required': 1,
         'SingleInstance': 1,
         'Attributes': {
            'title': {
               'Typecast': GTypecast.text,
               'Default': 'Untitled Form' },
            'readonly': {
               'Typecast': GTypecast.boolean,
               'Default': 0 },
            'tabbed': {
               'Typecast': GTypecast.name,
               'ValueSet': {
                  'left': {},
                  'right': {},
                  'bottom': {},
                  'top': {} },
               'Default': "" },
            'width': {
               'Typecast': GTypecast.whole,
               'Default': 40 },
            'height': {
               'Typecast': GTypecast.whole,
               'Default': 12 } },
         'ParentTags': None },

      'import': {
         'BaseClass': GFLibrary.GFImport,
         'Attributes': {
            'library': {
               'Required': 1,
               'Typecast': GTypecast.name },
            'datasources': {
               'Typecast': GTypecast.name,
               'Default': "" },
            'pages': {
               'Typecast': GTypecast.name,
               'Default': ""  },
            'triggers': {
               'Typecast': GTypecast.name,
               'Default': "" } },
         'ParentTags': ('form',) },

      'database': {
         'BaseClass': GFObjects.GFDatabase,
         'Attributes': {
            'name': {
               'Required': 1,
               'Unique': 1,
               'Typecast': GTypecast.name },
            'provider': {
               'Required': 1,
               'Typecast': GTypecast.name },
            'dbname': {
               'Required': 0,
               'Typecast': GTypecast.text },
            'service': {
               'Required': 0,
               'Typecast': GTypecast.text },
            'comment': {
               'Required': 0,
               'Typecast': GTypecast.text },
            'host': {
               'Required': 0,
               'Typecast': GTypecast.text } },
         'Deprecated': 'Use the external connections file format.',
         'ParentTags': ('form',) },

      'page': {
         'BaseClass': GFObjects.GFPage,
         'Required': 1,
         'Attributes': {
            'name': {
               'Unique': 1,
               'Typecast': GTypecast.name } },
         'ParentTags': ('form',) },

      'block': {
         'BaseClass': GFObjects.GFBlock,
         'Attributes': {
            'name': {
               'Required': 1,
               'Unique': 1,
               'Typecast': GTypecast.name },
            'rows': {
               'Typecast': GTypecast.whole},
            'rowSpacer': {
               'Typecast': GTypecast.whole },
            'transparentBlock':{
               'Typecast': GTypecast.boolean,
               'Default': 0 },
            'restrictDelete':{
               'Typecast': GTypecast.boolean,
               'Default': 0 },
            'restrictInsert':{
               'Typecast': GTypecast.boolean,
               'Default': 0 },
            'datasource': {
               'References': (('datasource','name'),),
               'Typecast': GTypecast.name } },
         'ParentTags': ('page',) },

      'label': {
         'BaseClass': GFObjects.GFLabel,
         'Attributes': {
            'name': {
               'Unique': 1,
               'Typecast': GTypecast.name },
            'text': {
               'Required': 1,
               'Typecast': GTypecast.text },
            'alignment': {
               'Typecast': GTypecast.name,
               'ValueSet': {
                  'left': {},
                  'right': {},
                  'center': {} },
               'Default': "left"},
            'width': {
               'Typecast': GTypecast.whole },
            'rows': {
               'Typecast': GTypecast.whole },
            'rowSpacer': {
               'Typecast': GTypecast.whole },
            'x': {
               'Required': 1,
               'Typecast': GTypecast.whole },
            'y': {
               'Required': 1,
               'Typecast': GTypecast.whole } },
         'ParentTags': ('page','block',) },

      # If you implement a new entry "style", add to the entryStyles
      # structure after this list
      'entry': {
         'BaseClass': GFObjects.GFEntry,
         'Attributes': {
            'name': {
               'Required': 1,
               'Unique': 1,
               'Typecast': GTypecast.name },
            'trigger': {
               'Typecast': GTypecast.name },
            'field': {
               'Typecast': GTypecast.name },
            'height': {
               'Typecast': GTypecast.whole,
               'Default': 1 },
            'width': {
               'Typecast': GTypecast.whole },
            'max_length': {
               'Typecast': GTypecast.whole },
            'visibleCount':{
               'Typecast': GTypecast.whole,
               'Deprecated': 'Use the block "rows" attribute instead.' },
            'rows': {
               'Typecast': GTypecast.whole},
            'rowSpacer': {
               'Typecast': GTypecast.whole },
            'readonly': {
               'Typecast': GTypecast.boolean,
               'Default': 0   },
            'required': {
               'Description': 'This object cannot have an empty value prior '
                              'to a commit.',
               'Typecast': GTypecast.boolean,
               'Default': 0   },
            'uppercase': {
               'Deprecated': 'Use case="upper" instead.',
               'Typecast': GTypecast.boolean,
               'Default': 0   },
            'lowercase': {
               'Deprecated': 'Use case="lower" instead.',
               'Typecast': GTypecast.boolean,
               'Default': 0   },
            'numeric': {
               'Deprecated': 'Use typecast="number" instead',
               'Typecast': GTypecast.boolean,
               'Default': 0   },
            'hidden': {
               'Typecast': GTypecast.boolean,
               'Default': 0   },
            'style': {
               'Typecast': GTypecast.name,
               'ValueSet': {
                  'default': {},
                  'dropdown': {},
                  'checkbox': {},
                  'textlookup': {} },
               'Default': 'default'},
            'case': {
               'Typecast': GTypecast.name,
               'ValueSet': {
                  'mixed': {},
                  'upper': {},
                  'lower': {} },
               'Default': 'mixed'},
            'typecast': {
               'Typecast': GTypecast.name,
               'ValueSet': {
                  'text': {},
                  'number': {},
                  'date': {} },
               'Default': 'text'},
            'formatmask': {
               'Typecast': GTypecast.text },
            'inputmask': {
               'Typecast': GTypecast.text },
            'displaymask': {
               'Typecast': GTypecast.text },
            'value': {
               'Typecast': GTypecast.text },
            'foreign_key': {
               'Typecast': GTypecast.name },
            'default': {
               'Typecast': GTypecast.text },
            'queryDefault':{
               'Typecast': GTypecast.text },
            'foreign_key_description': {
               'Typecast': GTypecast.text },
            'sloppyQuery': {
               'Typecast': GTypecast.text },
            'ignoreCaseOnQuery': {
               'Typecast': GTypecast.boolean,
               'Default': 0 },
            'editOnNull': {
               'Description': 'Only allow this object to be edited if it '
                              'is currently empty.',
               'Typecast': GTypecast.boolean,
               'Default': 0 },
            'no_ltrim': {
               'Description': 'Suppress trimming of extraneous space at '
                              'beginning of user input.',
               'Typecast': GTypecast.boolean,
               'Default': 0 },
            'no_rtrim': {
               'Description': 'Suppress trimming of extraneous space at end '
                              'of user input.',
               'Typecast': GTypecast.boolean,
               'Default': 0 },
            'x': {
               'Required': 1,
               'Typecast': GTypecast.whole },
            'y': {
               'Required': 1,
               'Typecast': GTypecast.whole } },
         'ParentTags': ('block',) },

      'scrollbar': {
         'BaseClass': GFObjects.GFScrollBar,
         'Attributes': {
            'width': {
               'Required': 1,
               'Typecast': GTypecast.whole },
            'height': {
               'Required': 1,
               'Typecast': GTypecast.whole },
            'x': {
               'Required': 1,
               'Typecast': GTypecast.whole },
            'y': {
               'Required': 1,
               'Typecast': GTypecast.whole },
            'page': {
              'Required': 1,
              'Typecast': GTypecast.whole
            },
         },
         'ParentTags': ('page','block',) },

      'box': {
         'BaseClass': GFObjects.GFBox,
         'Attributes': {
            'name': {
               'Unique': 1,
               'Typecast': GTypecast.name },
            'label': {
               'Typecast': GTypecast.text },
            'width': {
               'Required': 1,
               'Typecast': GTypecast.whole },
            'height': {
               'Required': 1,
               'Typecast': GTypecast.whole },
            'x': {
               'Required': 1,
               'Typecast': GTypecast.whole },
            'y': {
               'Required': 1,
               'Typecast': GTypecast.whole } },
         'ParentTags': ('page','block',) },

      'button': {
         'BaseClass': GFObjects.GFButton,
         'Attributes': {
            'name': {
               'Unique': 1,
               'Typecast': GTypecast.name },
            'trigger': {
               'Typecast': GTypecast.name },
            'label': {
               'Typecast': GTypecast.name },
            'width': {
               'Required': 1,
               'Typecast': GTypecast.whole },
            'height': {
               'Required': 1,
               'Typecast': GTypecast.whole },
            'x': {
               'Required': 1,
               'Typecast': GTypecast.whole },
            'y': {
               'Required': 1,
               'Typecast': GTypecast.whole } },
         'ParentTags': ('page','block',) },

      'trigger': {
         'BaseClass': GFTrigger.GFTrigger,
         'Attributes': {
            'name': {
               'Unique': 1,
               'Typecast': GTypecast.name },
            'id': {
               'Deprecated': 'Use name instead.',   # DEPRECATED: Use name instead
               'Typecast': GTypecast.name },
            'type': {
               'Typecast': GTypecast.name },
            'src': {
               'References': (('trigger','name'),),
               'Typecast': GTypecast.name },
            'language': {
               'Typecast': GTypecast.name,
               'ValueSet': {
                   'python': {} },
               'Default': 'python' } },
         'MixedContent': 1,
         'KeepWhitespace': 1,
         'UsableBySiblings': 1,
         'ParentTags': ('form',) },

      'options': {
         'BaseClass': GFObjects.GFOptions,
         'SingleInstance': 1,
         'UsableBySiblings': 1,
         'ParentTags': ('form',) },

      'option': {
         'BaseClass': GFObjects.GFOption,
         'Attributes': {
            'name': {
               'Required': 1,
               'Typecast': GTypecast.name },
            'value': {
               'Typecast': GTypecast.text } },
         'MixedContent': 1,
         'ParentTags': ('options',) },

      'title': {
         'BaseClass': GFObjects.GFOption,
         'Attributes': {
            'name': {
               'Typecast': GTypecast.name,
               'ValueSet': {
                   'title': {} },
               'Default': 'title' },
            'value': {
               'Typecast': GTypecast.text } },
         'MixedContent': 1,
         'SingleInstance': 1,
         'Deprecated': 'Use the <form> attribute "title" instead.',
         'ParentTags': ('options',) },

      'name': {
         'BaseClass': GFObjects.GFOption,
         'Attributes': {
            'name': {
               'Typecast': GTypecast.name,
               'ValueSet': {
                   'name': {} },
               'Default': 'name' },
            'value': {
               'Typecast': GTypecast.text } },
         'MixedContent': 1,
         'SingleInstance': 1,
         'ParentTags': ('options',) },

      'height': {
         'BaseClass': GFObjects.GFOption,
         'Attributes': {
            'name': {
               'Typecast': GTypecast.name,
               'ValueSet': {
                   'height': {} },
               'Default': 'height' },
            'value': {
               'Typecast': GTypecast.text } },
         'MixedContent': 1,
         'SingleInstance': 1,
         'Deprecated': 'Use the <form> attribute "height" instead.',
         'ParentTags': None },

      'width': {
         'BaseClass': GFObjects.GFOption,
         'Attributes': {
            'name': {
               'Typecast': GTypecast.name,
               'ValueSet': {
                   'width': {} },
               'Default': 'width' },
            'value': {
               'Typecast': GTypecast.text } },
         'MixedContent': 1,
         'SingleInstance': 1,
         'Deprecated': 'Use the <form> attribute "width" instead.',
         'ParentTags': ('options',) },

      'author': {
         'BaseClass': GFObjects.GFOption,
         'Attributes': {
            'name': {
               'Typecast': GTypecast.name,
               'ValueSet': {
                   'author': {} },
               'Default': 'author' },
            'value': {
               'Typecast': GTypecast.text } },
         'MixedContent': 1,
         'SingleInstance': 1,
         'ParentTags': ('options',) },

      'description':{
         'BaseClass': GFObjects.GFOption,
         'Attributes': {
            'name': {
               'Typecast': GTypecast.name,
               'ValueSet': {
                   'description': {} },
               'Default': 'description' },
            'value': {
               'Typecast': GTypecast.text } },
         'MixedContent': 1,
         'SingleInstance': 1,
         'ParentTags': ('options',) },

      'version': {
         'BaseClass': GFObjects.GFOption,
         'Attributes': {
            'name': {
               'Typecast': GTypecast.name,
               'ValueSet': {
                   'version': {} },
               'Default': 'version' },
            'value': {
               'Typecast': GTypecast.text } },
         'MixedContent': 1,
         'SingleInstance': 1,
         'ParentTags': ('options',) },

      'tip': {
         'BaseClass': GFObjects.GFOption,
         'Attributes': {
            'name': {
               'Typecast': GTypecast.name,
               'ValueSet': {
                   'tip': {} },
               'Default': 'tip' },
            'value': {
               'Typecast': GTypecast.text } },
         'MixedContent': 1,
         'SingleInstance': 1,
         'ParentTags': ('options',) },
      'parameter':    {
         'BaseClass': GFObjects.GFParameter,
         'Attributes': {
            'name':          {
               'Required': 1,
               'Unique': 1,
               'Typecast': GTypecast.name },
            'required':    {
               'Typecast': GTypecast.boolean,
               'Default': 0 },
##            'limited':     {
##               'Typecast': GTypecast.boolean,
##               'Default': 0 },
            'default':     {
               'Typecast': GTypecast.text },
            'description': {
##               'Required': 1,
               'Typecast': GTypecast.text },
##            'source':      {
##               'Typecast': GTypecast.name },
            'type':        {
               'Typecast': GTypecast.name,
               'Default': "char" } },
         'ParentTags':  ('form',) },

      'maintoolbar': {
        'BaseClass': GFForm.GFMainToolBar,
        'Required': 0,
        'SingleInstance': 1,
        'Attributes': {
          'enabled': {
            'Typecast': GTypecast.boolean,
            'Default': 1,
          },
        },
        'ParentTags': ('form',),
      },

      'mainmenubar': {
        'BaseClass': GFForm.GFMainMenuBar,
        'Required': 0,
        'SingleInstance': 1,
        'Attributes': {
          'enabled': {
            'Typecast': GTypecast.boolean,
            'Default': 1,
          },
        },
        'ParentTags': ('form',),
      },

    }


    #
    # Add DataSource elements
    #
    xmlElements.update(
      GDataSource.getXMLelements(
        updates={'datasource': {
                   'BaseClass': GFObjects.GFDataSource,
                   'ParentTags': ('form',) },
                 'cparam': {
                   'BaseClass': GFObjects.GFCParam }
                 }))

    #
    # Add menu elements
    #
    xmlElements.update(
        GMenu.getXMLelements(
                updates={'menu': {
                   ##'BaseClass': GFObjects.GFMenu,
                   'ParentTags': ('form',) },
                 }))

    #
    # Add importable elements
    # If an object needs to be importable,
    # simply add its tag name to the tuple below
    # and make sure it has a "name" attribute
    # (otherwise we don't know how to reference
    # it in the imported file).
    #
    for key in ('datasource','page','block','trigger','button','entry'):
      name = "import-%s" % key
      copy._deepcopy_dispatch[types.FunctionType] = copy._deepcopy_atomic
      copy._deepcopy_dispatch[types.ClassType] = copy._deepcopy_atomic
      copy._deepcopy_dispatch[type(int)] = copy._deepcopy_atomic

      p = copy.deepcopy(xmlElements[key])
      p['BaseClass'] = GFLibrary.GFImportItem
      p['Attributes']['library'] = {
         'Required': 1,
         'Typecast': GTypecast.name }
      p['MixedContent'] = 0
      p['Required'] = 0

      xmlElements[name] = p

  return xmlElements

#
# Different Types of Entries
#
entryStyles = {'text': 'Text Field',
               'dropdown': 'Drop Down Box',
               'checkbox': 'Check Box' ,
               'textlookup': 'Text Lookup'}

#######################################################
#
# xmlFormsHandler
#
# This class is called by the XML parser to
# process the xml file.
#
#######################################################

class xmlFormsHandler (GParser.xmlHandler):
  def __init__(self):

    GParser.xmlHandler.__init__(self)

    self.xmlElements = getXMLelements()


