#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it 
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either 
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied 
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public 
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place 
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
#
# FILE:
# GFTrigger.py
#
# DESCRIPTION:
# Base class for trigger objects
#
# NOTES:
#
from gnue.common.GParser import GContent, GDebug
from gnue.common.GObjects import GObj
from gnue.common.GTrigger import GTrigger, TriggerError

import string

#
# A list of all valid triggers, and their "pretty" names
#
VALIDTRIGGERS = { 'PRE-FOCUSOUT':   'Pre-FocusOut',
                  'POST-FOCUSOUT':  'Post-FocusOut',
                  'PRE-FOCUSIN':    'Pre-FocusIn',
                  'POST-FOCUSIN':   'Post-FocusIn',
                  'PRE-COMMIT':     'Pre-Commit',
                  'POST-COMMIT':    'Post-Commit',
                  'POST-QUERY':     'Post-Query',
                  'ON-SWITCH':      'On-Switch',
                  'PRE-CHANGE':     'Pre-Change',
                  'POST-CHANGE':    'Post-Change',
                  'PRE-UPDATE':     'Pre-Update',
                  'PRE-INSERT':     'Pre-Insert',
                  'PRE-DELETE':     'Pre-Delete',
                  'PRE-MODIFY':     'Pre-Modify', 
                  'ON-NEWRECORD':   'On-NewRecord',
                  'ON-STARTUP':     'On-Startup',
                  'ON-EXIT':        'On-Exit',
                }

#
# GFTriggerAware
#
# Objects that inherit this class will be
# capable of processing triggers
#
class GFTriggerAware:
  def __init__(self):
    self._trigger = {}
    self._validTriggers = VALIDTRIGGERS.keys()

  # addTrigger
  # Associates a trigger with the object
  def addTrigger(self, key, function):
    key = string.upper(key)
    if not self._trigger.has_key(key):
      self._trigger[key] = []
    self._trigger[key].append(function)

  # processTrigger
  # "fires" the trigger
  def processTrigger(self, key):
    key = string.upper(key)
    if self._validTriggers.count(key):
      if self._trigger.has_key(key):
        for function in self._trigger[key]:
          function(self)
      else:
        GDebug.printMesg(10, "No triggers to fire")
    else:
      print _("Invalid trigger "),key

#
# GFTrigger
#
# Class used to implement triggers
#
class GFTrigger(GTrigger):

  def __init__(self, parent=None, type=None, name=None, src=None, text=None, language='python'):
    GObj.__init__(self, parent)
    self._type = "GFTrigger"

    self._text=''
    self.language=language
    self._triggerns={}
    self.src = src
    self.type = type
    self.name = name
    self._inits   = [None,self.initialize]
    if text != None:
      GContent(self, text)
    if self.type != None:
      self._buildObject()








