#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2002 Free Software Foundation
#
# FILE:
# UIwxpython.py
#
# DESCRIPTION:
# A wxPython based user interface driver for GNUE forms.
#
# NOTES:
#  TODO: why does UIhelper not inherit from UIwidget?
#

from wxPython.wx import *
from wxPython.lib.splashscreen import SplashScreen
from gnue.forms.GFForm import *
from gnue.forms.GFEvent import *
from gnue.forms import GFKeyMapper
from gnue.forms.uidrivers._base.UIdriver import *
from gnue.forms.uidrivers._base.UIWidget import *

import os
import sys
import string
from gnue.forms import VERSION
from gnue.common import GDebug, GLoginHandler
from gnue.common import GConfig
from gnue.common import lineWrap

_LOOPTRAP = 0
_PROMPTFORRECORD = None
_charWidth = 0
_IdToWxObj = {}
_IdToGFObj = {}
_IdToUIObj = {}

# hack for wxPython 2.2 (2.3+ doesn't need this)
def _eventObjTowxWindow(event):
  eo = event.GetEventObject()
  return wxPyTypeCast(eo, 'wxWindow')

##
## Only create one instance of wxApp
## Use getWxApp() to get or create the
## single instance
##
## TODO: It seems this could all be folded
## TODO: into an __init__ inside GFwxApp
##
__wxApp = None

def getWxApp():
  global __wxApp

  if not __wxApp:
    GDebug.printMesg(7,"WxApp initializing")
    __wxApp = GFwxApp(0)
  return __wxApp

class GFwxApp(wxApp):
  _mainLoop= None

  def OnInit(self):
    GDebug.printMesg(7,"WxApp initializing")
    wxInitAllImageHandlers()
    return true

  def MainLoop (self):
    # just to launch one wx's mainloop
    # otherwise, the app does not finishes cleanly when several windows are open
    # 'cause the first mainloop gets all the events
    # and the others just sit and wait for something to happen
    if self._mainLoop==None:
      self._mainLoop= 1
      wxApp.MainLoop (self)

#
# Little global helper routine to set font according to options
#
# TODO: Not completed
#
def initFont(widget, affectsLayout=1):
  if int(gConfig('fixedWidthFont')):
    widget.SetFont(wxFont(int(gConfig('pointSize')),wxMODERN,wxNORMAL,wxNORMAL, 0, "lucidatypewriter"))

#
# GFUserInterface
#
# The public interface to the User Interface
# All UIs must provide this class
#
class GFUserInterface(GFUserInterfaceBase):
  def __init__(self, eventController, disableSplash = None, wId= 0):
    GDebug.printMesg (1, "wId: %d" % wId)

    self._WIDGETS = WIDGETS

    self._DIALOGS = {'about'       :UIAbout,
                     'messageBox'  :UIMessageBox,
                     'genericBox'  :UIGenericBox,
                     }

    self._aboutDialog = UIAbout
    self._messageDialog = UIMessageBox

    self._wxapp = getWxApp()
    self._disableSplash = disableSplash
    GFUserInterfaceBase.__init__(self,eventController)

    global _PROMPTFORRECORD
    _PROMPTFORRECORD = self.promptForRecordNumber

    self.forms = {}

    self._windowId= wId
    self.init()

  #
  # init
  #
  # Routine called by wxWindows when wxApp instance is created
  # It is used the build the basic UI
  #
  def init(self):

    #
    # SplashScreen
    #
    if not self._disableSplash:
      self.splash = SplashScreen(None,bitmapfile=images_dir+gConfig('splashScreenPNG'),
                                 duration = 4000, callback=self.onCloseSplash,
                                 style=wxSIMPLE_BORDER|wxCENTRE_ON_SCREEN )
      swidth,sheight = self.splash.GetSizeTuple()

      # Make the splash text print white
      self.splash.SetForegroundColour(wxColour(0xFF, 0xFF, 0xFF))
      self.splash.SetFont(wxFont(20,wxSWISS,wxNORMAL,wxBOLD))

      text = _("GNU Enterprise Forms")
      twidth,theight = self.splash.GetTextExtent(text)
      wxStaticText(parent=self.splash, id=-1, label=str(text),
                   size=wxSize(swidth,theight),pos=wxPoint(1,1),style=wxALIGN_CENTRE)

      self.splash.SetFont(wxFont(10,wxSWISS,wxNORMAL,wxBOLD))

      text = _("Version %s") % VERSION
      twidth,theight = self.splash.GetTextExtent(text)
      wxStaticText(parent=self.splash, id=-1, label=str(text),
                   size=wxSize(swidth,theight), pos=wxPoint(1,25),style=wxALIGN_CENTRE)

      text = "(c)2000-2002 Free Software Foundation"
      twidth,theight = self.splash.GetTextExtent(text)
      wxStaticText(self.splash, -1, str(text),
                   wxPoint(1,50),wxSize(swidth,theight),wxALIGN_CENTRE)

      text = _("Protected by GNU General Public Licence 2.0")
      wxStaticText(self.splash, -1, str(text),
                   wxPoint(1,sheight-(theight+10)),wxSize(swidth-5,theight),wxALIGN_RIGHT)

      self.splash.CenterOnScreen()
      self.splash.Show(true)

    #
    # Create the main window
    #
    self.mainWindow = wxFrame(NULL, -1, "", wxDefaultPosition)
    initFont(self.mainWindow,1)

    #
    # Find the max size of printable characters
    #  used to setup the grid
    #
    maxWidth, maxHeight, maxDescent, maxLeading = [0,0,0,0]
    checkchars = string.printable
    for letter in checkchars:
      width,height,descent,leading = self.mainWindow.GetFullTextExtent(letter)
      maxWidth = max(maxWidth,width)
      maxHeight = max(maxHeight,height)
      maxDescent = max(maxDescent,descent)
      maxLeading = max(maxLeading,leading)

    self.charWidth = int(maxWidth+maxLeading)
    self.charHeight = int(maxHeight+maxDescent)

    self.widgetWidth = self.charWidth       # The pixel width of a 1 char widget (for things like buttons)
    self.widgetHeight = self.charHeight + 3 # The pixel height of a 1 char widget (for things like buttons)
    self.textWidth = self.charWidth         # The pixel width of text inside a widget
    self.textHeight = self.charHeight       # The pixel height of text inside a widget

    global _charWidth
    _charWidth = self.charWidth             # Used by the mouse handler to compute char offset into widget


    self.menu_sb_space=4 # the extra spaces needed by the menu, toolbar and statusbar

    self.statusBar = self.mainWindow.CreateStatusBar()
    self.statusBar.SetFieldsCount(5)
    self.statusBar.SetStatusWidths([-1,50,50,75,75])
    initFont(self.statusBar,0)

    # this must be kept here so toolbar can work without menu
    eventBase= self._windowId*1000
    EVT_MENU(self._wxapp, eventBase+100, lambda event, l=self: l.dispatchEvent(GFEvent('requestCOMMIT')))
    EVT_MENU(self._wxapp, eventBase+102, lambda event, l=self: l.dispatchEvent(GFEvent('requestROLLBACK')))
    EVT_MENU(self._wxapp, eventBase+104, lambda event, l=self: l.dispatchEvent(GFEvent('requestPRINTOUT')))
    EVT_MENU(self._wxapp, eventBase+101,  self.menuExitEvent)

    EVT_MENU(self._wxapp, eventBase+401, lambda event, l=self: l.dispatchEvent(GFEvent('requestCUT')))
    EVT_MENU(self._wxapp, eventBase+402, lambda event, l=self: l.dispatchEvent(GFEvent('requestCOPY')))
    EVT_MENU(self._wxapp, eventBase+403, lambda event, l=self: l.dispatchEvent(GFEvent('requestPASTE')))

    EVT_MENU(self._wxapp, eventBase+211, lambda event, l=self: l.dispatchEvent(GFEvent('requestFIRSTRECORD')))
    EVT_MENU(self._wxapp, eventBase+202, lambda event, l=self: l.dispatchEvent(GFEvent('requestPREVRECORD')))
    EVT_MENU(self._wxapp, eventBase+201, lambda event, l=self: l.dispatchEvent(GFEvent('requestNEXTRECORD')))
    EVT_MENU(self._wxapp, eventBase+210, lambda event, l=self: l.dispatchEvent(GFEvent('requestLASTRECORD')))
    EVT_MENU(self._wxapp, eventBase+209, _PROMPTFORRECORD)
    EVT_MENU(self._wxapp, eventBase+203, lambda event, l=self: l.dispatchEvent(GFEvent('requestNEWRECORD')))
    EVT_MENU(self._wxapp, eventBase+204, lambda event, l=self: l.dispatchEvent(GFEvent('requestMARKFORDELETE')))
    EVT_MENU(self._wxapp, eventBase+205, lambda event, l=self: l.dispatchEvent(GFEvent('requestNEXTBLOCK')))
    EVT_MENU(self._wxapp, eventBase+206, lambda event, l=self: l.dispatchEvent(GFEvent('requestPREVBLOCK')))
    EVT_MENU(self._wxapp, eventBase+207, lambda event, l=self: l.dispatchEvent(GFEvent('requestENTERQUERY')))
    EVT_MENU(self._wxapp, eventBase+208, lambda event, l=self: l.dispatchEvent(GFEvent('requestEXECQUERY')))

    EVT_MENU(self._wxapp, eventBase+300, lambda event, l=self: l.dispatchEvent(GFEvent('requestABOUT')))

    self.mainWindow.panel = wxPanel(self.mainWindow,-1, wxDefaultPosition,self.mainWindow.GetSize())

    self.currentWidget = [self.mainWindow.panel]
    self._wxapp.SetTopWindow(self.mainWindow)

    EVT_CLOSE(self.mainWindow,self.closeTrap)

    return true

  def activateForm(self, form):
    width = int(self._form.width)
    height = int(self._form.height)

    formSize = wxSize(width*self.widgetWidth,
                      (height+self.menu_sb_space)*self.widgetHeight)

    self.mainWindow.SetSize(formSize)
    self.mainWindow.SetTitle(str(self._form.title))
    self.mainWindow.panel =  self._pageList[0]
    self.mainWindow.panel.SetSize(formSize)
    self._pageList[0].Show(TRUE)

    # Only one page at a time can be visible
    self.visiblePage = self._pageList[0]

    if self._notebook: # Adjust sizes
      self._notebook.SetSize(formSize)

      child = self._notebook.GetParent()
      while child:
        child.Fit()
        child = child.GetParent()

    self.mainWindow.CenterOnScreen()
    self.mainWindow.Show(true)


  #############################################################################
  #
  # Private UIBase support functions
  #
  # Called in UIbase functions to perform UI interaction with this specific
  # widget set.
  #

  #
  # _exit
  #
  # Tells the application to close it's main window
  #
  def _exit(self):
    self.mainWindow.Close(true)

  #
  # _setStatusBar
  #
  def _setStatusBar(self,tip, statusValue, insertValue, currentRecord, maxRecord, currentPage, maxPage):
    if tip: self.statusBar.SetStatusText(str(tip),0)

    if statusValue:
      self.statusBar.SetStatusText(statusValue,1)

    if insertValue:
      self.statusBar.SetStatusText(insertValue,2)

    if currentRecord and  maxRecord:
      self.statusBar.SetStatusText(string.strip("%s/%s" % (currentRecord,maxRecord)),3)

    if currentPage and  maxPage:
      self.statusBar.SetStatusText(string.strip("%s/%s" % (currentPage,maxPage)),4)

    #
    # Adjust the status fields to reflect width of text in them
    #
    # This is a hack
    #
    tipWidth = -1 # Fill whatever remains
    statusWidth,unused = self.statusBar.GetTextExtent(self.statusBar.GetStatusText(1))
    insertWidth,unused = self.statusBar.GetTextExtent(self.statusBar.GetStatusText(2))
    recordWidth,unused = self.statusBar.GetTextExtent(self.statusBar.GetStatusText(3))
    pageWidth,unused = self.statusBar.GetTextExtent(self.statusBar.GetStatusText(4))

    # Either double the width if it's set or default to 5
    statusWidth = statusWidth and (statusWidth * 2) or 5
    insertWidth = insertWidth and (insertWidth * 2) or 5
    recordWidth = recordWidth and (recordWidth * 2) or 5
    pageWidth   = pageWidth   and (pageWidth   * 2) or 5

    self.statusBar.SetStatusWidths([tipWidth,statusWidth,insertWidth,recordWidth,pageWidth])


  # Can be used by UI drivers to add more attributes to the creation event.
  def _updateCreateEvent(self, event):
    event.eventHandler = self.dispatchEvent
    return event


  #############################################################################
  #
  # Incoming Event Processors
  #
  # Processes the incoming events from other objects
  #

  #
  # mainLoop
  #
  # The primary loop of the user interface.  Called once the UI is
  # fully activated
  #
  def mainLoop(self):
    self._wxapp.MainLoop() # simply call the wxApp's MainLoop method

  #
  # gotoPage
  #
  # makes the requested page visible on the screen
  #
  def gotoPage(self, event):
    if not self._notebook:
      self.visiblePage.Show(FALSE)
      self.visiblePage = self._formToUI[event.data][0]
      # TODO: move the calculatoin out of here
      self.visiblePage.SetSize(wxSize(self._form.width*self.widgetWidth,
                                      (self._form.height+self.menu_sb_space)*self.widgetHeight))
    else:
      self.visiblePage = self._formToUI[event.data][0]
      for count in range(self._notebook.GetPageCount()):
        if self.visiblePage.GetId() == self._notebook.GetPage(count).GetId():
          self._notebook.SetSelection(count)
          break
    self.visiblePage.Show(TRUE)

  #
  # formAlert
  #
  # Rings a bell and alters the statusbar to display
  # a line of text
  #
  def formAlert(self, event):
    wxBell()
    self.statusBar.SetStatusText(event.data,0)

  #
  # Called whenever forms goes into a "wait" state in which user cannot
  # interact with interface (e.g., while waiting for a query or a commit)
  #
  def beginWait (self, event):
    wxBeginBusyCursor()

  #
  # Called whenever forms leaves a "wait" state
  #
  def endWait (self, event):
    wxEndBusyCursor()


  #
  # Clipboard routines
  #
  # If a particular UI has a system-wide clipboard,
  # these methods should be overridden to use that
  # clipboard.
  #
  def getClipboardContents(self, event):

    if wxTheClipboard.Open():
      data = wxTextDataObject()
      success = wxTheClipboard.GetData(data)
      wxTheClipboard.Close()
    else:
      success = 0
      GDebug.printMesg(5,'Unable to open clipboard for read')

    if success:
      value = data.GetText()
    else:
      GDebug.printMesg(5,'Unable to obtain clipboard contents. Defaulting to Empty.')
      value = None

    GDebug.printMesg(5, "Getting clipboard value '%s'" % value)
    event.__result__ = value


  def setClipboardContents(self, event):

    GDebug.printMesg(5,"Setting clipboard '%s'" % event.text)

    if wxTheClipboard.Open():
      value = wxTheClipboard.SetData(wxTextDataObject(event.text))
      GDebug.printMesg(10,"Set Clipboard Status: %s" % value)
      wxTheClipboard.Close()

    else:
      GDebug.printMesg(5,'Unable to open clipboard for write')


  #############################################################################
  #
  # Internal Event Processors
  #
  # Processes the events from the widget set
  #

  #
  # menuExitEvent
  #
  # Catches the Exit event from the menu and make sure it closes in such a way to
  # trigger the closeTrap method
  #
  def menuExitEvent(self, event):
    self.mainWindow.Close()

  #
  # promptForRecordNumber
  #
  # Presents a dialog to user asking for the record number to jump to
  #
  def promptForRecordNumber(self, event= None):
    dlg = wxTextEntryDialog(self.mainWindow, _('Enter record number to jump to'),
                                                        _('Which record?'))
    if dlg.ShowModal() == wxID_OK:
      action = self.dispatchEvent(GFEvent('requestRECORDNUMBER', dlg.GetValue()))
    else:
      action = None

    dlg.Destroy()
    return action

  #
  # closeTrap
  #
  # intercepts the applications closure and generates an event to the form requesting
  # closure.  Allows the form to control closure.  If the form approves it'll send
  # back an event closing the application
  #`
  def closeTrap(self,event):
    GDebug.printMesg (1, "trapping close")
    GDebug.printMesg (1, "%s" % dir (event))
    object = _eventObjTowxWindow(event)
    if event.CanVeto():
      self.dispatchEvent(GFEvent('requestEXIT'))
    else:
      GDebug.printMesg (1, "destroyed")
      object.Destroy()

  # Called by SplashScreen when timer expires
  def onCloseSplash(self):
    # For some reason, a simple Close() doesn't cooperate with
    # the modal login box and all hell breaks loose
    self.splash.Show(0)
    self.splash.Destroy()


#############################################################################
#############################################################################
#
# UI Objects
#
# A set of standard UI objects that allow us to move more code into UIbase
# These map generic functionality to the specific UI widget set in use by
# this driver.
#
#############################################################################
#############################################################################

#
# UIHelper
#
# Provides a UI widget set of std functions to reduce the number of functions
# require in each UIclass listed below
class UIHelper:
  def showModal(self):
    self.ShowModal()

  def show(self, flag=TRUE):
    self.Show(flag)

  def destroy(self):
    self.Destroy()

  def indexedFocus(self, index):
    self.widgets[index].SetFocus()

  def setValue(self, value, index=0):
    # These must be here or dropdown style controls
    # will get events they shouldn't and break.
    self.widgets[index].SetEvtHandlerEnabled(FALSE)
    self.widgets[index].SetValue(value)
    self.widgets[index].SetEvtHandlerEnabled(TRUE)

  def setCursorPosition(self, position, index=0):
    try:
      self.widgets[index].SetInsertionPoint(position)
    except AttributeError:
      pass  # For label-style entries

  def setSelectedArea(self, selection1, selection2, index=0):
    try:
      self.widgets[index].SetSelection(selection1, selection2)
    except (AttributeError, TypeError):
      pass  # For label-style & dropdown entries


  def _addToCrossRef(self, widget,gfobject, uiobject):
    id = widget.GetId()
    _IdToWxObj[id]=widget
    _IdToGFObj[id]=gfobject
    _IdToUIObj[id]=uiobject


  def _deleteFromCrossRef(self, widget, object):
    id = widget.GetId()
    try:
      del _IdToWxObj[id]
      del _IdToGFObj[id]
      del _IdToUIObj[id]
    except:
      pass


  def createWidget(self, event, spacer):

    newWidget = self._createWidget(event, spacer)
    initFont(newWidget)
    self._addToCrossRef(newWidget,event.object,self)
    return newWidget

  def cleanup(self, object):
    for widget in self.widgets[:]:
      id = widget.GetId()
      del _IdToWxObj[id]
      del _IdToGFObj[id]
      del _IdToUIObj[id]
      self.widgets.pop(0)
      self._deleteFromCrossRef(widget, object)


class UIMainMenuBar (UIHelper, UIWidget):
  def _createWidget (self, event, spacer):
    window= event.interface.mainWindow
    eventBase= event.interface._windowId*1000

    fileMenu = wxMenu()
    fileMenu.Append( eventBase+100, _("Save       (F6)"),_('Save all changes to database.'))
    fileMenu.Append( eventBase+102, _("Clear Form (F11)"),_('Clear Form'))
    fileMenu.AppendSeparator()
    fileMenu.Append( eventBase+104, _("&Print"),_('Perform print routine for this form.'))
    fileMenu.AppendSeparator()
    fileMenu.Append( eventBase+101, _("E&xit"),_('Leave the application.'))

    editMenu = wxMenu()
    editMenu.Append(eventBase+401, _("Cut"),_("Cut the selected text"))
    editMenu.Append(eventBase+402, _("Copy"),_("Copy the selected text"))
    editMenu.Append(eventBase+403, _("Paste"),_("Paste text"))

    dataMenu = wxMenu()
    dataMenu.Append( eventBase+211, _("First Record        (Shft+Up)"),_('Navigate to first record in memory.'))
    dataMenu.Append( eventBase+202, _("Previous Record         (Up)"),_('Navigate to previous record in memory.'))
    dataMenu.Append( eventBase+201, _("Next Record           (Down)"),_('Navigate to next record in memory.'))
    dataMenu.Append( eventBase+210, _("Last Record    (Shft+Down)"),_('Navigate to last record in memory.'))
    dataMenu.Append( eventBase+209, _("Jump to Record...        (F2)"),_('Prompts for a record number to which the system should jump.'))
    dataMenu.Append( eventBase+203, _("New Record              (F12)"),_('Create a new record for data input.'))
    dataMenu.Append( eventBase+204, _("Mark Record for Delete   (F5)"),_('Mark record for removal at next commit.'))
    dataMenu.AppendSeparator()
    dataMenu.Append( eventBase+205, _("Next Block              (PgDn)"),_('Navigate to next data block.'))
    dataMenu.Append( eventBase+206, _("Previous Block       (PgUp)"),_('Navigate to previous data block.'))
    dataMenu.AppendSeparator()
    dataMenu.Append( eventBase+207, _("Enter Query             (F8)"),_('Switch to input query mask mode.'))
    dataMenu.Append( eventBase+208, _("Execute Query           (F9)"), _('Execute query using current mask.'))

    helpMenu = wxMenu()
    helpMenu.Append( eventBase+300, _("&About..."), _('Display info about GNUe Forms.'))

    menuBar = wxMenuBar()
    menuBar.Append( fileMenu, _("&File") )
    menuBar.Append( editMenu, _("&Edit") )
    menuBar.Append( dataMenu, _("&Data") )
    menuBar.Append( helpMenu, _("&Help") )

    window.SetMenuBar( menuBar )

    return menuBar

class UIMainToolBar (UIHelper, UIWidget):
  def _createWidget (self, event, spacer):
    window= event.interface.mainWindow
    eventBase= event.interface._windowId*1000

    #
    #  There has got to be better way to call events to ride off the EVT_MENU
    #  Ok if i give buttons same id as menu's they share events
    #  I dont know if this is horribly wrong or design
    #
    # mdione: as maintoolbar can only be son of a form....
    if sys.platform == 'win32':
      mainToolBar = window.CreateToolBar()
    else:
      mainToolBar = window.CreateToolBar(wxTB_HORIZONTAL|wxTB_DOCKABLE)

    mainToolBar.SetToolBitmapSize(wxSize(24,24))
    mainToolBar.AddSimpleTool(eventBase+100, wxImage(images_dir+gConfig('tb_commit'), wxBITMAP_TYPE_PNG).ConvertToBitmap(), _("Save  Changes"), _('Save all changes to database.'))
    mainToolBar.AddSimpleTool(eventBase+203, wxImage(images_dir+gConfig('tb_insert'), wxBITMAP_TYPE_PNG).ConvertToBitmap(), _("Insert Record"), _('Create a new record for data input.'))
    mainToolBar.AddSimpleTool(eventBase+204, wxImage(images_dir+gConfig('tb_delete'), wxBITMAP_TYPE_PNG).ConvertToBitmap(), _("Delete Record"), _('Mark record for removal at next commit.'))
    mainToolBar.AddSeparator()
    mainToolBar.AddSimpleTool(eventBase+211, wxImage(images_dir+gConfig('tb_first'), wxBITMAP_TYPE_PNG).ConvertToBitmap(), _("First Record"),_('Navigate to first record in memory.'))
    mainToolBar.AddSimpleTool(eventBase+202, wxImage(images_dir+gConfig('tb_previous'), wxBITMAP_TYPE_PNG).ConvertToBitmap(), _("Previous Record"),_('Navigate to previous record in memory.'))
    mainToolBar.AddSimpleTool(eventBase+201, wxImage(images_dir+gConfig('tb_next'), wxBITMAP_TYPE_PNG).ConvertToBitmap(), _("Next Record"),_('Navigate to next record in memory.'))
    mainToolBar.AddSimpleTool(eventBase+210, wxImage(images_dir+gConfig('tb_last'), wxBITMAP_TYPE_PNG).ConvertToBitmap(), _("Last Record"),_('Navigate to last record in memory.'))
    mainToolBar.AddSimpleTool(eventBase+209, wxImage(images_dir+gConfig('tb_jump'), wxBITMAP_TYPE_PNG).ConvertToBitmap(), _("Jump to Record"),_('Navigate to user specified record in memory.'))
    mainToolBar.AddSeparator()
    mainToolBar.AddSimpleTool(eventBase+207, wxImage(images_dir+gConfig('tb_query_prep'), wxBITMAP_TYPE_PNG).ConvertToBitmap(), _("Prepare Query"), _('Switch to input query mask mode.'))
    mainToolBar.AddSimpleTool(eventBase+208, wxImage(images_dir+gConfig('tb_query'), wxBITMAP_TYPE_PNG).ConvertToBitmap(), _("Execute Query"), _('Execute query using current mask.'))
    mainToolBar.AddSeparator()
    mainToolBar.AddSimpleTool(eventBase+102, wxImage(images_dir+gConfig('tb_rollback'), wxBITMAP_TYPE_PNG).ConvertToBitmap(), _("Clear Form"), _('Clear Form'))
    mainToolBar.AddSimpleTool(eventBase+101, wxImage(images_dir+gConfig('tb_exit'), wxBITMAP_TYPE_PNG).ConvertToBitmap(), _("Exit Form"), _('Exit Form'))
    mainToolBar.Realize()

    return mainToolBar


#
# UILabel
#
# Widget set specific function that creates a single instance of a label
#
class UILabel(UIHelper, UIWidget):
  alignmentStyle = { 'left'  : wxALIGN_LEFT|wxST_NO_AUTORESIZE,
                     'center': wxALIGN_CENTRE|wxST_NO_AUTORESIZE,
                     'right' : wxALIGN_RIGHT|wxST_NO_AUTORESIZE,
                     }

  def _createWidget(self, event, spacer):
    object = event.object
    style = self.alignmentStyle[object.alignment]
    newWidget = wxStaticText(event.container, -1, str(object.text),
                             wxPoint(object.x*event.widgetWidth,
                                     (object.y+spacer+(spacer*object._gap))*event.widgetHeight),
                             wxSize(self.itemWidth,self.itemHeight), style)
    return newWidget

#
# UIBox
#
# Widget set specific function that creates a single instance of a border box
#
class UIBox(UIHelper, UIWidget):
  def _createWidget(self, event, spacer):
    object = event.object

    # this panel makes the GFBox be a container
    newPanel = wxPanel(event.container,-1,
                       wxPoint(object.x*event.widgetWidth+(event.widgetWidth/2),
                               (object.y+spacer+(spacer*object._gap))*event.widgetHeight+(event.widgetHeight/2)),
                       wxSize((object.width-1)*event.widgetWidth,
                              (object.height-1)*event.widgetHeight))
    newWidget = wxStaticBox(newPanel,-1,str(object.label),
                            wxPoint(0, 0),
                             wxSize((object.width-1)*event.widgetWidth,
                                    (object.height-1)*event.widgetHeight))
    return newPanel

#
# UIScrollBar
#
# Widget set specific function that creates a single instance of a scroll bar used
# to navigate multiple records
#
class UIScrollBar(UIHelper, UIWidget):
  # this one maps wx's scroll to gnue's scroll.
  wxToScroll= {
    10302: 0, # line up
    10304: 1, # page up
    10306: 2, # free scroll (handle move)
    10305: 3, # page down
    10303: 4  # line down
  }

  def _createWidget(self, event, spacer):
    object = event.object
    self._pos = 0
    self._page= object.page
    self._widgetHeight = event.widgetHeight
    self._widget= wxScrollBar(event.container,-1,
                           wxPoint(object.x*event.widgetWidth,(object.y+spacer)*event.widgetHeight),
                           wxSize(object.width*event.widgetWidth,object.height*event.widgetHeight),
                           wxSB_VERTICAL)
    # tie
    if event.initialize:
      # raise Exception
      # also, hook on our block's events
      # how? I need a eventAware thing!
      # registerEventListeners ()
      # or find the block and cross register.
      block= object.findParentOfType ('GFBlock')
      block.registerScrollBar (self)
      self._block = block
      # envetHandler is the gfuserinterface's dispatchEvent...
      self._eventHandler= event.eventHandler
      self.update (0, 0)
      EVT_SCROLL (self._widget, self.scroll)

    return self._widget

  def scroll (self, event):
    try:
      currPos = event.GetPosition()
      if currPos == self._pos:
        action = GFEvent('requestSCROLL', (self.wxToScroll[event.GetEventType()], self._page, self._block, self._pos, 0, float(self._page)/float(self._block.getRecordCount())*float(self._widgetHeight)))
      else:
        action = GFEvent('requestSCROLL', (self.wxToScroll[event.GetEventType()], self._page, self._block, self._pos, currPos, float(self._page)/float(self._block.getRecordCount())*float(self._widgetHeight)))
      self._eventHandler(action)
    except KeyError, e:
      # ignore; it's not a event we want to handle
      # e.g., click on the sb's handle
      pass

  # int position, int thumbSize, int range, int pageSize, const bool refresh = TRUE
  def update (self, pos, range):
    self._pos = pos
    try:
      tumbSize = float(self._page)/float(range)*float(self._widgetHeight)
      if tumbSize < 1:
        tumbSize = 1
    except ZeroDivisionError:
      tumbSize = 0
      pass
    #print pos, tumbSize, range, self._page
    self._widget.SetScrollbar (pos, tumbSize, range, self._page)


#
# UIButton
#
# Widget set specific function that creates a single instance of a button
#
class UIButton(UIHelper, UIWidget):
  def _createWidget(self, event, spacer):
    object = event.object
    newWidget = wxButton(event.container,-1,str(object.label),
                         wxPoint(object.x*event.widgetWidth,(object.y+spacer)*event.widgetHeight),
                         wxSize(object.width*event.widgetWidth,object.height*event.widgetHeight)
                         )
    if event.initialize:
      self._eventHandler = event.eventHandler
      EVT_BUTTON(newWidget,newWidget.GetId(),self.buttonHandler)
      _setDefaultEventHandlers(newWidget, event.eventHandler, event.initialize)

    return newWidget

  def buttonHandler(self,event):
    if event.GetEventType() == wxEVT_COMMAND_BUTTON_CLICKED:
      action = None
      object = _eventObjTowxWindow(event)

      id = object.GetId()
      gfObject     = _IdToGFObj[id]
      action = GFEvent('fireTRIGGER',gfObject.trigger)
      self._eventHandler(action)


#
# UIPage
#
# Widget set specific function that creates a single instance of a page
#
# Note: The event trap for changing notebook pages is attached to the UIForm
#
class UIPage(UIHelper, UIWidget):
  def _createWidget(self, event, spacer):
    object = event.object
    if event.interface._notebook:
      newWidget = wxPanel(event.interface._notebook, -1,wxDefaultPosition,event.interface._notebook.GetSize())
      event.interface._notebook.AddPage(newWidget,object.name)
      newWidget.Show(1)
    else:
      newWidget = wxPanel(event.container, -1,wxDefaultPosition,event.container.GetSize())
      newWidget.Show(0)

    _setDefaultEventHandlers(newWidget, event.eventHandler, event.initialize)

    event.interface._pageList.append(newWidget)
    return newWidget

#
# UIEntry
#
# Widget set specific function that creates a single instance of a data entry widget
#
class UIEntry(UIHelper, UIWidget):
  def _createWidget(self, event, spacer):
    object = event.object

    style = object.style
    defaultPoint =  wxPoint(object.x*event.widgetWidth,
                            (object.y+spacer+(object._gap*spacer))*event.widgetHeight)
    defaultSize  = wxSize(self.itemWidth,self.itemHeight+1)

    if style == 'dropdown':
      if event.initialize:
        choices = object.allowedValues()[1]
      else:
        choices = [""]

      newWidget = wxComboBox(event.container, -1, "",defaultPoint,
                              wxSize(self.itemWidth,self.itemHeight),
                             choices, wxCB_DROPDOWN,) # |wxWANTS_CHARS,)
      newWidget.SetValue("")
      EVT_COMBOBOX(newWidget, newWidget.GetId(), self.comboHandler)

    elif style == 'textlookup':
      newWidget = wxStaticText(event.container, -1, "",defaultPoint,defaultSize,
                               wxST_NO_AUTORESIZE)
      # To make the label consistent with the other entry styles...
      newWidget.SetValue = newWidget.SetLabel
      newWidget.GetValue = newWidget.GetLabel

    elif style == 'label':
      newWidget = wxStaticText(event.container, -1, "",defaultPoint,defaultSize,
                               wxST_NO_AUTORESIZE)
      # To make the label consistent with the other entry styles...
      newWidget.SetValue = newWidget.SetLabel
      newWidget.GetValue = newWidget.GetLabel

    elif style == 'checkbox':
      newWidget = wxCheckBox(event.container, -1, "",defaultPoint,
                             wxSize(2*event.textWidth,event.textHeight),
                             wxNO_BORDER)
      # print "cW: %s" % object.value
      EVT_CHECKBOX (newWidget, newWidget.GetId (), self.checkboxHandler)
      newWidget.SetValue(0)

    else: # Normal text box
      if event.initialize:
        styles = wxTE_PROCESS_TAB
        if object.height > 1:
          styles = styles|wxTE_MULTILINE
        value = ""  ## object.getValue()
      else:
        styles = 0
        value = ""

      newWidget = wxTextCtrl(event.container, -1, value, defaultPoint, defaultSize, styles)
      if object.readonly:
        # set the font color to some shade of gray
        # newWidget.SetForegroundColour (wxColour (255, 127, 127))
        # almost there.
        # newWidget.SetDefaultStyle (wxTextAttr (wxColour (191, 191, 191)))
        pass

    self._eventHandler = event.eventHandler
    _setDefaultEventHandlers(newWidget, event.eventHandler, event.initialize)
    return newWidget

  def comboHandler(self, event):
    action = None
    if event.GetEventType() == wxEVT_COMMAND_COMBOBOX_SELECTED:
      object = _eventObjTowxWindow(event)
      id = object.GetId()

      selection = event.GetSelection()

      gfObject     = _IdToGFObj[id]
      eventdata = [gfObject, event.GetString()]
      action = GFEvent('requestREPLACEVALUE',object=gfObject,
                              index=selection, text=event.GetString())
    self._eventHandler(action)

  def checkboxHandler(self,event):
    action = None
    if event.GetEventType()==wxEVT_COMMAND_CHECKBOX_CLICKED:
      object= _eventObjTowxWindow(event)
      id= object.GetId()
      gfObject= _IdToGFObj[id]

      eventdata = [gfObject, event.GetString()]
      action = GFEvent('requestTOGGLECHKBOX')

    self._eventHandler(action)
      # hack for swig shortcomming
#       eo = event.GetEventObject()
#       object = wxPyTypeCast(eo, 'wxCheckBox')
      # TODO: this doesn't work yet :)
      #for key in self._formToUI.keys():
      #  for count in range(len(self._formToUI[key])):
      #    if self._formToUI[key][count].GetId() == object.GetId():
      #      self.dispatchEvent(GFEvent('requestFOCUS',key))
      #      eventdata = [key, object.GetValue()]
      #      self.dispatchEvent(GFEvent('requestJUMPRECORD',count - key._visibleIndex))
      #      self.dispatchEvent(GFEvent('requestTOGGLECHKBOX',eventdata))
      #      break

      #if action:
      #  self.dispatchEvent(action)



#
# UIForm
#
# Widget set specific function that creates a single instance of a Form widget
#
class UIForm(UIHelper, UIWidget):
  tabStyles = {'left':wxNB_LEFT,
               'right':wxNB_RIGHT,
               'bottom':wxNB_BOTTOM,
               'top':0,
               }

  def _createWidget(self, event, spacer):
    object = event.object

    if object.tabbed:
      tabstyle =self.tabStyles[object.tabbed]
      newWidget = wxNotebook(event.container,-1,wxPoint(10,10),size=event.container.GetSize(),style=tabstyle)
      newWidget.Show(TRUE)
      EVT_NOTEBOOK_PAGE_CHANGING(newWidget, newWidget.GetId(), self.notebookTabHandler)

      event.interface._notebook  = newWidget
    else:
      newWidget = wxPanel(event.interface.mainWindow.panel,-1, wxDefaultPosition)
      event.interface._notebook = None

    if event.initialize:
      EVT_CLOSE(event.interface.mainWindow,event.interface.closeTrap)

    self._eventHandler = event.eventHandler
    return newWidget

  def notebookTabHandler(self,event):
    if event.GetEventType() == wxEVT_COMMAND_NOTEBOOK_PAGE_CHANGING:
      action = None
      global _LOOPTRAP

      if _LOOPTRAP == 0:
        action = GFEvent('requestPAGE',event.GetSelection())

      _LOOPTRAP = not _LOOPTRAP # flip the value

    if action:
      self._eventHandler(action)

#####################################################################
##
## Required Dialog box support
##
#####################################################################

#
# UIAbout
#
# A dialog box providing about info
#
class UIAbout(wxMessageDialog, UIHelper):
  def __init__(self, interface, programVersion, formName, formVersion, author, description):
    text = _("GNUE Forms")+"\n"+      \
    _("  Version : ")+"%s\n"+         \
    _("  Driver  : UIwxpython")+"\n"+ \
    _("-= Form Info =-")+"\n"+        \
    _("  Name   : ")+"%s\n"+          \
    _("  Version: ")+"%s\n"+          \
    _("  Author : ")+"%s\n"+          \
    _("  Description:")+"%s\n"
    wxMessageDialog.__init__(self, interface,
                             text % (programVersion, formName ,formVersion,author,description),
                             _("About"), wxOK | wxICON_INFORMATION)
#
# UImessageBox
#
# A dialog box providing basic message info
#
class UIMessageBox(wxMessageDialog, UIHelper):
  def __init__(self, interface, message, caption):
    message = lineWrap(message,60)
    wxMessageDialog.__init__(self, interface, message, caption, style=wxOK|wxICON_EXCLAMATION)


#
# UIGenericBox
#
# A generic box providing feedback info
#
class UIGenericBox(wxDialog, UIHelper):
  def __init__(self, interface, caption):
    self.interface = interface
    self.caption = caption

  def createGenericBox(self, message, buttonSet):
      self.genericBox = dlg = wxDialog(self.interface, -1, self.caption)
      buttonCount = len(buttonSet)

      dlg.SetAutoLayout(true)

      #bmp = wxImage(images_dir+gConfig('smallPNG'), wxBITMAP_TYPE_PNG).ConvertToBitmap()

      messageField = WrappedStaticText(dlg, -1, str(message), 300, style=wxALIGN_CENTER)

      dlgWidth = messageField.GetSize().GetWidth() + 30

      dlgHeight = messageField.GetSize().GetHeight() + 80

      idx = 0
      buttonSetWidth = 0
      maxButtonHeight = 0
      button = []

      for buttonLabel in buttonSet:
        buttonId = wxNewId()
        button.append(wxButton(dlg,buttonId,_(buttonLabel)))
        EVT_BUTTON(dlg, buttonId, lambda event, s=self, butt=idx+1: s.setResult(butt) )
        buttonSetWidth += button[idx].GetSize().GetWidth()
        if maxButtonHeight < button[idx].GetSize().GetHeight():
          maxButtonHeight = button[idx].GetSize().GetHeight()
        idx += 1

      button[0].SetDefault()
      dlgWidth = max(dlgWidth, buttonSetWidth + 15 * (buttonCount- 1) + 40)

      dlgHeight += maxButtonHeight - 6

      # Create and position the logo
      #wxStaticBitmap(dlg,-1, bmp,
      #               wxPoint((dlgWidth-bmp.GetWidth())/2, 12),
      #               wxSize(bmp.GetWidth(), bmp.GetHeight()))

      # Move the various widgets into position
      messageField.SetPosition(
        wxPoint(dlgWidth/2 - messageField.GetSize().GetWidth()/2,
                30 ))
      idx = 0
      incr = (dlgWidth - 40 - buttonSetWidth)
      if buttonCount > 1:
        incr = incr / (buttonCount - 1)

      if buttonCount > 1:
        currentX = 15
      else:
        currentX = (dlgWidth - buttonSetWidth) / 2

      for buttonLabel in buttonSet:
        button[idx].SetPosition(wxPoint(currentX, dlgHeight - 10 - maxButtonHeight))
        currentX += button[idx].GetSize().GetWidth() + incr
        idx += 1

      dlg.SetSize(wxSize(dlgWidth, dlgHeight))

      dlg.Refresh()
      dlg.Fit()
      dlg.Raise()
      dlg.CenterOnScreen()

      # If user cancels, this will be set to 0
      self._result = 0

      return dlg

  def getResult(self):
    return self._result

  def setResult(self, value):
    self._result = value
    self.genericBox.EndModal(1)
    self.genericBox.Destroy()


#####################################################################
##
## Keymapper Support
##
#####################################################################
from gnue.forms.GFKeyMapper import vk

# Translate from wx keystrokes to our virtual keystrokes
wxKeyTranslations = {
   vk.F1     : WXK_F1,        vk.F2        : WXK_F2,
   vk.F3     : WXK_F3,        vk.F4        : WXK_F4,
   vk.F5     : WXK_F5,        vk.F6        : WXK_F6,
   vk.F7     : WXK_F7,        vk.F8        : WXK_F8,
   vk.F9     : WXK_F9,        vk.F10       : WXK_F10,
   vk.F11    : WXK_F11,       vk.F12       : WXK_F12,
   vk.INSERT : WXK_INSERT,    vk.DELETE    : WXK_DELETE,
   vk.HOME   : WXK_HOME,      vk.END       : WXK_END,
   vk.PAGEUP : WXK_PRIOR,     vk.PAGEDOWN  : WXK_NEXT,
   vk.UP     : WXK_UP,        vk.DOWN      : WXK_DOWN,
   vk.LEFT   : WXK_LEFT,      vk.RIGHT     : WXK_RIGHT,
   vk.TAB    : WXK_TAB,
   vk.ENTER  : WXK_RETURN,    vk.BACKSPACE : WXK_BACK }

GFKeyMapper.KeyMapper.setUIKeyMap(wxKeyTranslations)

#####################################################################
##
## Basic Event Processing
##
#####################################################################

def _setDefaultEventHandlers(newWidget, eventHandler, initialize):
  if initialize:
    # TODO: this should use one instance
    newWidget.PushEventHandler(mouseEvtHandler(eventHandler))
    newWidget.PushEventHandler(keyboardEvtHandler(eventHandler))

class uiBaseEvtHandler(wxEvtHandler):
  def __init__(self,eventList,eventProcessor):
    wxEvtHandler.__init__(self)
    self._eventProcessor = eventProcessor
    for eventType in eventList:
      self.Connect(-1, -1, eventType,self._processEvent)

  def _processEvent(self,event):
    GDebug.printMesg(0,"uiBaseEvtHandler _processEvent was called?!?")

class mouseEvtHandler(uiBaseEvtHandler):
  def __init__(self,eventProcessor):
    eventList = [wxEVT_LEFT_DOWN]
    uiBaseEvtHandler.__init__(self, eventList, eventProcessor)

  def _processEvent(self,event):
    global _charWidth
    object = _eventObjTowxWindow(event)

    # compute the location of the character in the widget
    x,y = event.GetPositionTuple()
    cursorPosition = x/_charWidth

    # Move to proper object
    id = object.GetId()
    gfObject     = _IdToGFObj[id]
    screenWidget = _IdToWxObj[id]
    count        = _IdToUIObj[id].widgets.index(screenWidget)
    self._eventProcessor(GFEvent('requestFOCUS',gfObject))
    self._eventProcessor(GFEvent('requestJUMPRECORD',count - gfObject._visibleIndex))
    self._eventProcessor(GFEvent('requestCURSORMOVE',position=cursorPosition))
#    if gfObject._type == 'GFButton':
#      self._eventProcessor(GFEvent('fireTRIGGER',gfObject.trigger))

    event.Skip()

#
# keyboardEvtHandler
#
# Note: wxEVT_KEY_DOWN event seems to only return uppercase chars so
#       i don't trap it.
#

class keyboardEvtHandler(uiBaseEvtHandler):
  def __init__(self, eventProcessor):
    eventList = [wxEVT_CHAR]
    uiBaseEvtHandler.__init__(self, eventList, eventProcessor)

  def _processEvent(self,event):
    action = None
    # hack for swig shortcomming needed by wxPython 2.2.x (2.3 does not need this)
    # compute the location of the character in the widget
    object = _IdToGFObj[_eventObjTowxWindow(event).GetId()]

    keycode = event.KeyCode()

    #
    # Sigh... a hack for using <enter> in multiline entries
    #
    if  keycode == 13 and \
        not event.ShiftDown() and \
        not event.ControlDown() and \
        not event.AltDown() and \
        int (gConfig('enterIsNewLine')) and \
        object.height > 1 and object._type != 'GFButton':

      command = 'NEWLINE'

    if keycode==32 and object._type=='GFEntry' and object.style=='checkbox':
      #
      command= 'TOGGLECHKBOX'

    else:


      #
      # Get the event to process from the KeyMapper
      #
      command = GFKeyMapper.KeyMapper.getEvent(
        keycode,
        event.ShiftDown(),
        event.ControlDown(),
        event.AltDown())

    if command == 'JUMPRECORD':
      global _PROMPTFORRECORD
      action = _PROMPTFORRECORD()

    elif command == 'NEWLINE':
      action = GFEvent('requestKEYPRESS', '\n',
                     text='\n',
                     code=10)

    elif command:
      action = GFEvent('request%s' % command)

    elif object._type == 'GFButton':
      action = GFEvent('fireTRIGGER',object.trigger)

    else:
      try:
        char = chr(keycode)
        if char in string.printable or char == "\n" or \
           128 <= keycode <= 255:
          action = GFEvent('requestKEYPRESS', char,
                         text=char,
                         code=keycode)
      except ValueError:
        pass #event.Skip()

    if action:
      self._eventProcessor(action)

#####################################################################
##
## Login Support
##
#####################################################################

#
# UILoginHandler
#
class UILoginHandler(GLoginHandler.LoginHandler):
  def __init__(self):
    self._wxapp = getWxApp()
    self.dlg = None

  def getLogin(self, loginData, errortext=None):#"Invalid username/password"):
    if 1: # TODO: This is left here to prevent me from having to unindent this code
          # TODO: Since the UI stuff is currently being gutted I'm not wasting the time
          # TODO: to do the right thing
      if len(loginData[1]):
        loginMesg = _('Login required for\n"%s"') % (loginData[1])
      else:
        loginMesg = _('Login required for %s') % (loginData[0])

      self.dlg = wxDialog(NULL, -1, _("GNU Enterprise: Login to %s") % loginData[0])
      self.dlg.SetAutoLayout(true)

      bmp = wxImage(images_dir+gConfig('smallPNG'), wxBITMAP_TYPE_PNG).ConvertToBitmap()

      messageField = WrappedStaticText(self.dlg, -1, str(loginMesg), 300, style=wxALIGN_CENTER)

      self.textctrlList = []
      labelList = []

      dlgWidth = max(bmp.GetWidth(), messageField.GetSize().GetWidth()+20)

      dlgHeight = bmp.GetHeight() + messageField.GetSize().GetHeight() + 80

      xSpacing = 0
      ySpacing = 0
      fieldLabelWidth = 0

      for prompt in loginData[2]:
        s = wxStaticText(self.dlg, -1, '%s:' % prompt[1])
        labelList.append(s)
        if prompt[2]:
          t = wxTextCtrl(self.dlg, -1,"",wxPoint(1, 1), wxSize(150, 20),
                         style=wxTE_PASSWORD|wxTE_PROCESS_ENTER)
        else:
          t = wxTextCtrl(self.dlg, -1,"",wxPoint(1, 1), wxSize(150, 20),
                         style=wxTE_PROCESS_ENTER)

        myID = len(self.textctrlList)
        self.textctrlList.append(t)
        EVT_CHAR(t, LoginFieldHandler(self, myID).loginFieldEventTrap)

        fieldLabelWidth = max(fieldLabelWidth,
                  s.GetSize().GetWidth() + t.GetSize().GetWidth() + 10)

        dlgWidth = max(dlgWidth, \
                       s.GetSize().GetWidth() + t.GetSize().GetWidth() + 20)

        xSpacing = max(xSpacing, s.GetSize().GetWidth())
        ySpacing = max(ySpacing, s.GetSize().GetHeight())
        ySpacing = max(ySpacing, t.GetSize().GetHeight())


      loginId = wxNewId()
      cancelId = wxNewId()

      loginButton = wxButton(self.dlg,loginId,_('Login'))
      cancelButton = wxButton(self.dlg,cancelId,_('Cancel'))

      EVT_BUTTON(self.dlg, loginId, self.loginButtonEventTrap)
      EVT_BUTTON(self.dlg, cancelId, self.loginCancelEventTrap)
      dlgWidth = max(dlgWidth, loginButton.GetSize().GetWidth() +
                            cancelButton.GetSize().GetWidth() + 6) + 20

      dlgHeight += max(loginButton.GetSize().GetHeight(),
                       cancelButton.GetSize().GetHeight()) - 6

      if errortext:
        errorField = WrappedStaticText(self.dlg, -1, str(errortext), 300,
                                       style=wxALIGN_CENTER)
        errorField.SetForegroundColour(wxColour(223,0,0))

        dlgWidth = max(dlgWidth, errorField.GetSize().width+10)
        dlgHeight += errorField.GetSize().height + 6


      firstY = bmp.GetHeight() + messageField.GetSize().GetHeight() + 50
      lastY = firstY
      xSpacing += 10  # Add whitespace between widgets
      ySpacing += 6  # Add whitespace between widgets
      xPos = dlgWidth/2 - fieldLabelWidth/2


      # Move the fields and labels into position
      for i in range(0, len(self.textctrlList)):
        dlgHeight = dlgHeight + ySpacing
        labelList[i].SetPosition(wxPoint(xPos, lastY))
        self.textctrlList[i].SetPosition(wxPoint(xPos + xSpacing, lastY))
        lastY = lastY + ySpacing

      if errortext:
        errorField.SetPosition(
          wxPoint(dlgWidth/2 - errorField.GetSize().width/2,
                lastY+3))

      # Set the focus to the first text entry field
      self.textctrlList[0].SetFocus()

      # Create and position the logo
      wxStaticBitmap(self.dlg,-1, bmp,
                     wxPoint((dlgWidth-bmp.GetWidth())/2, 12),
                     wxSize(bmp.GetWidth(), bmp.GetHeight()))

      # Move the various widgets into position
      messageField.SetPosition(
        wxPoint(dlgWidth/2 - messageField.GetSize().GetWidth()/2,
                30 + bmp.GetHeight()))

      cancelButton.SetPosition(
        wxPoint(dlgWidth - 10 - cancelButton.GetSize().GetWidth(),
                dlgHeight - 10 - max(loginButton.GetSize().GetHeight(),
                                            cancelButton.GetSize().GetHeight())))
      loginButton.SetPosition(
        wxPoint(dlgWidth - 16 - cancelButton.GetSize().GetWidth() - \
                loginButton.GetSize().GetWidth(),
                dlgHeight - 10 - max(loginButton.GetSize().GetHeight(),
                                            cancelButton.GetSize().GetHeight())))

      self.loginButton = loginButton

      self.dlg.SetSize(wxSize(dlgWidth, dlgHeight))

      self.dlg.Refresh()
      self.dlg.Fit()
      self.dlg.Raise()
      self.dlg.CenterOnScreen()

    # If user cancels, this will be set to 0
    self._completed = 0
    self.dlg.ShowModal()

    if not self._completed:
      raise GLoginHandler.UserCanceledLogin

    rv = {}
    for i in range(0, len(loginData[2])):
      rv[loginData[2][i][0]] = self.textctrlList[i].GetValue()

    return rv

  #
  # Login is completed, for whatever reason
  #
  def loginCompleted(self, successful):
    self._completed = successful
    self.dlg.EndModal(1)

  #
  # Called when user clicks "login"
  #
  def loginButtonEventTrap(self, event):
    self.loginCompleted(1)

  #
  # Called when user clicks "cancel"
  #
  def loginCancelEventTrap(self, event):
    self.loginCompleted(0)

  #
  # TODO: This is a hack required because windows
  # TODO: seems to have issues with wxWindows dialogs
  #
  def destroyLoginDialog(self):
    self.dlg.Destroy()

#
# LoginFieldHandler
#
# Used by the login handler
# enables the user to press return and have it jump to the next box
#
class LoginFieldHandler:
  def __init__(self, app, seq):
    self.app = app
    self.seq = seq

  def loginFieldEventTrap(self, event):
     if event.KeyCode() in (WXK_RETURN, WXK_TAB):
       if self.seq < len(self.app.textctrlList) - 1:
         self.app.textctrlList[self.seq+1].SetFocus()
       else:
         if event.KeyCode() == WXK_TAB:
           if event.ShiftDown():
             self.app.textctrlList[self.seq-1].SetFocus()
           else:
             self.app.loginButton.SetFocus()
         else:
           self.app.loginCompleted(1)
     else:
      event.Skip()



class WrappedStaticText(wxStaticText):
  def __init__(self, parent, id, label, width, *args, **params):
    wxStaticText.__init__(self, parent, id, "bah!", *args, **params)


    textSoFar = ""
    thisLine = ""
    for part in string.split(label,'\n'):
      for word in string.split(part):
        self.SetLabel(thisLine + word)
        if self.GetSize().width > width:
          textSoFar += thisLine + " \n"
          thisLine = word + " "
        else:
          thisLine += word + " "

      textSoFar += thisLine + " \n"
      thisLine = ""

    if len(textSoFar):
      self.SetLabel(string.replace(textSoFar,' \n','\n')[:-1])
    else:
      self.SetLabel("")





# Needed by Designer
WIDGETS = {'GFLabel'     : UILabel,
           'GFBox'       : UIBox,
           'GFPage'      : UIPage,
           'GFForm'      : UIForm,
           'GFEntry'     : UIEntry,
           'GFButton'    : UIButton,
           'GFScrollBar' : UIScrollBar,
           'GFMainToolBar': UIMainToolBar,
           'GFMainMenuBar': UIMainMenuBar,
          }


# Shortcut
images_dir = GConfig.getInstalledBase('forms_images','common_images') + '/'


