CREATE TABLE _product_measure_data (
    product int,
    product_measure_type int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    measure numeric(12,4),
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _product_file_data (
    product int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    file int,
    _table bigint NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _alien_entity_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    entity int,
    PRIMARY KEY (id)
);

CREATE TABLE barter (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _product_measure_type_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    name varchar(80),
    PRIMARY KEY (id)
);

CREATE TABLE movement_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE own_pos (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE erm_task (
    erm int,
    task int,
    id serial NOT NULL,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE entity (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE persona_tax_profile (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE task_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE credit_card_movement (
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    movement bigint,
    id bigserial NOT NULL,
    entry bigint,
    plan bigint,
    PRIMARY KEY (id)
);

CREATE TABLE _category_entity_type_data (
    category_type int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    entity_type int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE alien_document_proceeding (
    amount numeric(12,4),
    proceeding bigint,
    entry bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    id bigserial NOT NULL,
    alien_document bigint,
    PRIMARY KEY (id)
);

CREATE TABLE _bank_account_data (
    credit numeric(12,4),
    credit_timestamp timestamp,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    overdraft numeric(12,4),
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    bank_branch int,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _account_plan_data (
    balance numeric(12,4),
    code varchar(40) NOT NULL,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(40) NOT NULL,
    parent int,
    _state int DEFAULT -1 NOT NULL,
    chargeable boolean,
    path varchar(255),
    _table int NOT NULL,
    id bigserial NOT NULL,
    _end_t timestamp,
    PRIMARY KEY (id)
);
CREATE UNIQUE INDEX account_plan_code_unique_idx ON _account_plan_data(code) WHERE _end_t is null;

CREATE TABLE _employee_comm_client_data (
    commission int,
    client int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    employee int,
    _table bigint NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE product_classification (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE admin_period (
    end_time timestamp,
    closed boolean DEFAULT 'f',
    admin_period_type int,
    start_time timestamp,
    id serial NOT NULL,
    name varchar(40) NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE bank (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _product_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    commission int,
    country int,
    salable boolean,
    _state int DEFAULT -1 NOT NULL,
    account_plan int,
    product_classification int,
    stockable boolean,
    is_perishable boolean,
    canonical_name varchar(80),
    location int,
    _table int NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE document4document (
    own_document bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    entry bigint,
    id bigserial NOT NULL,
    alien_document bigint,
    PRIMARY KEY (id)
);

CREATE TABLE employee (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE logical_structure (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE address_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _employee_comm_prod_data (
    commission int,
    product int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    employee int,
    _table bigint NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _entity_relation_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    entity_from int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    entity_to int,
    PRIMARY KEY (id)
);

CREATE TABLE product_store (
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE own_item_tax (
    amount numeric(12,4),
    entry bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    tax bigint,
    own_item bigint,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE cheque_movement (
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    movement bigint,
    id bigserial NOT NULL,
    entry bigint,
    cheque bigint,
    PRIMARY KEY (id)
);

CREATE TABLE _document_type_data (
    code char(4),
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    description text,
    name varchar(80),
    id bigserial NOT NULL,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    account_plan int,
    _end_t timestamp,
    sign boolean DEFAULT NULL,
    PRIMARY KEY (id)
);
CREATE UNIQUE INDEX document_type_code_unique_idx ON _document_type_data(code) WHERE _end_t is null;

CREATE TABLE _product_name_data (
    product int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    product_name_type int,
    maximum_length int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    _end_t timestamp,
    name varchar(80),
    PRIMARY KEY (id)
);

CREATE TABLE _tax_formula_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    graft text,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    from_profile int,
    to_profile int,
    PRIMARY KEY (id)
);

CREATE TABLE entity_type_uid_conf (
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    id serial NOT NULL,
    entity_uid_type bigint,
    entity_type bigint,
    PRIMARY KEY (id)
);

CREATE TABLE alien_product_item (
    product bigint,
    account_plan bigint,
    entry bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    qty numeric(12,4),
    id bigserial NOT NULL,
    unit_price numeric(12,4),
    alien_item bigint,
    PRIMARY KEY (id)
);

CREATE TABLE category_entity_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE ticket (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE document_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _tax_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(80),
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    account_plan int,
    _end_t timestamp,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _credit_card_data (
    _table int NOT NULL,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(80),
    accreditation_account int,
    ownid serial,
    accreditation_date timestamp,
    _state int DEFAULT -1 NOT NULL,
    account_plan int,
    _end_t timestamp,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE bank_account (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE task_recipient (
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    task int,
    id serial NOT NULL,
    to_make_to bigint,
    PRIMARY KEY (id)
);

CREATE TABLE product_code_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _entity_attribute_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    category_attribute int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    entity int,
    PRIMARY KEY (id)
);

CREATE TABLE alien_descriptive_item (
    amount numeric(12,4),
    description text,
    entry bigint,
    id bigserial NOT NULL,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    account_plan bigint,
    alien_item bigint,
    PRIMARY KEY (id)
);

CREATE TABLE third_party_cheque_tax (
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    third_party_cheque bigint,
    tax bigint,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE tax_formula (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE relation_types (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE own_doc4own_doc (
    document_to bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    id serial NOT NULL,
    relation_type int,
    document_from bigint,
    PRIMARY KEY (id)
);

CREATE TABLE _entity_type_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(80),
    parent int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    _end_t timestamp,
    PRIMARY KEY (id)
);

CREATE TABLE own_document_tax (
    amount numeric(12,4),
    taxed_amount numeric(12,4),
    own_document bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    tax bigint,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _relation_types_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    entity_relation int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    relation_type int,
    PRIMARY KEY (id)
);

CREATE TABLE task_priority (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _tax_profile_data (
    profile varchar(80),
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    tax int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _own_pos_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    group_documents boolean DEFAULT 'f',
    _state int DEFAULT -1 NOT NULL,
    _table int NOT NULL,
    own_entity int,
    id bigserial NOT NULL,
    stock_by_invoice boolean,
    number char(4),
    PRIMARY KEY (id)
);
CREATE UNIQUE INDEX own_pos_number_unique_idx ON _own_pos_data(number) WHERE _end_t is null;

CREATE TABLE _own_branch_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    own_entity int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    number char(4),
    PRIMARY KEY (id)
);
CREATE UNIQUE INDEX own_branch_number_unique_idx ON _own_branch_data(number) WHERE _end_t is null;

CREATE TABLE _bank_data (
    persona int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(80),
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    _end_t timestamp,
    PRIMARY KEY (id)
);

CREATE TABLE physical_structure (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE bank_branch (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE own_branch (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _obligation_type_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    description text,
    id bigserial NOT NULL,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    account_plan int,
    _end_t timestamp,
    name varchar(80),
    PRIMARY KEY (id)
);

CREATE TABLE product_code (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE employee_comm_prod_client (
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE product_provider (
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE proceeding (
    responsable bigint,
    branch bigint,
    entry bigint,
    date timestamp,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _product_code_data (
    code varchar(30),
    product int,
    product_code_type int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);
CREATE UNIQUE INDEX product_code_unique_idx ON _product_code_data(code, product_code_type) WHERE _end_t is null;

CREATE TABLE _logical2physical_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    logical int,
    physical int,
    PRIMARY KEY (id)
);

CREATE TABLE entity_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE entity_uid (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _price_data (
    currency_type int,
    product int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    excluded boolean,
    price_type int,
    _table bigint NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    unit_price numeric(12,4),
    PRIMARY KEY (id)
);

CREATE TABLE _employee_data (
    commission int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    own_entity int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE category_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE alien_document_tax (
    amount numeric(12,4),
    taxed_amount numeric(12,4),
    entry bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    tax bigint,
    id bigserial NOT NULL,
    alien_document bigint,
    PRIMARY KEY (id)
);

CREATE TABLE product_structure (
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE provider (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE own_item (
    discount numeric(12,4),
    entry bigint,
    bonus numeric(12,4),
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    own_document bigint,
    account_plan bigint,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _role_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    name varchar(80),
    PRIMARY KEY (id)
);

CREATE TABLE movement (
    amount numeric(12,4),
    proceeding bigint,
    parent bigint,
    movement_type bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE third_party_cheque (
    bank_branch bigint,
    entry bigint,
    credit_date timestamp,
    emmited_by varchar(80),
    third_party_cheque_delay int,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    id bigserial NOT NULL,
    account_plan bigint,
    PRIMARY KEY (id)
);

CREATE TABLE _alien_storehouse_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    alien_entity int,
    PRIMARY KEY (id)
);

CREATE TABLE entity_attribute (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _persona_tax_profile_data (
    tax_profile int,
    persona int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE own_document (
    due_date timestamp,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    own_document_type bigint,
    related_to varchar(255),
    number char(30),
    is_void boolean,
    salesperson bigint,
    address varchar(255),
    date timestamp,
    authorized_amount numeric(12,4),
    recipient bigint,
    id bigserial NOT NULL,
    transport bigint,
    end_timestamp timestamp,
    tax_total numeric(12,4),
    start_timestamp timestamp,
    own_entity bigint,
    bundles int,
    net_total numeric(12,4),
    observations text,
    discount_percent numeric(12,4),
    authorized_by bigint,
    entry bigint,
    guide_number char(30),
    PRIMARY KEY (id)
);
CREATE UNIQUE INDEX own_document_number_unique_idx ON own_document(number, own_document_type, own_entity);

CREATE TABLE _employee_comm_prod_client_data (
    commission int,
    product int,
    client int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    employee int,
    _table bigint NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE own_product_stock_item (
    own_item bigint,
    observations text,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    difference numeric(12,4),
    product bigint,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE erm_log (
    text text,
    id serial NOT NULL,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    erm int,
    PRIMARY KEY (id)
);

CREATE TABLE address (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _task_priority_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    name varchar(80),
    PRIMARY KEY (id)
);

CREATE TABLE obligation (
    obligation_type bigint,
    amount numeric(12,4),
    date timestamp,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    account_plan bigint,
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE own_storehouse (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE alien_document (
    due_date timestamp,
    start_timestamp timestamp,
    sender bigint,
    alien_document_type bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    tax_total numeric(12,4),
    own_entity bigint,
    bundles int,
    number char(30),
    received_by bigint,
    is_void boolean,
    net_total numeric(12,4),
    salesperson bigint,
    observations text,
    discount_percent numeric(12,4),
    date timestamp,
    entry bigint,
    guide_number char(30),
    id bigserial NOT NULL,
    transport bigint,
    end_timestamp timestamp,
    PRIMARY KEY (id)
);
CREATE UNIQUE INDEX alien_document_number_unique_idx ON alien_document(number, sender, alien_document_type);

CREATE TABLE _product_tax_data (
    product int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    entry bigint,
    _state int DEFAULT -1 NOT NULL,
    _table bigint NOT NULL,
    tax int,
    amount numeric(12,4),
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE alien_document_admin_period (
    admin_period int,
    id bigserial NOT NULL,
    alien_document bigint,
    PRIMARY KEY (id)
);

CREATE TABLE price_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE country (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _bank_branch_data (
    persona int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _address_data (
    comment varchar(80),
    _table int NOT NULL,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    address varchar(255),
    address_type int,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    entity int,
    PRIMARY KEY (id)
);

CREATE TABLE credit_card_plan (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE product_measure_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _relation_type_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    name varchar(255),
    PRIMARY KEY (id)
);

CREATE TABLE _physical_structure_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(80),
    parent int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    _end_t timestamp,
    PRIMARY KEY (id)
);

CREATE TABLE _logical_structure_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(80),
    parent int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    _end_t timestamp,
    PRIMARY KEY (id)
);

CREATE TABLE task_status (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE alien_storehouse (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _currency_type_data (
    rate numeric(10,2),
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(50),
    format varchar(50),
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    account_plan int,
    _end_t timestamp,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);
CREATE UNIQUE INDEX currency_type_name_unique_idx ON _currency_type_data(name) WHERE _end_t is null;

CREATE TABLE _country_data (
    phone integer,
    iso3 char(3),
    iso2 char(2),
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(80),
    _table int NOT NULL,
    un integer,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    _end_t timestamp,
    PRIMARY KEY (id)
);

CREATE TABLE employee_role (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _fisco_data (
    persona int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE ticket_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE employee_comm_prod (
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE cheque (
    chequebook int,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    date timestamp,
    entry bigint,
    account_plan bigint,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE own_pos_document (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE currency_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE bank_account_movement (
    origin bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    bank_account_movement_delay int,
    destination bigint,
    id bigserial NOT NULL,
    movement bigint,
    PRIMARY KEY (id)
);

CREATE TABLE _task_status_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    name varchar(80),
    PRIMARY KEY (id)
);

CREATE TABLE _file_data (
    mimetype char(80),
    application char(15),
    _table int NOT NULL,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    checksum char(32),
    filename varchar(255),
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    size int,
    PRIMARY KEY (id)
);

CREATE TABLE alien_item (
    discount numeric(12,4),
    alien_document bigint,
    entry bigint,
    bonus numeric(12,4),
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    account_plan bigint,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE alien_item_tax (
    amount numeric(12,4),
    entry bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    tax bigint,
    id bigserial NOT NULL,
    alien_item bigint,
    PRIMARY KEY (id)
);

CREATE TABLE _credit_limit_data (
    credit_limit numeric(12,4),
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    movement_type int,
    _table int NOT NULL,
    time_stamp timestamp,
    balance numeric(12,4),
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    entity int,
    PRIMARY KEY (id)
);

CREATE TABLE account_plan (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _ticket_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    ticket_type int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    account_plan int,
    PRIMARY KEY (id)
);

CREATE TABLE commission (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE sys_account2account_plan (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE document_relation_type (
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    id serial NOT NULL,
    name varchar(80),
    PRIMARY KEY (id)
);

CREATE TABLE admin_period_type (
    id serial NOT NULL,
    name varchar(40) NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _entity_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(255),
    is_active boolean,
    start_time timestamp,
    _state int DEFAULT -1 NOT NULL,
    account_plan int,
    anniversary date,
    observations text,
    _table int NOT NULL,
    id bigserial NOT NULL,
    _end_t timestamp,
    PRIMARY KEY (id)
);

CREATE TABLE currency_movement (
    currency_type bigint,
    entry bigint,
    id bigserial NOT NULL,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    movement bigint,
    PRIMARY KEY (id)
);

CREATE TABLE _own_entity_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    entity int,
    PRIMARY KEY (id)
);

CREATE TABLE comm_prod_client (
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE cheque_tax (
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    tax bigint,
    id bigserial NOT NULL,
    cheque bigint,
    PRIMARY KEY (id)
);

CREATE TABLE _own_storehouse_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    own_entity int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    stock_by_invoice boolean,
    PRIMARY KEY (id)
);

CREATE TABLE barter_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _product_store_data (
    hiwater numeric(12,4),
    product int,
    lowater numeric(12,4),
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    storehouse int,
    _table bigint NOT NULL,
    level numeric(12,4),
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE barter_movement (
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    entry bigint,
    id bigserial NOT NULL,
    barter bigint,
    movement bigint,
    PRIMARY KEY (id)
);

CREATE TABLE _ticket_type_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    description text,
    id bigserial NOT NULL,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    account_plan int,
    _end_t timestamp,
    name varchar(80),
    PRIMARY KEY (id)
);

CREATE TABLE obligation_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE tax_profile (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE credit_limit (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _provider_data (
    persona int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE employee_comm_client (
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE own_product_item (
    own_item bigint,
    account_plan bigint,
    entry bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    price bigint,
    id bigserial NOT NULL,
    unit_price numeric(12,4),
    qty numeric(12,4),
    PRIMARY KEY (id)
);

CREATE TABLE logical2physical (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE organization (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _barter_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(80),
    _state int DEFAULT -1 NOT NULL,
    _table int NOT NULL,
    barter_type int,
    account_plan int,
    _end_t timestamp,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE entity_relation (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _price_type_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(80),
    parent int,
    is_active boolean,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    branch int,
    formula text,
    end_date timestamp,
    _table int NOT NULL,
    start_date timestamp,
    _end_t timestamp,
    PRIMARY KEY (id)
);

CREATE TABLE obligation_movement (
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    obligation int,
    entry bigint,
    id bigserial NOT NULL,
    movement bigint,
    PRIMARY KEY (id)
);

CREATE TABLE _organization_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    own_entity int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE own_document_admin_period (
    own_document bigint,
    id bigserial NOT NULL,
    admin_period int,
    PRIMARY KEY (id)
);

CREATE TABLE category_attribute (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE erm (
    erm_type bigint,
    description varchar(80),
    time timestamp,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    made_by bigint,
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE entity_types (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _entity_uid_type_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    name varchar(255),
    PRIMARY KEY (id)
);

CREATE TABLE _comm_prod_client_data (
    commission int,
    product int,
    client int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table bigint NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE ticket_movement (
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    ticket bigint,
    entry bigint,
    id bigserial NOT NULL,
    movement bigint,
    PRIMARY KEY (id)
);

CREATE TABLE _product_code_type_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    name varchar(80),
    PRIMARY KEY (id)
);

CREATE TABLE erm_file (
    erm int,
    id serial NOT NULL,
    file bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _task_type_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    name varchar(80),
    PRIMARY KEY (id)
);

CREATE TABLE third_party_cheque_movement (
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    third_party_cheque bigint,
    entry bigint,
    id bigserial NOT NULL,
    movement bigint,
    PRIMARY KEY (id)
);

CREATE TABLE tax (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE erm_entity (
    erm int,
    id serial NOT NULL,
    made_to bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _contact_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    alien_entity int,
    PRIMARY KEY (id)
);

CREATE TABLE _entity_types_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    entity_type int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    entity int,
    PRIMARY KEY (id)
);

CREATE TABLE _own_pos_document_data (
    document_type int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    print_command varchar(255),
    _state int DEFAULT -1 NOT NULL,
    _table int NOT NULL,
    own_pos int,
    id bigserial NOT NULL,
    number char(30),
    PRIMARY KEY (id)
);

CREATE TABLE __papo__ (
    date timestamp DEFAULT now(),
    value varchar(255),
    key varchar(255),
    PRIMARY KEY (key)
);

CREATE TABLE file (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _product_name_type_data (
    is_required boolean,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(80),
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    _end_t timestamp,
    PRIMARY KEY (id)
);

CREATE TABLE own_entity (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _credit_card_plan_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(80),
    financial_charge numeric(10,2),
    _state int DEFAULT -1 NOT NULL,
    account_plan int,
    credit_card int,
    payments int,
    _table int NOT NULL,
    id bigserial NOT NULL,
    _end_t timestamp,
    surcharge numeric(10,2),
    PRIMARY KEY (id)
);

CREATE TABLE _product_structure_data (
    product int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _state int DEFAULT -1 NOT NULL,
    _table bigint NOT NULL,
    component int,
    id bigserial NOT NULL,
    qty numeric(12,4),
    PRIMARY KEY (id)
);

CREATE TABLE alien_doc4alien_doc (
    document_to bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    id serial NOT NULL,
    relation_type int,
    document_from bigint,
    PRIMARY KEY (id)
);

CREATE TABLE alien_entity (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _employee_role_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    employee int,
    role int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE __state__ (
    state varchar(40) NOT NULL,
    id int NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _erm_type_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    name varchar(80),
    PRIMARY KEY (id)
);

CREATE TABLE _sys_account2account_plan_data (
    account int,
    code varchar(40),
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(40) NOT NULL,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    _end_t timestamp,
    PRIMARY KEY (id)
);

CREATE TABLE client (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE role (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE product_name_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _address_type_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    name varchar(80),
    PRIMARY KEY (id)
);

CREATE TABLE product_name (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _persona_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    alien_entity int,
    PRIMARY KEY (id)
);

CREATE TABLE product (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE persona (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _entity_currency_type_data (
    currency_type int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    entity int,
    PRIMARY KEY (id)
);

CREATE TABLE _product_classification_data (
    _end_t timestamp,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(255),
    parent int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    is_product boolean,
    description text,
    PRIMARY KEY (id)
);

CREATE TABLE _entity_uid_data (
    _end_t timestamp,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    uid varchar(50),
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    entity_uid_type int,
    entity int,
    PRIMARY KEY (id)
);
CREATE UNIQUE INDEX entity_uid_unique_idx ON _entity_uid_data(uid, entity_uid_type) WHERE _end_t is null;

CREATE TABLE price (
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _commission_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(80),
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    rule text,
    _end_t timestamp,
    PRIMARY KEY (id)
);

CREATE TABLE _client_data (
    commission int,
    score int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    price_type int,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    persona int,
    PRIMARY KEY (id)
);

CREATE TABLE relation_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE credit_card (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE chequebook (
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    account bigint,
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE erm_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE own_descriptive_item (
    amount numeric(12,4),
    description text,
    entry bigint,
    id bigserial NOT NULL,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    own_item bigint,
    account_plan bigint,
    PRIMARY KEY (id)
);

CREATE TABLE _category_type_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    name varchar(80),
    PRIMARY KEY (id)
);

CREATE TABLE entry (
    date timestamp,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    cost_center bigint,
    id bigserial NOT NULL,
    entry bigint,
    PRIMARY KEY (id)
);

CREATE TABLE _barter_type_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    description text,
    id bigserial NOT NULL,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    account_plan int,
    _end_t timestamp,
    name varchar(80),
    PRIMARY KEY (id)
);

CREATE TABLE own_document_proceeding (
    amount numeric(12,4),
    proceeding bigint,
    entry bigint,
    own_document bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE entity_currency_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE fisco (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _product_provider_data (
    product int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    catalog_number char(30),
    provider int,
    _table bigint NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    _end_t timestamp,
    PRIMARY KEY (id)
);

CREATE TABLE task (
    assigned_time timestamp,
    task_type bigint,
    name varchar(80),
    parent int,
    assigned_by bigint,
    timestamp timestamp DEFAULT current_timestamp NOT NULL,
    task_priority bigint,
    planned_start_time timestamp,
    assigned_to bigint,
    task_status bigint,
    planned_end_time timestamp,
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE _category_attribute_data (
    category_type int,
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    name varchar(80),
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    _end_t timestamp,
    PRIMARY KEY (id)
);

CREATE TABLE product_file (
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE product_measure (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE contact (
    id serial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE product_tax (
    id bigserial NOT NULL,
    PRIMARY KEY (id)
);

CREATE TABLE __papo_patches__ (
    date timestamp DEFAULT now(),
    patch varchar(255),
    PRIMARY KEY (patch)
);

CREATE TABLE _movement_type_data (
    _start_t timestamp DEFAULT '1970/1/1 00:00:00 UTC' NOT NULL,
    _end_t timestamp,
    _table int NOT NULL,
    _state int DEFAULT -1 NOT NULL,
    id bigserial NOT NULL,
    name varchar(80),
    PRIMARY KEY (id)
);

CREATE TABLE entity_uid_type (
    id serial NOT NULL,
    PRIMARY KEY (id)
);
ALTER TABLE _product_measure_data
    ADD FOREIGN KEY (product) REFERENCES product
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_measure_data
    ADD FOREIGN KEY (product_measure_type) REFERENCES product_measure_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_measure_data
    ADD FOREIGN KEY (_table) REFERENCES product_measure
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_measure_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _product_file_data
    ADD FOREIGN KEY (product) REFERENCES product
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_file_data
    ADD FOREIGN KEY (file) REFERENCES file
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_file_data
    ADD FOREIGN KEY (_table) REFERENCES product_file
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_file_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _alien_entity_data
    ADD FOREIGN KEY (_table) REFERENCES alien_entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _alien_entity_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _alien_entity_data
    ADD FOREIGN KEY (entity) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _product_measure_type_data
    ADD FOREIGN KEY (_table) REFERENCES product_measure_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_measure_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;



ALTER TABLE erm_task
    ADD FOREIGN KEY (erm) REFERENCES erm
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE erm_task
    ADD FOREIGN KEY (task) REFERENCES task
    ON DELETE RESTRICT ON UPDATE RESTRICT;




ALTER TABLE credit_card_movement
    ADD FOREIGN KEY (movement) REFERENCES movement
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE credit_card_movement
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE credit_card_movement
    ADD FOREIGN KEY (plan) REFERENCES _credit_card_plan_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _category_entity_type_data
    ADD FOREIGN KEY (category_type) REFERENCES category_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _category_entity_type_data
    ADD FOREIGN KEY (entity_type) REFERENCES entity_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _category_entity_type_data
    ADD FOREIGN KEY (_table) REFERENCES category_entity_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _category_entity_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE alien_document_proceeding
    ADD FOREIGN KEY (proceeding) REFERENCES proceeding
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_document_proceeding
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_document_proceeding
    ADD FOREIGN KEY (alien_document) REFERENCES alien_document
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _bank_account_data
    ADD FOREIGN KEY (_table) REFERENCES bank_account
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _bank_account_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _bank_account_data
    ADD FOREIGN KEY (bank_branch) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _account_plan_data
    ADD FOREIGN KEY (parent) REFERENCES account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _account_plan_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _account_plan_data
    ADD FOREIGN KEY (_table) REFERENCES account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _employee_comm_client_data
    ADD FOREIGN KEY (commission) REFERENCES commission
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_comm_client_data
    ADD FOREIGN KEY (client) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_comm_client_data
    ADD FOREIGN KEY (employee) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_comm_client_data
    ADD FOREIGN KEY (_table) REFERENCES employee_comm_client
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_comm_client_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE admin_period
    ADD FOREIGN KEY (admin_period_type) REFERENCES admin_period_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _product_data
    ADD FOREIGN KEY (commission) REFERENCES commission
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_data
    ADD FOREIGN KEY (country) REFERENCES country
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_data
    ADD FOREIGN KEY (account_plan) REFERENCES account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_data
    ADD FOREIGN KEY (product_classification) REFERENCES product_classification
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_data
    ADD FOREIGN KEY (location) REFERENCES physical_structure
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_data
    ADD FOREIGN KEY (_table) REFERENCES product
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE document4document
    ADD FOREIGN KEY (own_document) REFERENCES own_document
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE document4document
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE document4document
    ADD FOREIGN KEY (alien_document) REFERENCES alien_document
    ON DELETE RESTRICT ON UPDATE RESTRICT;




ALTER TABLE _employee_comm_prod_data
    ADD FOREIGN KEY (commission) REFERENCES commission
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_comm_prod_data
    ADD FOREIGN KEY (product) REFERENCES product
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_comm_prod_data
    ADD FOREIGN KEY (employee) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_comm_prod_data
    ADD FOREIGN KEY (_table) REFERENCES employee_comm_prod
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_comm_prod_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _entity_relation_data
    ADD FOREIGN KEY (entity_from) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_relation_data
    ADD FOREIGN KEY (_table) REFERENCES entity_relation
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_relation_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_relation_data
    ADD FOREIGN KEY (entity_to) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE own_item_tax
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_item_tax
    ADD FOREIGN KEY (tax) REFERENCES _tax_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_item_tax
    ADD FOREIGN KEY (own_item) REFERENCES own_item
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE cheque_movement
    ADD FOREIGN KEY (movement) REFERENCES movement
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE cheque_movement
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE cheque_movement
    ADD FOREIGN KEY (cheque) REFERENCES cheque
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _document_type_data
    ADD FOREIGN KEY (_table) REFERENCES document_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _document_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _document_type_data
    ADD FOREIGN KEY (account_plan) REFERENCES account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _product_name_data
    ADD FOREIGN KEY (product) REFERENCES product
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_name_data
    ADD FOREIGN KEY (product_name_type) REFERENCES product_name_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_name_data
    ADD FOREIGN KEY (_table) REFERENCES product_name
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_name_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _tax_formula_data
    ADD FOREIGN KEY (_table) REFERENCES tax_formula
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _tax_formula_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _tax_formula_data
    ADD FOREIGN KEY (from_profile) REFERENCES tax_profile
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _tax_formula_data
    ADD FOREIGN KEY (to_profile) REFERENCES tax_profile
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE entity_type_uid_conf
    ADD FOREIGN KEY (entity_uid_type) REFERENCES _entity_uid_type_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE entity_type_uid_conf
    ADD FOREIGN KEY (entity_type) REFERENCES _entity_type_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE alien_product_item
    ADD FOREIGN KEY (product) REFERENCES _product_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_product_item
    ADD FOREIGN KEY (account_plan) REFERENCES _account_plan_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_product_item
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_product_item
    ADD FOREIGN KEY (alien_item) REFERENCES alien_item
    ON DELETE RESTRICT ON UPDATE RESTRICT;




ALTER TABLE _tax_data
    ADD FOREIGN KEY (_table) REFERENCES tax
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _tax_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _tax_data
    ADD FOREIGN KEY (account_plan) REFERENCES account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _credit_card_data
    ADD FOREIGN KEY (_table) REFERENCES credit_card
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _credit_card_data
    ADD FOREIGN KEY (accreditation_account) REFERENCES bank_account
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _credit_card_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _credit_card_data
    ADD FOREIGN KEY (account_plan) REFERENCES account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE task_recipient
    ADD FOREIGN KEY (task) REFERENCES task
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE task_recipient
    ADD FOREIGN KEY (to_make_to) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _entity_attribute_data
    ADD FOREIGN KEY (category_attribute) REFERENCES category_attribute
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_attribute_data
    ADD FOREIGN KEY (_table) REFERENCES entity_attribute
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_attribute_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_attribute_data
    ADD FOREIGN KEY (entity) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE alien_descriptive_item
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_descriptive_item
    ADD FOREIGN KEY (account_plan) REFERENCES _account_plan_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_descriptive_item
    ADD FOREIGN KEY (alien_item) REFERENCES alien_item
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE third_party_cheque_tax
    ADD FOREIGN KEY (third_party_cheque) REFERENCES third_party_cheque
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE third_party_cheque_tax
    ADD FOREIGN KEY (tax) REFERENCES _tax_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;



ALTER TABLE own_doc4own_doc
    ADD FOREIGN KEY (document_to) REFERENCES own_document
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_doc4own_doc
    ADD FOREIGN KEY (relation_type) REFERENCES document_relation_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_doc4own_doc
    ADD FOREIGN KEY (document_from) REFERENCES own_document
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _entity_type_data
    ADD FOREIGN KEY (parent) REFERENCES entity_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_type_data
    ADD FOREIGN KEY (_table) REFERENCES entity_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE own_document_tax
    ADD FOREIGN KEY (own_document) REFERENCES own_document
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_document_tax
    ADD FOREIGN KEY (tax) REFERENCES _tax_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _relation_types_data
    ADD FOREIGN KEY (entity_relation) REFERENCES entity_relation
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _relation_types_data
    ADD FOREIGN KEY (_table) REFERENCES relation_types
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _relation_types_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _relation_types_data
    ADD FOREIGN KEY (relation_type) REFERENCES relation_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _tax_profile_data
    ADD FOREIGN KEY (tax) REFERENCES tax
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _tax_profile_data
    ADD FOREIGN KEY (_table) REFERENCES tax_profile
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _tax_profile_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _own_pos_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _own_pos_data
    ADD FOREIGN KEY (_table) REFERENCES own_pos
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _own_pos_data
    ADD FOREIGN KEY (own_entity) REFERENCES own_entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _own_branch_data
    ADD FOREIGN KEY (own_entity) REFERENCES own_entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _own_branch_data
    ADD FOREIGN KEY (_table) REFERENCES own_branch
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _own_branch_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _bank_data
    ADD FOREIGN KEY (persona) REFERENCES persona
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _bank_data
    ADD FOREIGN KEY (_table) REFERENCES bank
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _bank_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;




ALTER TABLE _obligation_type_data
    ADD FOREIGN KEY (_table) REFERENCES obligation_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _obligation_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _obligation_type_data
    ADD FOREIGN KEY (account_plan) REFERENCES account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;




ALTER TABLE proceeding
    ADD FOREIGN KEY (responsable) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE proceeding
    ADD FOREIGN KEY (branch) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE proceeding
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _product_code_data
    ADD FOREIGN KEY (product) REFERENCES product
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_code_data
    ADD FOREIGN KEY (product_code_type) REFERENCES product_code_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_code_data
    ADD FOREIGN KEY (_table) REFERENCES product_code
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_code_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _logical2physical_data
    ADD FOREIGN KEY (_table) REFERENCES logical2physical
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _logical2physical_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _logical2physical_data
    ADD FOREIGN KEY (logical) REFERENCES logical_structure
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _logical2physical_data
    ADD FOREIGN KEY (physical) REFERENCES physical_structure
    ON DELETE RESTRICT ON UPDATE RESTRICT;



ALTER TABLE _price_data
    ADD FOREIGN KEY (currency_type) REFERENCES currency_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _price_data
    ADD FOREIGN KEY (product) REFERENCES product
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _price_data
    ADD FOREIGN KEY (price_type) REFERENCES price_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _price_data
    ADD FOREIGN KEY (_table) REFERENCES price
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _price_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _employee_data
    ADD FOREIGN KEY (commission) REFERENCES commission
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_data
    ADD FOREIGN KEY (own_entity) REFERENCES own_entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_data
    ADD FOREIGN KEY (_table) REFERENCES employee
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE alien_document_tax
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_document_tax
    ADD FOREIGN KEY (tax) REFERENCES _tax_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_document_tax
    ADD FOREIGN KEY (alien_document) REFERENCES alien_document
    ON DELETE RESTRICT ON UPDATE RESTRICT;



ALTER TABLE own_item
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_item
    ADD FOREIGN KEY (own_document) REFERENCES own_document
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_item
    ADD FOREIGN KEY (account_plan) REFERENCES _account_plan_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _role_data
    ADD FOREIGN KEY (_table) REFERENCES role
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _role_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE movement
    ADD FOREIGN KEY (proceeding) REFERENCES proceeding
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE movement
    ADD FOREIGN KEY (parent) REFERENCES movement
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE movement
    ADD FOREIGN KEY (movement_type) REFERENCES _movement_type_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE third_party_cheque
    ADD FOREIGN KEY (bank_branch) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE third_party_cheque
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE third_party_cheque
    ADD FOREIGN KEY (account_plan) REFERENCES _account_plan_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _alien_storehouse_data
    ADD FOREIGN KEY (_table) REFERENCES alien_storehouse
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _alien_storehouse_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _alien_storehouse_data
    ADD FOREIGN KEY (alien_entity) REFERENCES alien_entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _persona_tax_profile_data
    ADD FOREIGN KEY (tax_profile) REFERENCES tax_profile
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _persona_tax_profile_data
    ADD FOREIGN KEY (persona) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _persona_tax_profile_data
    ADD FOREIGN KEY (_table) REFERENCES persona_tax_profile
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _persona_tax_profile_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE own_document
    ADD FOREIGN KEY (own_document_type) REFERENCES _document_type_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_document
    ADD FOREIGN KEY (salesperson) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_document
    ADD FOREIGN KEY (recipient) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_document
    ADD FOREIGN KEY (transport) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_document
    ADD FOREIGN KEY (own_entity) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_document
    ADD FOREIGN KEY (authorized_by) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_document
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _employee_comm_prod_client_data
    ADD FOREIGN KEY (commission) REFERENCES commission
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_comm_prod_client_data
    ADD FOREIGN KEY (product) REFERENCES product
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_comm_prod_client_data
    ADD FOREIGN KEY (client) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_comm_prod_client_data
    ADD FOREIGN KEY (employee) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_comm_prod_client_data
    ADD FOREIGN KEY (_table) REFERENCES employee_comm_prod_client
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_comm_prod_client_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE own_product_stock_item
    ADD FOREIGN KEY (own_item) REFERENCES own_item
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_product_stock_item
    ADD FOREIGN KEY (product) REFERENCES _product_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE erm_log
    ADD FOREIGN KEY (erm) REFERENCES erm
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _task_priority_data
    ADD FOREIGN KEY (_table) REFERENCES task_priority
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _task_priority_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE obligation
    ADD FOREIGN KEY (obligation_type) REFERENCES _obligation_type_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE obligation
    ADD FOREIGN KEY (account_plan) REFERENCES _account_plan_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE alien_document
    ADD FOREIGN KEY (sender) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_document
    ADD FOREIGN KEY (alien_document_type) REFERENCES _document_type_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_document
    ADD FOREIGN KEY (own_entity) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_document
    ADD FOREIGN KEY (received_by) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_document
    ADD FOREIGN KEY (salesperson) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_document
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_document
    ADD FOREIGN KEY (transport) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _product_tax_data
    ADD FOREIGN KEY (product) REFERENCES product
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_tax_data
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_tax_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_tax_data
    ADD FOREIGN KEY (_table) REFERENCES product_tax
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_tax_data
    ADD FOREIGN KEY (tax) REFERENCES tax
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE alien_document_admin_period
    ADD FOREIGN KEY (admin_period) REFERENCES admin_period
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_document_admin_period
    ADD FOREIGN KEY (alien_document) REFERENCES alien_document
    ON DELETE RESTRICT ON UPDATE RESTRICT;



ALTER TABLE _bank_branch_data
    ADD FOREIGN KEY (persona) REFERENCES persona
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _bank_branch_data
    ADD FOREIGN KEY (_table) REFERENCES bank_branch
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _bank_branch_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _address_data
    ADD FOREIGN KEY (_table) REFERENCES address
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _address_data
    ADD FOREIGN KEY (address_type) REFERENCES address_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _address_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _address_data
    ADD FOREIGN KEY (entity) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;



ALTER TABLE _relation_type_data
    ADD FOREIGN KEY (_table) REFERENCES relation_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _relation_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _physical_structure_data
    ADD FOREIGN KEY (parent) REFERENCES physical_structure
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _physical_structure_data
    ADD FOREIGN KEY (_table) REFERENCES physical_structure
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _physical_structure_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _logical_structure_data
    ADD FOREIGN KEY (parent) REFERENCES logical_structure
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _logical_structure_data
    ADD FOREIGN KEY (_table) REFERENCES logical_structure
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _logical_structure_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;



ALTER TABLE _currency_type_data
    ADD FOREIGN KEY (_table) REFERENCES currency_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _currency_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _currency_type_data
    ADD FOREIGN KEY (account_plan) REFERENCES account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _country_data
    ADD FOREIGN KEY (_table) REFERENCES country
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _country_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _fisco_data
    ADD FOREIGN KEY (persona) REFERENCES persona
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _fisco_data
    ADD FOREIGN KEY (_table) REFERENCES fisco
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _fisco_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;



ALTER TABLE cheque
    ADD FOREIGN KEY (chequebook) REFERENCES chequebook
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE cheque
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE cheque
    ADD FOREIGN KEY (account_plan) REFERENCES _account_plan_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;



ALTER TABLE bank_account_movement
    ADD FOREIGN KEY (origin) REFERENCES _bank_account_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE bank_account_movement
    ADD FOREIGN KEY (destination) REFERENCES _bank_account_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE bank_account_movement
    ADD FOREIGN KEY (movement) REFERENCES movement
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _task_status_data
    ADD FOREIGN KEY (_table) REFERENCES task_status
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _task_status_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _file_data
    ADD FOREIGN KEY (_table) REFERENCES file
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _file_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE alien_item
    ADD FOREIGN KEY (alien_document) REFERENCES alien_document
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_item
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_item
    ADD FOREIGN KEY (account_plan) REFERENCES _account_plan_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE alien_item_tax
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_item_tax
    ADD FOREIGN KEY (tax) REFERENCES _tax_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_item_tax
    ADD FOREIGN KEY (alien_item) REFERENCES alien_item
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _credit_limit_data
    ADD FOREIGN KEY (movement_type) REFERENCES movement_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _credit_limit_data
    ADD FOREIGN KEY (_table) REFERENCES credit_limit
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _credit_limit_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _credit_limit_data
    ADD FOREIGN KEY (entity) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _ticket_data
    ADD FOREIGN KEY (ticket_type) REFERENCES ticket_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _ticket_data
    ADD FOREIGN KEY (_table) REFERENCES ticket
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _ticket_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _ticket_data
    ADD FOREIGN KEY (account_plan) REFERENCES account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;





ALTER TABLE _entity_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_data
    ADD FOREIGN KEY (account_plan) REFERENCES account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_data
    ADD FOREIGN KEY (_table) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE currency_movement
    ADD FOREIGN KEY (currency_type) REFERENCES _currency_type_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE currency_movement
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE currency_movement
    ADD FOREIGN KEY (movement) REFERENCES movement
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _own_entity_data
    ADD FOREIGN KEY (_table) REFERENCES own_entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _own_entity_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _own_entity_data
    ADD FOREIGN KEY (entity) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE cheque_tax
    ADD FOREIGN KEY (tax) REFERENCES _tax_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE cheque_tax
    ADD FOREIGN KEY (cheque) REFERENCES cheque
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _own_storehouse_data
    ADD FOREIGN KEY (own_entity) REFERENCES own_entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _own_storehouse_data
    ADD FOREIGN KEY (_table) REFERENCES own_storehouse
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _own_storehouse_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _product_store_data
    ADD FOREIGN KEY (product) REFERENCES product
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_store_data
    ADD FOREIGN KEY (storehouse) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_store_data
    ADD FOREIGN KEY (_table) REFERENCES product_store
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_store_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE barter_movement
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE barter_movement
    ADD FOREIGN KEY (barter) REFERENCES _barter_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE barter_movement
    ADD FOREIGN KEY (movement) REFERENCES movement
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _ticket_type_data
    ADD FOREIGN KEY (_table) REFERENCES ticket_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _ticket_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _ticket_type_data
    ADD FOREIGN KEY (account_plan) REFERENCES account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;




ALTER TABLE _provider_data
    ADD FOREIGN KEY (persona) REFERENCES persona
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _provider_data
    ADD FOREIGN KEY (_table) REFERENCES provider
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _provider_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE own_product_item
    ADD FOREIGN KEY (own_item) REFERENCES own_item
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_product_item
    ADD FOREIGN KEY (account_plan) REFERENCES _account_plan_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_product_item
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_product_item
    ADD FOREIGN KEY (price) REFERENCES _price_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;



ALTER TABLE _barter_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _barter_data
    ADD FOREIGN KEY (_table) REFERENCES barter
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _barter_data
    ADD FOREIGN KEY (barter_type) REFERENCES barter_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _barter_data
    ADD FOREIGN KEY (account_plan) REFERENCES account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _price_type_data
    ADD FOREIGN KEY (parent) REFERENCES price_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _price_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _price_type_data
    ADD FOREIGN KEY (branch) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _price_type_data
    ADD FOREIGN KEY (_table) REFERENCES price_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE obligation_movement
    ADD FOREIGN KEY (obligation) REFERENCES obligation
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE obligation_movement
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE obligation_movement
    ADD FOREIGN KEY (movement) REFERENCES movement
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _organization_data
    ADD FOREIGN KEY (own_entity) REFERENCES own_entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _organization_data
    ADD FOREIGN KEY (_table) REFERENCES organization
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _organization_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE own_document_admin_period
    ADD FOREIGN KEY (own_document) REFERENCES own_document
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_document_admin_period
    ADD FOREIGN KEY (admin_period) REFERENCES admin_period
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE erm
    ADD FOREIGN KEY (erm_type) REFERENCES _erm_type_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE erm
    ADD FOREIGN KEY (made_by) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _entity_uid_type_data
    ADD FOREIGN KEY (_table) REFERENCES entity_uid_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_uid_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _comm_prod_client_data
    ADD FOREIGN KEY (commission) REFERENCES commission
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _comm_prod_client_data
    ADD FOREIGN KEY (product) REFERENCES product
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _comm_prod_client_data
    ADD FOREIGN KEY (client) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _comm_prod_client_data
    ADD FOREIGN KEY (_table) REFERENCES comm_prod_client
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _comm_prod_client_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE ticket_movement
    ADD FOREIGN KEY (ticket) REFERENCES _ticket_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE ticket_movement
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE ticket_movement
    ADD FOREIGN KEY (movement) REFERENCES movement
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _product_code_type_data
    ADD FOREIGN KEY (_table) REFERENCES product_code_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_code_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE erm_file
    ADD FOREIGN KEY (erm) REFERENCES erm
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE erm_file
    ADD FOREIGN KEY (file) REFERENCES _file_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _task_type_data
    ADD FOREIGN KEY (_table) REFERENCES task_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _task_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE third_party_cheque_movement
    ADD FOREIGN KEY (third_party_cheque) REFERENCES third_party_cheque
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE third_party_cheque_movement
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE third_party_cheque_movement
    ADD FOREIGN KEY (movement) REFERENCES movement
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE erm_entity
    ADD FOREIGN KEY (erm) REFERENCES erm
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE erm_entity
    ADD FOREIGN KEY (made_to) REFERENCES _entity_types_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _contact_data
    ADD FOREIGN KEY (_table) REFERENCES contact
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _contact_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _contact_data
    ADD FOREIGN KEY (alien_entity) REFERENCES alien_entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _entity_types_data
    ADD FOREIGN KEY (entity_type) REFERENCES entity_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_types_data
    ADD FOREIGN KEY (_table) REFERENCES entity_types
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_types_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_types_data
    ADD FOREIGN KEY (entity) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _own_pos_document_data
    ADD FOREIGN KEY (document_type) REFERENCES document_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _own_pos_document_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _own_pos_document_data
    ADD FOREIGN KEY (_table) REFERENCES own_pos_document
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _own_pos_document_data
    ADD FOREIGN KEY (own_pos) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;



ALTER TABLE _product_name_type_data
    ADD FOREIGN KEY (_table) REFERENCES product_name_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_name_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _credit_card_plan_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _credit_card_plan_data
    ADD FOREIGN KEY (account_plan) REFERENCES account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _credit_card_plan_data
    ADD FOREIGN KEY (credit_card) REFERENCES credit_card
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _credit_card_plan_data
    ADD FOREIGN KEY (_table) REFERENCES credit_card_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _product_structure_data
    ADD FOREIGN KEY (product) REFERENCES product
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_structure_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_structure_data
    ADD FOREIGN KEY (_table) REFERENCES product_structure
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_structure_data
    ADD FOREIGN KEY (component) REFERENCES product
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE alien_doc4alien_doc
    ADD FOREIGN KEY (document_to) REFERENCES alien_document
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_doc4alien_doc
    ADD FOREIGN KEY (relation_type) REFERENCES document_relation_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE alien_doc4alien_doc
    ADD FOREIGN KEY (document_from) REFERENCES alien_document
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _employee_role_data
    ADD FOREIGN KEY (employee) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_role_data
    ADD FOREIGN KEY (role) REFERENCES role
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_role_data
    ADD FOREIGN KEY (_table) REFERENCES employee_role
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _employee_role_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _erm_type_data
    ADD FOREIGN KEY (_table) REFERENCES erm_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _erm_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _sys_account2account_plan_data
    ADD FOREIGN KEY (account) REFERENCES account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _sys_account2account_plan_data
    ADD FOREIGN KEY (_table) REFERENCES sys_account2account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _sys_account2account_plan_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;




ALTER TABLE _address_type_data
    ADD FOREIGN KEY (_table) REFERENCES address_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _address_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _persona_data
    ADD FOREIGN KEY (_table) REFERENCES persona
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _persona_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _persona_data
    ADD FOREIGN KEY (alien_entity) REFERENCES alien_entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;



ALTER TABLE _entity_currency_type_data
    ADD FOREIGN KEY (currency_type) REFERENCES currency_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_currency_type_data
    ADD FOREIGN KEY (_table) REFERENCES entity_currency_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_currency_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_currency_type_data
    ADD FOREIGN KEY (entity) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _product_classification_data
    ADD FOREIGN KEY (parent) REFERENCES product_classification
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_classification_data
    ADD FOREIGN KEY (_table) REFERENCES product_classification
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_classification_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _entity_uid_data
    ADD FOREIGN KEY (_table) REFERENCES entity_uid
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_uid_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_uid_data
    ADD FOREIGN KEY (entity_uid_type) REFERENCES entity_uid_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _entity_uid_data
    ADD FOREIGN KEY (entity) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE _commission_data
    ADD FOREIGN KEY (_table) REFERENCES commission
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _commission_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _client_data
    ADD FOREIGN KEY (commission) REFERENCES commission
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _client_data
    ADD FOREIGN KEY (price_type) REFERENCES price_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _client_data
    ADD FOREIGN KEY (_table) REFERENCES client
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _client_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _client_data
    ADD FOREIGN KEY (persona) REFERENCES persona
    ON DELETE RESTRICT ON UPDATE RESTRICT;



ALTER TABLE chequebook
    ADD FOREIGN KEY (account) REFERENCES _bank_account_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;


ALTER TABLE own_descriptive_item
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_descriptive_item
    ADD FOREIGN KEY (own_item) REFERENCES own_item
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_descriptive_item
    ADD FOREIGN KEY (account_plan) REFERENCES _account_plan_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _category_type_data
    ADD FOREIGN KEY (_table) REFERENCES category_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _category_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE entry
    ADD FOREIGN KEY (cost_center) REFERENCES _logical_structure_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE entry
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _barter_type_data
    ADD FOREIGN KEY (_table) REFERENCES barter_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _barter_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _barter_type_data
    ADD FOREIGN KEY (account_plan) REFERENCES account_plan
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE own_document_proceeding
    ADD FOREIGN KEY (proceeding) REFERENCES proceeding
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_document_proceeding
    ADD FOREIGN KEY (entry) REFERENCES entry
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE own_document_proceeding
    ADD FOREIGN KEY (own_document) REFERENCES own_document
    ON DELETE RESTRICT ON UPDATE RESTRICT;



ALTER TABLE _product_provider_data
    ADD FOREIGN KEY (product) REFERENCES product
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_provider_data
    ADD FOREIGN KEY (provider) REFERENCES entity
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_provider_data
    ADD FOREIGN KEY (_table) REFERENCES product_provider
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _product_provider_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE task
    ADD FOREIGN KEY (task_type) REFERENCES _task_type_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE task
    ADD FOREIGN KEY (parent) REFERENCES task
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE task
    ADD FOREIGN KEY (assigned_by) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE task
    ADD FOREIGN KEY (task_priority) REFERENCES _task_priority_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE task
    ADD FOREIGN KEY (assigned_to) REFERENCES _entity_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE task
    ADD FOREIGN KEY (task_status) REFERENCES _task_status_data
    ON DELETE RESTRICT ON UPDATE RESTRICT;

ALTER TABLE _category_attribute_data
    ADD FOREIGN KEY (category_type) REFERENCES category_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _category_attribute_data
    ADD FOREIGN KEY (_table) REFERENCES category_attribute
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _category_attribute_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;






ALTER TABLE _movement_type_data
    ADD FOREIGN KEY (_table) REFERENCES movement_type
    ON DELETE RESTRICT ON UPDATE RESTRICT;
ALTER TABLE _movement_type_data
    ADD FOREIGN KEY (_state) REFERENCES __state__
    ON DELETE RESTRICT ON UPDATE RESTRICT;


-- 01-__state__.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

INSERT INTO __state__ (id, state) VALUES (-1, 'UNDEFINED');-- 01-account-plan.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

begin;
insert into account_plan (id) values (1);
insert into _account_plan_data (id,code,name,_table,parent) values (1,'1','ACTIVO',1,NULL);
insert into account_plan (id) values (2);
insert into _account_plan_data (id,code,name,_table,parent) values (2,'1.1','ACTIVO CORRIENTE',2,1);
insert into account_plan (id) values (3);
insert into _account_plan_data (id,code,name,_table,parent) values (3,'1.1.1','DISPONIBILIDADES',3,2);
insert into account_plan (id) values (4);
insert into _account_plan_data (id,code,name,_table,parent) values (4,'1.1.1.1','CAJA',4,3);
insert into account_plan (id) values (5);
insert into _account_plan_data (id,code,name,_table,parent) values (5,'1.1.1.2','BANCOS',5,3);
insert into account_plan (id) values (6);
insert into _account_plan_data (id,code,name,_table,parent) values (6,'1.1.2','INVERSIONES CORRIENTES',6,2);
insert into account_plan (id) values (7);
insert into _account_plan_data (id,code,name,_table,parent) values (7,'1.1.2.1','TTULO VALORES',7,6);
insert into account_plan (id) values (8);
insert into _account_plan_data (id,code,name,_table,parent) values (8,'1.1.2.2','OTRAS INVERSIONES',8,6);
insert into account_plan (id) values (9);
insert into _account_plan_data (id,code,name,_table,parent) values (9,'1.1.3','CRDITOS POR VENTA',9,2);
insert into account_plan (id) values (10);
insert into _account_plan_data (id,code,name,_table,parent) values (10,'1.1.3.1','DEUDORES POR VENTA',10,9);
insert into account_plan (id) values (11);
insert into _account_plan_data (id,code,name,_table,parent) values (11,'1.1.3.2','DEUDORES EN GESTIN EXTRAJUDICIAL',11,9);
insert into account_plan (id) values (12);
insert into _account_plan_data (id,code,name,_table,parent) values (12,'1.1.3.3','DEUDORES EN GESTIN JUDICIAL',12,9);
insert into account_plan (id) values (13);
insert into _account_plan_data (id,code,name,_table,parent) values (13,'1.1.3.4','DOCUMENTOS A COBRAR',13,9);
insert into account_plan (id) values (14);
insert into _account_plan_data (id,code,name,_table,parent) values (14,'1.1.3.5','PREVISIN PARA DEUDORES INCOBRABLES',14,9);
insert into account_plan (id) values (15);
insert into _account_plan_data (id,code,name,_table,parent) values (15,'1.1.4','CRDITOS FISCALES Y SOCIALES',15,2);
insert into account_plan (id) values (16);
insert into _account_plan_data (id,code,name,_table,parent) values (16,'1.1.4.1','SALDO DDJJ IVA A FAVOR',16,15);
insert into account_plan (id) values (17);
insert into _account_plan_data (id,code,name,_table,parent) values (17,'1.1.4.2','RETENCIONES IVA SUIFRIDAS',17,15);
insert into account_plan (id) values (18);
insert into _account_plan_data (id,code,name,_table,parent) values (18,'1.1.4.3','CRDITO DISCAL IVA',18,15);
insert into account_plan (id) values (19);
insert into _account_plan_data (id,code,name,_table,parent) values (19,'1.1.5','ANTICIPOS',19,2);
insert into account_plan (id) values (20);
insert into _account_plan_data (id,code,name,_table,parent) values (20,'1.1.5.1','ANTICIPOS VARIOS',20,19);
insert into account_plan (id) values (21);
insert into _account_plan_data (id,code,name,_table,parent) values (21,'1.1.5.2','GASTOS PAGADOS POR ADELANTADO',21,19);
insert into account_plan (id) values (22);
insert into _account_plan_data (id,code,name,_table,parent) values (22,'1.1.5.3','CUENTA PARTICULAR SOCIOS',22,19);
insert into account_plan (id) values (23);
insert into _account_plan_data (id,code,name,_table,parent) values (23,'1.1.6','BIENES DE CAMBIO',23,2);
insert into account_plan (id) values (24);
insert into _account_plan_data (id,code,name,_table,parent) values (24,'1.1.6.1','MATERIA PRIMA',24,23);
insert into account_plan (id) values (25);
insert into _account_plan_data (id,code,name,_table,parent) values (25,'1.1.6.2','PRODUCTOS EN PROCESO',25,23);
insert into account_plan (id) values (26);
insert into _account_plan_data (id,code,name,_table,parent) values (26,'1.1.6.3','PRODUCTOS TERMINADOS',26,23);
insert into account_plan (id) values (27);
insert into _account_plan_data (id,code,name,_table,parent) values (27,'1.1.6.4','PRODUCTOS EN CONSIGNACIN',27,23);
insert into account_plan (id) values (28);
insert into _account_plan_data (id,code,name,_table,parent) values (28,'1.2','ACTIVO NO CORRIENTE',28,1);
insert into account_plan (id) values (29);
insert into _account_plan_data (id,code,name,_table,parent) values (29,'1.2.1','INVERSIONES NO CORRIENTES',29,28);
insert into account_plan (id) values (30);
insert into _account_plan_data (id,code,name,_table,parent) values (30,'1.2.1.1','TTULOS VALORES',30,29);
insert into account_plan (id) values (31);
insert into _account_plan_data (id,code,name,_table,parent) values (31,'1.2.1.2','INMUEBLES',31,29);
insert into account_plan (id) values (32);
insert into _account_plan_data (id,code,name,_table,parent) values (32,'1.2.1.99','INMUEBLES - AJTE CBIOS PO ADQ. MONEDA',32,NULL);
insert into account_plan (id) values (33);
insert into _account_plan_data (id,code,name,_table,parent) values (33,'1.2.2','BIENES DE USO - VALOR DE ORGEN',33,28);
insert into account_plan (id) values (34);
insert into _account_plan_data (id,code,name,_table,parent) values (34,'1.2.2.1','EQUIPOS',34,33);
insert into account_plan (id) values (35);
insert into _account_plan_data (id,code,name,_table,parent) values (35,'1.2.2.2','INMUEBLES',35,33);
insert into account_plan (id) values (36);
insert into _account_plan_data (id,code,name,_table,parent) values (36,'1.2.2.3','RODADOS',36,33);
insert into account_plan (id) values (37);
insert into _account_plan_data (id,code,name,_table,parent) values (37,'1.2.2.99','BIENES DE USO - AJTE CBIOS PO ADQ. MON',37,NULL);
insert into account_plan (id) values (38);
insert into _account_plan_data (id,code,name,_table,parent) values (38,'1.2.3','AMORTIZACIN ACUMULADA BIENES DE USO',38,28);
insert into account_plan (id) values (39);
insert into _account_plan_data (id,code,name,_table,parent) values (39,'1.2.3.1','A.A. EQUIPOS',39,38);
insert into account_plan (id) values (40);
insert into _account_plan_data (id,code,name,_table,parent) values (40,'1.2.3.2','A.A. INMUEBLES',40,38);
insert into account_plan (id) values (41);
insert into _account_plan_data (id,code,name,_table,parent) values (41,'1.2.3.3','A.A. RODADOS',41,38);
insert into account_plan (id) values (42);
insert into _account_plan_data (id,code,name,_table,parent) values (42,'1.2.3.99','AMORT. ACUM. - CBIOS PO ADQ. MONEDA',42,NULL);
insert into account_plan (id) values (43);
insert into _account_plan_data (id,code,name,_table,parent) values (43,'2','PASIVO',43,NULL);
insert into account_plan (id) values (44);
insert into _account_plan_data (id,code,name,_table,parent) values (44,'2.1','PASIVO CORRIENTE',44,43);
insert into account_plan (id) values (45);
insert into _account_plan_data (id,code,name,_table,parent) values (45,'2.1.1','DEUDAS COMERCIALES',45,44);
insert into account_plan (id) values (46);
insert into _account_plan_data (id,code,name,_table,parent) values (46,'2.1.1.1','PROVEEDORES',46,45);
insert into account_plan (id) values (47);
insert into _account_plan_data (id,code,name,_table,parent) values (47,'2.1.1.2','DOCUMENTOS A PAGAR',47,45);
insert into account_plan (id) values (48);
insert into _account_plan_data (id,code,name,_table,parent) values (48,'2.1.2','DEUDAS BANCARIAS Y FINANCIERAS',48,44);
insert into account_plan (id) values (49);
insert into _account_plan_data (id,code,name,_table,parent) values (49,'2.1.2.1','DEUDAS BANCARIAS',49,48);
insert into account_plan (id) values (50);
insert into _account_plan_data (id,code,name,_table,parent) values (50,'2.1.2.2','DEUDAS FINANCIERAS',50,48);
insert into account_plan (id) values (51);
insert into _account_plan_data (id,code,name,_table,parent) values (51,'2.1.3','DEUDAS FISCALES',51,44);
insert into account_plan (id) values (52);
insert into _account_plan_data (id,code,name,_table,parent) values (52,'2.1.3.1','IMPUESTO A LAS GANANCIAS - DDJJ A PAGAR',52,51);
insert into account_plan (id) values (53);
insert into _account_plan_data (id,code,name,_table,parent) values (53,'2.1.3.2','RETENCION GANANCIAS A DEPOSITAR',53,51);
insert into account_plan (id) values (54);
insert into _account_plan_data (id,code,name,_table,parent) values (54,'2.1.3.3','IVA DBITO FISCAL',54,51);
insert into account_plan (id) values (55);
insert into _account_plan_data (id,code,name,_table,parent) values (55,'2.1.3.4','IVA - SALDO DDJJ A PAGAR',55,51);
insert into account_plan (id) values (56);
insert into _account_plan_data (id,code,name,_table,parent) values (56,'2.1.3.5','RETENCIN IVA A DEPOSITAR',56,51);
insert into account_plan (id) values (57);
insert into _account_plan_data (id,code,name,_table,parent) values (57,'2.1.3.6','PERCEPCIN IVA A DEPOSITAR',57,51);
insert into account_plan (id) values (58);
insert into _account_plan_data (id,code,name,_table,parent) values (58,'2.1.3.7','DDJJ IIBB',58,51);
insert into account_plan (id) values (59);
insert into _account_plan_data (id,code,name,_table,parent) values (59,'2.1.3.8','RENTENCIONES INGRESOS BRUTOS A DEPOSITAR',59,51);
insert into account_plan (id) values (60);
insert into _account_plan_data (id,code,name,_table,parent) values (60,'2.1.3.9','IMPUESTO INMOBILIARIO',60,51);
insert into account_plan (id) values (61);
insert into _account_plan_data (id,code,name,_table,parent) values (61,'2.1.3.10','COMERCIO E INDUSTRIA',61,NULL);
insert into account_plan (id) values (62);
insert into _account_plan_data (id,code,name,_table,parent) values (62,'2.1.4','DEUDAS SOCIALES',62,44);
insert into account_plan (id) values (63);
insert into _account_plan_data (id,code,name,_table,parent) values (63,'2.1.4.1','SUELDOS Y JORNALES A PAGAR',63,62);
insert into account_plan (id) values (64);
insert into _account_plan_data (id,code,name,_table,parent) values (64,'2.1.4.2','CARGA SOCIALES A PAGAR',64,62);
insert into account_plan (id) values (65);
insert into _account_plan_data (id,code,name,_table,parent) values (65,'2.1.4.3','PREVISION VACACIONES',65,62);
insert into account_plan (id) values (66);
insert into _account_plan_data (id,code,name,_table,parent) values (66,'2.1.4.4','PREVISIN SAC',66,62);
insert into account_plan (id) values (67);
insert into _account_plan_data (id,code,name,_table,parent) values (67,'2.1.5','OTROS PASIVOS CP',67,44);
insert into account_plan (id) values (68);
insert into _account_plan_data (id,code,name,_table,parent) values (68,'2.1.5.1','ANTICIPOS DE CLIENTES',68,67);
insert into account_plan (id) values (69);
insert into _account_plan_data (id,code,name,_table,parent) values (69,'2.1.6','PREVISIONES',69,44);
insert into account_plan (id) values (70);
insert into _account_plan_data (id,code,name,_table,parent) values (70,'2.2','PASIVO NO CORRIENTE',70,43);
insert into account_plan (id) values (71);
insert into _account_plan_data (id,code,name,_table,parent) values (71,'2.2.1','DEUDAS COMERCIALES',71,70);
insert into account_plan (id) values (72);
insert into _account_plan_data (id,code,name,_table,parent) values (72,'2.2.1.1','PROVEEDORES LARGO PLAZO',72,71);
insert into account_plan (id) values (73);
insert into _account_plan_data (id,code,name,_table,parent) values (73,'2.2.2','DEUDAS BANCARIAS Y FINANCIERAS',73,70);
insert into account_plan (id) values (74);
insert into _account_plan_data (id,code,name,_table,parent) values (74,'2.2.2.1','DEUDAS FINANCIERAS LP',74,73);
insert into account_plan (id) values (75);
insert into _account_plan_data (id,code,name,_table,parent) values (75,'2.2.3','DEUDAS FISCALES',75,70);
insert into account_plan (id) values (76);
insert into _account_plan_data (id,code,name,_table,parent) values (76,'2.2.3.1','MORATORIA LP',76,75);
insert into account_plan (id) values (77);
insert into _account_plan_data (id,code,name,_table,parent) values (77,'2.2.4','DEUDAS SOCIALES',77,70);
insert into account_plan (id) values (78);
insert into _account_plan_data (id,code,name,_table,parent) values (78,'2.2.4.1','INDEMNIZACIONES LP',78,77);
insert into account_plan (id) values (79);
insert into _account_plan_data (id,code,name,_table,parent) values (79,'2.2.5','PREVISIONES',79,70);
insert into account_plan (id) values (80);
insert into _account_plan_data (id,code,name,_table,parent) values (80,'2.2.5.1','PREVISIN',80,79);
insert into account_plan (id) values (81);
insert into _account_plan_data (id,code,name,_table,parent) values (81,'3','PATRIMONIO NETO',81,NULL);
insert into account_plan (id) values (82);
insert into _account_plan_data (id,code,name,_table,parent) values (82,'4','INGRESOS',82,NULL);
insert into account_plan (id) values (83);
insert into _account_plan_data (id,code,name,_table,parent) values (83,'4.1','INGRESOS OPERATIVOS',83,82);
insert into account_plan (id) values (84);
insert into _account_plan_data (id,code,name,_table,parent) values (84,'4.1.1','INGRESOS POR VENTA',84,83);
insert into account_plan (id) values (85);
insert into _account_plan_data (id,code,name,_table,parent) values (85,'4.1.1.1','INGRESOS POR VENTA',85,84);
insert into account_plan (id) values (86);
insert into _account_plan_data (id,code,name,_table,parent) values (86,'4.1.1.99','INGRESOS X VTA - AJTE PODER ADQ. MONEDA',86,NULL);
insert into account_plan (id) values (87);
insert into _account_plan_data (id,code,name,_table,parent) values (87,'4.1.2','INGRESOS VARIOS',87,83);
insert into account_plan (id) values (88);
insert into _account_plan_data (id,code,name,_table,parent) values (88,'4.1.3','INGRESOS FINANCIEROS',88,83);
insert into account_plan (id) values (89);
insert into _account_plan_data (id,code,name,_table,parent) values (89,'4.1.3.1','INTERESES GANADOS',89,88);
insert into account_plan (id) values (90);
insert into _account_plan_data (id,code,name,_table,parent) values (90,'4.1.3.2','DIF. DE CAMBIO GANADA',90,88);
insert into account_plan (id) values (91);
insert into _account_plan_data (id,code,name,_table,parent) values (91,'4.1.3.99','INGRESOS FINANCIEROS - AJTE PO ADQ. MON',91,NULL);
insert into account_plan (id) values (92);
insert into _account_plan_data (id,code,name,_table,parent) values (92,'4.2','INGRESOS EXTRAORDINARIOS',92,82);
insert into account_plan (id) values (93);
insert into _account_plan_data (id,code,name,_table,parent) values (93,'4.2.1','RESULTADO VENTA DE ACTIVOS OPERATIVOS',93,92);
insert into account_plan (id) values (94);
insert into _account_plan_data (id,code,name,_table,parent) values (94,'4.2.1.1','RESULTADO VENTA DE BIENES DE USO',94,93);
insert into account_plan (id) values (95);
insert into _account_plan_data (id,code,name,_table,parent) values (95,'5','EGRESOS',95,NULL);
insert into account_plan (id) values (96);
insert into _account_plan_data (id,code,name,_table,parent) values (96,'5.1','EGRESOS OPERATIVOS',96,95);
insert into account_plan (id) values (97);
insert into _account_plan_data (id,code,name,_table,parent) values (97,'5.1.1','COSTOS VARIABLES',97,96);
insert into account_plan (id) values (98);
insert into _account_plan_data (id,code,name,_table,parent) values (98,'5.1.1.1','COMPRA DE MATERIA PRIMA Y MATERIALES',98,97);
insert into account_plan (id) values (99);
insert into _account_plan_data (id,code,name,_table,parent) values (99,'5.1.1.2','SUELDOS Y JORNALES',99,97);
insert into account_plan (id) values (100);
insert into _account_plan_data (id,code,name,_table,parent) values (100,'5.1.1.3','CARGAS SOCIALES',100,97);
insert into account_plan (id) values (101);
insert into _account_plan_data (id,code,name,_table,parent) values (101,'5.1.1.4','GASTOS DE PERSONAL',101,97);
insert into account_plan (id) values (102);
insert into _account_plan_data (id,code,name,_table,parent) values (102,'5.1.1.5','SUELDO ANUAL COMPLEMENTARIO',102,97);
insert into account_plan (id) values (103);
insert into _account_plan_data (id,code,name,_table,parent) values (103,'5.1.1.6','VACACIONES',103,97);
insert into account_plan (id) values (104);
insert into _account_plan_data (id,code,name,_table,parent) values (104,'5.1.1.7','AMORTIZACIN',104,97);
insert into account_plan (id) values (105);
insert into _account_plan_data (id,code,name,_table,parent) values (105,'5.1.1.8','HONORARIOS',105,97);
insert into account_plan (id) values (106);
insert into _account_plan_data (id,code,name,_table,parent) values (106,'5.1.1.9','MANTENIMIENTO',106,97);
insert into account_plan (id) values (107);
insert into _account_plan_data (id,code,name,_table,parent) values (107,'5.1.1.10','ELECTRICIDAD',107,NULL);
insert into account_plan (id) values (108);
insert into _account_plan_data (id,code,name,_table,parent) values (108,'5.1.1.11','GAS',108,NULL);
insert into account_plan (id) values (109);
insert into _account_plan_data (id,code,name,_table,parent) values (109,'5.1.1.12','COMBUSTIBLE',109,NULL);
insert into account_plan (id) values (110);
insert into _account_plan_data (id,code,name,_table,parent) values (110,'5.1.1.13','GASTOS DE MOVILIDAD Y FLETE',110,NULL);
insert into account_plan (id) values (111);
insert into _account_plan_data (id,code,name,_table,parent) values (111,'5.1.1.14','INSUMOS DE OFICINA',111,NULL);
insert into account_plan (id) values (112);
insert into _account_plan_data (id,code,name,_table,parent) values (112,'5.1.1.15','INCOBRABLES',112,NULL);
insert into account_plan (id) values (113);
insert into _account_plan_data (id,code,name,_table,parent) values (113,'5.1.1.16','IMPUESOS TASAS Y CONTRIBUCIONES',113,NULL);
insert into account_plan (id) values (114);
insert into _account_plan_data (id,code,name,_table,parent) values (114,'5.1.1.17','GASTOS DE PUBLICIDAD',114,NULL);
insert into account_plan (id) values (115);
insert into _account_plan_data (id,code,name,_table,parent) values (115,'5.1.1.18','CORREO',115,NULL);
insert into account_plan (id) values (116);
insert into _account_plan_data (id,code,name,_table,parent) values (116,'5.1.1.19','GASTOS VARIOS',116,NULL);
insert into account_plan (id) values (117);
insert into _account_plan_data (id,code,name,_table,parent) values (117,'5.1.1.20','INTERESES PERDIDOS',117,NULL);
insert into account_plan (id) values (118);
insert into _account_plan_data (id,code,name,_table,parent) values (118,'5.1.1.21','DIFERENCIA DE CAMBIO PERDIDA',118,NULL);
insert into account_plan (id) values (119);
insert into _account_plan_data (id,code,name,_table,parent) values (119,'5.1.1.22','GASTOS BANCARIOS',119,NULL);
insert into account_plan (id) values (120);
insert into _account_plan_data (id,code,name,_table,parent) values (120,'5.1.2','COSTOS FIJOS',120,96);
insert into account_plan (id) values (121);
insert into _account_plan_data (id,code,name,_table,parent) values (121,'5.1.2.1','ALQUILERES',121,120);
insert into account_plan (id) values (122);
insert into _account_plan_data (id,code,name,_table,parent) values (122,'5.1.2.2','SEGUROS',122,120);
insert into account_plan (id) values (123);
insert into _account_plan_data (id,code,name,_table,parent) values (123,'5.1.2.3','SUELDOS Y JORNALES',123,120);
insert into account_plan (id) values (124);
insert into _account_plan_data (id,code,name,_table,parent) values (124,'5.1.2.4','CARGAS SOCIALES',124,120);
insert into account_plan (id) values (125);
insert into _account_plan_data (id,code,name,_table,parent) values (125,'5.1.2.5','GASTOS DE PERSONAL',125,120);
insert into account_plan (id) values (126);
insert into _account_plan_data (id,code,name,_table,parent) values (126,'5.1.2.6','SUELDO ANUAL COMPLEMENTARIO',126,120);
insert into account_plan (id) values (127);
insert into _account_plan_data (id,code,name,_table,parent) values (127,'5.1.2.7','VACACIONES',127,120);
insert into account_plan (id) values (128);
insert into _account_plan_data (id,code,name,_table,parent) values (128,'5.1.2.8','AMORTIZACIN',128,120);
insert into account_plan (id) values (129);
insert into _account_plan_data (id,code,name,_table,parent) values (129,'5.1.2.9','HONORARIOS',129,120);
insert into account_plan (id) values (130);
insert into _account_plan_data (id,code,name,_table,parent) values (130,'5.1.2.10','MANTENIMIENTO',130,NULL);
insert into account_plan (id) values (131);
insert into _account_plan_data (id,code,name,_table,parent) values (131,'5.1.2.11','ELECTRICIDAD',131,NULL);
insert into account_plan (id) values (132);
insert into _account_plan_data (id,code,name,_table,parent) values (132,'5.1.2.12','GAS',132,NULL);
insert into account_plan (id) values (133);
insert into _account_plan_data (id,code,name,_table,parent) values (133,'5.1.2.13','COMBUSTIBLE',133,NULL);
insert into account_plan (id) values (134);
insert into _account_plan_data (id,code,name,_table,parent) values (134,'5.1.2.14','GASTOS DE MOVILIDAD Y FLETE',134,NULL);
insert into account_plan (id) values (135);
insert into _account_plan_data (id,code,name,_table,parent) values (135,'5.1.2.15','INSUMOS DE OFICINA',135,NULL);
insert into account_plan (id) values (136);
insert into _account_plan_data (id,code,name,_table,parent) values (136,'5.1.2.16','INCOBRABLES',136,NULL);
insert into account_plan (id) values (137);
insert into _account_plan_data (id,code,name,_table,parent) values (137,'5.1.2.17','IMPUESOS TASAS Y CONTRIBUCIONES',137,NULL);
insert into account_plan (id) values (138);
insert into _account_plan_data (id,code,name,_table,parent) values (138,'5.1.2.18','GASTOS DE PUBLICIDAD',138,NULL);
insert into account_plan (id) values (139);
insert into _account_plan_data (id,code,name,_table,parent) values (139,'5.1.2.19','CORREO',139,NULL);
insert into account_plan (id) values (140);
insert into _account_plan_data (id,code,name,_table,parent) values (140,'5.1.2.20','GASTOS VARIOS',140,NULL);
insert into account_plan (id) values (141);
insert into _account_plan_data (id,code,name,_table,parent) values (141,'5.1.2.21','INTERESES PERDIDOS',141,NULL);
insert into account_plan (id) values (142);
insert into _account_plan_data (id,code,name,_table,parent) values (142,'5.1.2.22','DIFERENCIA DE CAMBIO PERDIDA',142,NULL);
insert into account_plan (id) values (143);
insert into _account_plan_data (id,code,name,_table,parent) values (143,'5.1.2.23','GASTOS BANCARIOS',143,NULL);
insert into account_plan (id) values (144);
insert into _account_plan_data (id,code,name,_table,parent) values (144,'5.2','EGRESOS EXTRAORDINARIOS',144,95);
insert into account_plan (id) values (145);
insert into _account_plan_data (id,code,name,_table,parent) values (145,'5.2.1','VENTA DE ACTIVOS OPERATIVOS',145,144);
insert into account_plan (id) values (146);
insert into _account_plan_data (id,code,name,_table,parent) values (146,'5.2.1.1','VENTA DE BIENES DE USO',146,145);
insert into account_plan (id) values (147);
insert into _account_plan_data (id,code,name,_table,parent) values (147,'5.2.2','OTROS EGRESOS EXTRAORDINARIOS',147,144);
insert into account_plan (id) values (148);
insert into _account_plan_data (id,code,name,_table,parent) values (148,'5.2.2.1','SINIESTROS',148,147);
select setval('"account_plan_id_seq"',149,true);
select setval('"_account_plan_data_id_seq"',149,true);
commit;
-- 01-sys_account2account_plan.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

begin;
insert into sys_account2account_plan (id) values (1);
insert into _sys_account2account_plan_data (id,_table,name,account,code) values (1,1,'Impuestos',113,'5.1.1.16%');
insert into sys_account2account_plan (id) values (2);
insert into _sys_account2account_plan_data (id,_table,name,account,code) values (2,2,'Proveedores',46,'2.1.1.1%');
insert into sys_account2account_plan (id) values (3);
insert into _sys_account2account_plan_data (id,_table,name,account,code) values (3,3,'Clientes',84,'4.1.1%');
insert into sys_account2account_plan (id) values (4);
insert into _sys_account2account_plan_data (id,_table,name,account,code) values (4,4,'Productos',23,'1.1.6%');
insert into sys_account2account_plan (id) values (5);
insert into _sys_account2account_plan_data (id,_table,name,account,code) values (5,5,'Fisco',51,'2.1.3%');
insert into sys_account2account_plan (id) values (6);
insert into _sys_account2account_plan_data (id,_table,name,account,code) values (6,6,'Gastos',96,'5.1%');
commit;-- 01-address_type.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

insert into address_type (id) values (1);
insert into _address_type_data (id,_table,name) values (1,1,'Direccin');
insert into address_type (id) values (2);
insert into _address_type_data (id,_table,name) values (2,2,'Telfono');
insert into address_type (id) values (3);
insert into _address_type_data (id,_table,name) values (3,3,'URL');
insert into address_type (id) values (4);
insert into _address_type_data (id,_table,name) values (4,4,'e-mail');
insert into address_type (id) values (5);
insert into _address_type_data (id,_table,name) values (5,5,'Chat');
insert into address_type (id) values (6);
insert into _address_type_data (id,_table,name) values (6,6,'Fax');
insert into address_type (id) values (7);
insert into _address_type_data (id,_table,name) values (7,7,'Tel/Fax');
insert into address_type (id) values (8);
insert into _address_type_data (id,_table,name) values (8,8,'Cdigo postal');
-- 01-country.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

-- this is a comment

DELETE FROM "_country_data";
DELETE FROM "country";
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Afghanistan',4,'AF','AFG',12,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Albania',8,'AL','ALB',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Algeria',12,'DZ','DZA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'American Samoa',16,'AS','ASM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Andorra',20,'AD','AND',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Angola',24,'AO','AGO',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Azerbaijan',31,'AZ','AZE',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Australia',36,'AU','AUS',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Austria',40,'AT','AUT',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Bahamas',44,'BS','BHS',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Bahrain',48,'BH','BHR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Bangladesh',50,'BD','BGD',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Armenia',51,'AM','ARM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Barbados',52,'BB','BRB',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Belgium',56,'BE','BEL',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Bermuda',60,'BM','BMU',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Bhutan',64,'BT','BTN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Bolivia',68,'BO','BOL',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Bosnia and Herzegovina',70,'BA','BIH',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Botswana',72,'BW','BWA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Brazil',76,'BR','BRA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Belize',84,'BZ','BLZ',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Solomon Islands',90,'SB','SLB',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Virgin Islands, British',92,'VG','VGB',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Brunei Daraussalam',96,'BN','BRN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Bulgaria',100,'BG','BGR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Myanmar',104,'MM','MMR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Burundi',108,'BI','BDI',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Belarus',112,'BY','BLR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Cambodia',116,'KH','KHM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Cameroon',120,'CM','CMR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Canada',124,'CA','CAN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Cape Verde',132,'CV','CPV',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Cayman Islands',136,'KY','CYM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Central African Republic',140,'CF','CAF',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Sri Lanka',144,'LK','LKA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Chad',148,'TD','TCD',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Chile',152,'CL','CHL',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'China',156,'CN','CHN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Taiwan Province of China',158,'TW','TWN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Colombia',170,'CO','COL',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Comoros',174,'KM','COM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Congo',178,'CG','COG',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Congo, The Democratic Republic of the',180,'CD','COD',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Cook Islands',184,'CK','COK',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Costa Rica',188,'CR','CRI',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Croatia',191,'HR','HRV',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Cuba',192,'CU','CUB',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Cyprus',196,'CY','CYP',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Czech Republic',203,'CZ','CZE',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Benin',204,'BJ','BEN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Denmark',208,'DK','DNK',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Dominica',212,'DM','DMA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Dominican Republic',214,'DO','DOM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Ecuador',218,'EC','ECU',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'El Salvador',222,'SV','SLV',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Equatorial Guinea',226,'GQ','GNQ',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Ethiopia',230,'ET','ETH',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Eritrea',232,'ER','ERI',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Estonia',233,'EE','EST',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Faroe Islands',234,'FO','FRO',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Falkland Islands (Malvinas)',238,'FK','FLK',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Fiji',242,'FJ','FJI',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Finland',246,'FI','FIN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'France',250,'FR','FRA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'French Guiana',254,'GF','GUF',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'French Polynesia',258,'PF','PYF',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Djibouti',262,'DJ','DJI',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Gabon',266,'GA','GAB',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Georgia',268,'GE','GEO',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Gambia',270,'GM','GMB',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Palestinian Territory, Occupied',275,'PS','PSE',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Germany',276,'DE','DEU',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Ghana',288,'GH','GHA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Gibraltar',292,'GI','GIB',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Kiribati',296,'KI','KIR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Greece',300,'GR','GRC',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Greenland',304,'GL','GRL',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Grenada',308,'GD','GRD',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Guadeloupe',312,'GP','GLP',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Guam',316,'GU','GUM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Guatemala',320,'GT','GTM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Guinea',324,'GN','GIN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Guyana',328,'GY','GUY',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Haiti',332,'HT','HTI',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Holy See (Vatican City State)',336,'VA','VAT',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Honduras',340,'HN','HND',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Hong Kong',344,'HK','HKG',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Hungary',348,'HU','HUN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Iceland',352,'IS','ISL',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'India',356,'IN','IND',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Indonesia',360,'ID','IDN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Iran, Islamic Republic of',364,'IR','IRN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Iraq',368,'IQ','IRQ',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Ireland',372,'IE','IRL',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Israel',376,'IL','ISR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Italy',380,'IT','ITA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Cote d''Ivoire',384,'CI','CIV',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Jamaica',388,'JM','JAM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Japan',392,'JP','JPN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Kazakhstan',398,'KZ','KAZ',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Jordan',400,'JO','JOR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Kenya',404,'KE','KEN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Korea, Democratic People''s Republic of',408,'KP','PRK',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Korea, Republic of',410,'KR','KOR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Kuwait',414,'KW','KWT',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Kyrgyzstan',417,'KG','KGZ',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Lao People''s Democratic Republic',418,'LA','LAO',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Lebanon',422,'LB','LBN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Lesotho',426,'LS','LSO',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Latvia',428,'LV','LVA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Liberia',430,'LR','LBR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Libyan Arab Jamahiriya',434,'LY','LBY',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Liechtenstein',438,'LI','LIE',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Lithuania',440,'LT','LTU',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Luxembourg',442,'LU','LUX',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Macau',446,'MO','MAC',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Madagascar',450,'MG','MDG',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Malawi',454,'MW','MWI',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Malaysia',458,'MY','MYS',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Maldives',462,'MV','MDV',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Mali',466,'ML','MLI',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Malta',470,'MT','MLT',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Martinique',474,'MQ','MTQ',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Mauritania',478,'MR','MRT',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Mauritius',480,'MU','MUS',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Mexico',484,'MX','MEX',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Monaco',492,'MC','MCO',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Mongolia',496,'MN','MNG',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Moldova',498,'MD','MDA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Montserrat',500,'MS','MSR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Morocco',504,'MA','MAR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Mozambique',508,'MZ','MOZ',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Oman',512,'OM','OMN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Namibia',516,'NA','NAM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Nauru',520,'NR','NRU',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Nepal',524,'NP','NPL',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Netherlands',528,'NL','NLD',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Netherlands Antilles',530,'AN','ANT',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Aruba',533,'AW','ABW',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'New Caledonia',540,'NC','NCL',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Vanuatu',548,'VU','VUT',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'New Zealand',554,'NZ','NZL',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Nicaragua',558,'NI','NIC',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Niger',562,'NE','NER',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Nigeria',566,'NG','NGA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Niue',570,'NU','NIU',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Norfolk Island',574,'NF','NFK',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Norway',578,'NO','NOR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Northern Mariana Islands',580,'MP','MNP',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Micronesia, Federated States of',583,'FM','FSM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Marshall Islands',584,'MH','MHL',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Palau',585,'PW','PLW',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Pakistan',586,'PK','PAK',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Panama',591,'PA','PAN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Papua New Guinea',598,'PG','PNG',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Paraguay',600,'PY','PRY',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Peru',604,'PE','PER',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Philippines',608,'PH','PHL',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Pitcairn',612,'PN','PCN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Poland',616,'PL','POL',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Portugal',620,'PT','PRT',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Guinea-Bissau',624,'GW','GNB',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'East Timor',626,'TP','TMP',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Puerto Rico',630,'PR','PRI',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Qatar',634,'QA','QAT',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Réunion',638,'RE','REU',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Romania',642,'RO','ROM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Russian Federation',643,'RU','RUS',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Rwanda',646,'RW','RWA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Saint Helena',654,'SH','SHN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Saint Kitts and Nevis',659,'KN','KNA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Saint Lucia',662,'LC','LCA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Saint Pierre and Miquelon',666,'PM','SPM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Saint Vincent and the Grenadines',670,'VC','VCT',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'San Marino',674,'SM','SMR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'São Tomé and Principe',678,'ST','STP',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Saudi Arabia',682,'SA','SAU',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Senegal',686,'SN','SEN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Seychelles',690,'SC','SYC',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Sierra Leone',694,'SL','SLE',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Singapore',702,'SG','SGP',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Slovakia',703,'SK','SVK',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Viet Nam',704,'VN','VNM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Slovenia',705,'SI','SVN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Somalia',706,'SO','SOM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'South Africa',710,'ZA','ZAF',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Zimbabwe',716,'ZW','ZWE',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Spain',724,'ES','ESP',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Sudan',736,'SD','SDN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Suriname',740,'SR','SUR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Svalbard and Jan Mayen Islands',744,'SJ','SJM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Swaziland',748,'SZ','SWZ',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Sweden',752,'SE','SWE',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Switzerland',756,'CH','CHE',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Syrian Arab Republic',760,'SY','SYR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Tajikistan',762,'TJ','TJK',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Thailand',764,'TH','THA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Togo',768,'TG','TGO',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Tokelau',772,'TK','TKL',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Tonga',776,'TO','TON',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Trinidad and Tobago',780,'TT','TTO',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'United Arab Emirates',784,'AE','ARE',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Tunisia',788,'TN','TUN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Turkey',792,'TR','TUR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Turkmenistan',795,'TM','TKM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Turks and Caicos Islands',796,'TC','TCA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Tuvalu',798,'TV','TUV',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Uganda',800,'UG','UGA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Ukraine',804,'UA','UKR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Macedonia, The former Yugoslav Republic of',807,'MK','MKD',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Egypt',818,'EG','EGY',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'United Kingdom',826,'GB','GBR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'United States',840,'US','USA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Burkina Faso',854,'BF','BFA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Uruguay',858,'UY','URY',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Uzbekistan',860,'UZ','UZB',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Venezuela',862,'VE','VEN',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Channel Islands',830,'  ','   ',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Isle of Man',833,'IM','   ',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Tanzania, United Republic of',834,'TZ','TZA',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Virgin Islands, U.S.',850,'VI','VIR',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Wallis and Futuna Islands',876,'WF','WLF',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Samoa',882,'WS','WSM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Yemen',887,'YE','YEM',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Yugoslavia',891,'YU','YUG',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Zambia',894,'ZM','ZMB',NULL,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Argentina',32,'AR','ARG',540,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Anguilla',660,'AI','AIA',2,currval('country_id_seq'));
INSERT INTO "country" VALUES (nextval('country_id_seq'));
INSERT INTO "_country_data" ("_start_t","_end_t","name","un","iso2","iso3","phone","_table") VALUES (current_timestamp,NULL,'Antigua and Barbuda',28,'AG','ATG',1,currval('country_id_seq'));
UPDATE "_country_data" SET "_start_t" = current_timestamp;

-- 01-currency_type.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

insert into currency_type (id) values (1);
insert into _currency_type_data (id,_table,name,rate,format) values (1,1,'Pesos','1.00','$ %10.2f');
insert into currency_type (id) values (2);
insert into _currency_type_data (id,_table,name,rate,format) values (2,2,'Dolares','2.95','U$D %10.2f');
insert into currency_type (id) values (3);
insert into _currency_type_data (id,_table,name,rate,format) values (3,3,'Euros','3.25','EUR %10.2f');
select setval('"_currency_type_data_id_seq"',4,true);


-- 01-document_type.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

INSERT INTO document_type (id) VALUES (2);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (2, 2, 'FCAe', 'Factura A de proveedor', true);
INSERT INTO document_type (id) VALUES (3);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (3, 3, 'RCXi', 'Recibo', true);
INSERT INTO document_type (id) VALUES (4);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (4, 4, 'FCAi', 'Factura A', false);
INSERT INTO document_type (id) VALUES (6);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (6, 6, 'OPXi', 'Orden de pago', false);
INSERT INTO document_type (id) VALUES (7);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (7, 7, 'ASXi', 'Ajuste de stock', Null);
INSERT INTO document_type (id) VALUES (8);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (8, 8, 'FCBi', 'Factura B', false);
INSERT INTO document_type (id) VALUES (9);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (9, 9, 'FCCi', 'Factura C', false);
INSERT INTO document_type (id) VALUES (10);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (10, 10, 'FCBe', 'Factura B de proveedor', true);
INSERT INTO document_type (id) VALUES (11);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (11, 11, 'FCCe', 'Factura C de proveedor', true);
INSERT INTO document_type (id) VALUES (12);
INSERT INTO _document_type_data (id, _table, code, name, sign ) VALUES (12, 12, 'PPXi', 'Presupuesto', Null);
INSERT INTO document_type (id) VALUES (13);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (13, 13, 'RMXi', 'Remito X', true);
INSERT INTO document_type (id) VALUES (14);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (14, 14, 'NCAi', 'Nota de Crdito A', true);
INSERT INTO document_type (id) VALUES (15);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (15, 15, 'NCBi', 'Nota de Crdito B', true);
INSERT INTO document_type (id) VALUES (16);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (16, 16, 'NCCi', 'Nota de Crdito C', true);
INSERT INTO document_type (id) VALUES (17);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (17, 17, 'RMXe', 'Remito X de proveedor', false);
INSERT INTO document_type (id) VALUES (18);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (18, 18, 'NCAe', 'Nota de Crdito A de proveedor', false);
INSERT INTO document_type (id) VALUES (19);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (19, 19, 'NCBe', 'Nota de Crdito B de proveedor', false);
INSERT INTO document_type (id) VALUES (20);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (20, 20, 'NCCe', 'Nota de Crdito C de proveedor', false);
INSERT INTO document_type (id) VALUES (21);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (21, 21, 'DMXi', 'Devolucin de mercadera', false);
INSERT INTO document_type (id) VALUES (22);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (22, 22, 'RTVi', 'Retencin IVA', true);
INSERT INTO document_type (id) VALUES (23);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (23, 23, 'RTVe', 'Retencin IVA sufrida', true);
INSERT INTO document_type (id) VALUES (24);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (24, 24, 'RTGi', 'Retencin Ganancias', true);
INSERT INTO document_type (id) VALUES (25);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (25, 25, 'RTGe', 'Retencin Ganancias sufrida', true);
INSERT INTO document_type (id) VALUES (26);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (26, 26, 'RTBi', 'Retencin Ingresos Brutos', true);
INSERT INTO document_type (id) VALUES (27);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (27, 27, 'RTBe', 'Retencin Ingresos Brutos sufrida', true);
INSERT INTO document_type (id) VALUES (28);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (28, 28, 'RMRi', 'Remito R', true);
INSERT INTO document_type (id) VALUES (29);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (29, 29, 'RMRe', 'Remito R de proveedor', false);
INSERT INTO document_type (id) VALUES (30);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (30, 30, 'NDAi', 'Nota de Ddito A', false);
INSERT INTO document_type (id) VALUES (31);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (31, 31, 'NDBi', 'Nota de Ddito B', false);
INSERT INTO document_type (id) VALUES (32);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (32, 32, 'NDCi', 'Nota de Ddito C', false);
INSERT INTO document_type (id) VALUES (33);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (33, 33, 'NDAe', 'Nota de Ddito A de Proveedor', false);
INSERT INTO document_type (id) VALUES (34);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (34, 34, 'NDBe', 'Nota de Ddito B de Proveedor', false);
INSERT INTO document_type (id) VALUES (35);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (35, 35, 'NDCe', 'Nota de Ddito C de Proveedor', false);
INSERT INTO document_type (id) VALUES (36);
INSERT INTO _document_type_data (id, _table, code, name, sign) VALUES (36, 36, 'IMPe', 'Impuestos', false);-- 01-entity_type.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


INSERT INTO entity_type VALUES (11);
INSERT INTO _entity_type_data ("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'entity',NULL,11);
INSERT INTO entity_type VALUES (12);
INSERT INTO _entity_type_data 	("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'own_entity',11,12);
INSERT INTO entity_type VALUES (13);
INSERT INTO _entity_type_data ("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'alien_entity',11,13);
INSERT INTO entity_type VALUES (1);
INSERT INTO _entity_type_data ("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'alien_storehouse',13,1);
INSERT INTO entity_type VALUES (2);
INSERT INTO _entity_type_data ("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'contact',13,2);
INSERT INTO entity_type VALUES (3);
INSERT INTO _entity_type_data ("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'persona',13,3);
INSERT INTO entity_type VALUES (4);
INSERT INTO _entity_type_data ("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'employee',12,4);
INSERT INTO entity_type VALUES (5);
INSERT INTO _entity_type_data ("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'own_storehouse',12,5);
INSERT INTO entity_type VALUES (6);
INSERT INTO _entity_type_data ("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'own_branch',12,6);
INSERT INTO entity_type VALUES (7);
INSERT INTO _entity_type_data ("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'client',3,7);
INSERT INTO entity_type VALUES (8);
INSERT INTO _entity_type_data ("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'provider',3,8);
INSERT INTO entity_type VALUES (9);
INSERT INTO _entity_type_data ("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'bank_branch',3,9);
INSERT INTO entity_type VALUES (10);
INSERT INTO _entity_type_data ("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'bank',3,10);
INSERT INTO entity_type VALUES (14);
INSERT INTO _entity_type_data ("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'organization',12,14);
INSERT INTO entity_type VALUES (15);
INSERT INTO _entity_type_data ("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'own_pos',12,15);
INSERT INTO entity_type VALUES (16);
INSERT INTO _entity_type_data ("_start_t","_end_t","name","parent",_table) VALUES (current_timestamp,NULL,'fisco',3,16);-- 01-entity_uid_type.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


insert into entity_uid_type (id) values (1);
insert into _entity_uid_type_data (id,_table,name) values (1,1,'CUIT');
insert into entity_uid_type (id) values (2);
insert into _entity_uid_type_data (id,_table,name) values (2,2,'Ingresos Brutos');
insert into entity_uid_type (id) values (3);
insert into _entity_uid_type_data (id,_table,name) values (3,3,'Codigo');

-- 01-erm_type.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

insert into erm_type (id) values (1);
insert into _erm_type_data (id,_table,name) values (1,1,'email enviado');
insert into erm_type (id) values (2);
insert into _erm_type_data (id,_table,name) values (2,2,'email recibido');
insert into erm_type (id) values (3);
insert into _erm_type_data (id,_table,name) values (3,3,'conversacin');
insert into erm_type (id) values (4);
insert into _erm_type_data (id,_table,name) values (4,4,'llamada telef');
insert into erm_type (id) values (5);
insert into _erm_type_data (id,_table,name) values (5,5,'entrevista');
insert into erm_type (id) values (6);
insert into _erm_type_data (id,_table,name) values (6,6,'documentos');


-- 01-movement_type.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

insert into movement_type (id) values (1);
insert into _movement_type_data (id,_table,name) values (1,1,'Cuenta Corriente');

-- 01-price_type.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

insert into price_type (id) values (nextval('price_type_id_seq'));
insert into _price_type_data (id,_table,name,is_active) values (nextval('_price_type_data_id_seq'),currval('price_type_id_seq'),'Lista base','t');

-- 01-product_code_type.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

insert into product_code_type (id) values (1);
insert into _product_code_type_data (id,_table,name) values (1,1,'Interno');

-- 01-tax.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

insert into tax (id) values (1);
insert into _tax_data (id,_table,name) values (1,1,'IVA');
insert into tax (id) values (2);
insert into _tax_data (id,_table,name) values (2,2,'IVA RNI');
insert into tax (id) values (3);
insert into _tax_data (id,_table,name) values (3,3,'Otros');
insert into tax (id) values (4);
insert into _tax_data (id,_table,name) values (4,4,'Percepcin IVA');
insert into tax (id) values (5);
insert into _tax_data (id,_table,name) values (5,5,'Percepcin IIBB');
insert into tax (id) values (6);
insert into _tax_data (id,_table,name) values (6,6,'Impuestos Internos');
-- 01-relation_type.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

INSERT INTO relation_type VALUES (1);
INSERT INTO _relation_type_data ("_start_t","_end_t","name",_table) VALUES (current_timestamp,NULL,'branch_of_organization',1);
INSERT INTO relation_type VALUES (2);
INSERT INTO _relation_type_data ("_start_t","_end_t","name",_table) VALUES (current_timestamp,NULL,'pos_of_branch',2);
INSERT INTO relation_type VALUES (3);
INSERT INTO _relation_type_data ("_start_t","_end_t","name",_table) VALUES (current_timestamp,NULL,'storehouse_of_branch',3);
INSERT INTO relation_type VALUES (4);
INSERT INTO _relation_type_data ("_start_t","_end_t","name",_table) VALUES (current_timestamp,NULL,'pos_of_storehouse',4);

-- 01-document_relation_type.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

INSERT INTO document_relation_type ("id","name") VALUES (1,'invoice_of_despatch_note');
--INSERT INTO document_relation_type ("id","name") VALUES (2,'_of_');
--INSERT INTO document_relation_type ("id","name") VALUES (3,'_of_');
-- 01-admin_period_type.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

begin;
INSERT INTO admin_period_type ("id","name") VALUES (1,'libro iva');
commit;-- 02-entity_type_uid_conf.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

INSERT INTO entity_type_uid_conf ("entity_type","entity_uid_type") VALUES (7,3);
INSERT INTO entity_type_uid_conf ("entity_type","entity_uid_type") VALUES (8,3);
INSERT INTO entity_type_uid_conf ("entity_type","entity_uid_type") VALUES (2,3);
INSERT INTO entity_type_uid_conf ("entity_type","entity_uid_type") VALUES (4,3);
INSERT INTO entity_type_uid_conf ("entity_type","entity_uid_type") VALUES (16,3);-- 02-tax_profile.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


insert into tax_profile (id) values (1);
insert into _tax_profile_data (id,_table,profile,tax) values (1,1,'Responsable Inscripto',1);
insert into tax_profile (id) values (2);
insert into _tax_profile_data (id,_table,profile,tax) values (2,2,'Responsable No Inscripto',1);
insert into tax_profile (id) values (3);
insert into _tax_profile_data (id,_table,profile,tax) values (3,3,'Exento',1);
insert into tax_profile (id) values (4);
insert into _tax_profile_data (id,_table,profile,tax) values (4,4,'Monotributo',1);
insert into tax_profile (id) values (5);
insert into _tax_profile_data (id,_table,profile,tax) values (5,5,'Consumidor Final',1);
insert into tax_profile (id) values (6);
insert into _tax_profile_data (id,_table,profile,tax) values (6,6,'Responsable Inscripto',2);
insert into tax_profile (id) values (7);
insert into _tax_profile_data (id,_table,profile,tax) values (7,7,'Responsable No Inscripto',2);
-- 03-tax_formula.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

insert into tax_formula (id) values (1);
insert into _tax_formula_data (id,_table,from_profile,to_profile,graft)values (1,1,1,1,'def graft(amount,taxPercent):\n\treturn amount * taxPercent / 100\n');
insert into tax_formula (id) values (2);
insert into _tax_formula_data (id,_table,from_profile,to_profile,graft)values (2,2,1,2,'def graft(amount,taxPercent):\n\treturn amount * taxPercent / 100\n');
insert into tax_formula (id) values (3);
insert into _tax_formula_data (id,_table,from_profile,to_profile,graft)values (3,3,1,3,'def graft(amount,taxPercent):\n\treturn amount * taxPercent / 100\n');
insert into tax_formula (id) values (4);
insert into _tax_formula_data (id,_table,from_profile,to_profile,graft)values (4,4,1,4,'def graft(amount,taxPercent):\n\treturn amount * taxPercent / 100\n');
insert into tax_formula (id) values (5);
insert into _tax_formula_data (id,_table,from_profile,to_profile,graft)values (5,5,1,5,'def graft(amount,taxPercent):\n\treturn amount * taxPercent / 100\n');
insert into tax_formula (id) values (6);
insert into _tax_formula_data (id,_table,from_profile,to_profile,graft)values (6,6,6,7,'def graft(amount,taxPercent):\n\treturn amount * taxPercent / 100 /2 \n');



-- 03-organization.sql
-- Copyright 2003 Fundacion Via Libre
--
-- This file is part of PAPO.
-- 
-- PAPO is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
-- 
-- PAPO is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with PAPO; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

insert into entity (id) values (2);
insert into _entity_data (name,_table) values ('Empresa',2);
insert into own_entity (id) values (2);
insert into _own_entity_data (_start_t,_table,entity) values (now(),2,2);
insert into organization (id) values (nextval('organization_id_seq'));
insert into _organization_data (_table,own_entity) values (currval('organization_id_seq'),2);
--cargo tax profile
insert into persona_tax_profile (id) values (nextval('persona_tax_profile_id_seq'));
insert into _persona_tax_profile_data (_table,persona,tax_profile) values (currval('persona_tax_profile_id_seq'),2,1);
--cargo entity types
insert into entity_types (id) values (nextval('entity_types_id_seq'));
insert into _entity_types_data (_table,entity,entity_type) values (currval('entity_types_id_seq'),2,14);
--ajusto los valores de seq
select setval('"entity_id_seq"',3,true);
select setval('"own_entity_id_seq"',3,true);




INSERT INTO __papo__ (key, value) VALUES ('version', '0.2.0');
